/*
 * (c) Copyright 2015 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>
#include <stdbool.h>

#include <argyllio.h>

struct patch_set *
argyllio_ti3_parse(const char filename[],
                   bool error_is_fatal)
{
    double *bands = NULL;
    double *fields = NULL;
    char **strings = NULL;
    struct patch_set *ti3 = NULL;
    FILE *stream = fopen(filename, "rt");
    if (stream == NULL) {
        fprintf(stderr, "could not open \"%s\" for reading: %s.\n", filename, strerror(errno));
        goto fail;
    }
    char s[8192];
    bool found_begin = false;
    int class_offset = -1;
    int xyz_offset = -1;
    int rgb_offset = -1;
    int spec_offset = -1;
    int spec_band_count = -1;
    double spectral_scaling = 1.0;
    const char *spec_prefixes[] = {
        "SPEC_",
        "SPECTRAL_NM_",
        NULL
    };
    while (fgets(s, sizeof(s), stream) != NULL) {
        if (strstr(s, "CTI3") == s) {
            spectral_scaling = 100.0;
        } else if (strstr(s, "SPEC_SCALE") == s) {
            spectral_scaling = atof(&s[strlen("SPEC_SCALE")]);
        } else if (strstr(s, "BEGIN_DATA_FORMAT\n") == s) {
            if (fgets(s, sizeof(s), stream) == NULL) {
                fprintf(stderr, "Could not read data format\n");
                goto fail;
            }
            char *p = s;
            int field = 0;
            for (;;) {
                while (isblank(*p)) p++;
                if (*p == '\n' || *p == '\0') {
                    break;
                }
                if (strncmp(p, "RGB_R", 5) == 0) {
                    rgb_offset = field;
                    p = strstr(p, "RGB_B");
                    if (p == NULL) {
                        fprintf(stderr, "Missing RGB_B or bad field order\n");
                        goto fail;
                    }
                    p += 5;
                    field += 3;
                } else if (strncmp(p, "XYZ_X", 5) == 0) {
                    xyz_offset = field;
                    p = strstr(p, "XYZ_Z");
                    if (p == NULL) {
                        fprintf(stderr, "Missing XYZ_Z or bad field order\n");
                        goto fail;
                    }
                    p += 5;
                    field += 3;
                } else if (strncmp(p, "SAMPLE_CLASS", 12) == 0) {
                    class_offset = field;
                    p += 12;
                    field++;
                } else {
                    int sp;
                    for (sp = 0; spec_prefixes[sp] != NULL; sp++) {
                        const int prefix_len = strlen(spec_prefixes[sp]);
                        if (strncmp(p, spec_prefixes[sp], prefix_len) == 0) {
                            spec_offset = field;
                            bands = malloc((strlen(s) / 7) * sizeof(bands[0]));
                            int bc = 0;
                            for (;;) {
                                p += prefix_len;
                                bands[bc++] = (double)atoi(p);
                                while (!isblank(*p) && *p != '\0') p++;
                                while (isblank(*p)) p++;
                                if (strncmp(p, spec_prefixes[sp], prefix_len) != 0) {
                                    break;
                                }
                            }
                            for (int i = 0; i < bc; i++) {
                                if (bands[i] <= 0 || (i > 0 && bands[i] <= bands[i-1])) {
                                    fprintf(stderr, "Bad spectral bands (%f).\n", bands[i]);
                                    goto fail;
                                }
                            }
                            if (bc < 3) {
                                fprintf(stderr, "Too few spectral bands.\n");
                                goto fail;
                            }
                            spec_band_count = bc;
                            field += bc;
                            break;
                        }
                    }
                    if (spec_prefixes[sp] == NULL) {
                        while (!isblank(*p) && *p != '\0') p++;
                        field++;
                    }
                }
            }
        } else if (strstr(s, "BEGIN_DATA\n") == s) {
            found_begin = true;
            break;
        }
    }
    if (xyz_offset < 0) {
        fprintf(stderr, "Missing XYZ fields, creating dummy values.\n");
    }
    if (rgb_offset < 0) {
        fprintf(stderr, "Missing RGB fields, creating dummy values.\n");
    }
    if (!found_begin) {
        fprintf(stderr, "BEGIN_DATA not found.\n");
        goto fail;
    }
    int alloc_size = 1;
    ti3 = calloc(1, sizeof(*ti3) + alloc_size * sizeof(struct cam2xyz_patch));
    fields = malloc(sizeof(s) * sizeof(fields[0]));
    strings = malloc(sizeof(s) * sizeof(strings[0]));
    for (;;) {
        if (fgets(s, sizeof(s), stream) == NULL) {
            break;
        }
        if (strstr(s, "END_DATA") == s) {
            break;
        }
        char *p = s;
        while (isblank(*p)) p++;
        int fi = 0;
        const int i = ti3->patch_count;
        while (*p != '\0' && *p != '\n') {
            if (*p == '\"') {
                fields[fi] = 0;
                strings[fi++] = ++p;
                if ((p = strchr(p, '\"')) == NULL) {
                    fprintf(stderr, "Bad string on sample data line %d.\n", i + 1);
                    goto fail;
                }
                *p = '\0'; p++;
            } else {
                strings[fi] = NULL;
                fields[fi++] = atof(p);
                while (!isblank(*p) && *p != '\0') p++;
            }
            while (isblank(*p)) p++;
        }
        if (fi < xyz_offset + 3 || fi < rgb_offset + 3 || fi < spec_offset + spec_band_count) {
            fprintf(stderr, "Too few data fields on sample data line %d.\n", i + 1);
            goto fail;
        }
        if (xyz_offset >= 0) {
            ti3->patch[i].xyz.v[0] = fields[xyz_offset+0];
            ti3->patch[i].xyz.v[1] = fields[xyz_offset+1];
            ti3->patch[i].xyz.v[2] = fields[xyz_offset+2];
        }
        if (rgb_offset >= 0) {
            ti3->patch[i].cam.v[0] = fields[rgb_offset+0];
            ti3->patch[i].cam.v[1] = fields[rgb_offset+1];
            ti3->patch[i].cam.v[2] = fields[rgb_offset+2];
        }
        if (class_offset >= 0) {
            if (strings[class_offset] == NULL) {
                fprintf(stderr, "Bad class name on sample data line %d.\n", i + 1);
                goto fail;
            }
            ti3->patch[i].class = -1;
            int j;
            for (j = 0; j < TARGET_MAX_CLASS_COUNT && ti3->class_names[j][0] != '\0'; j++) {
                if (strcasecmp(ti3->class_names[j], strings[class_offset]) == 0) {
                    ti3->patch[i].class = j;
                    break;
                }
            }
            if (ti3->patch[i].class == -1) {
                if (j == TARGET_MAX_CLASS_COUNT) {
                    fprintf(stderr, "Too many class names, max %d supported.\n", TARGET_MAX_CLASS_COUNT);
                    goto fail;
                }
                if (strlen(strings[class_offset]) >= TARGET_CLASS_NAME_SIZE) {
                    fprintf(stderr, "Too long class name on sample data line %d.\n", i+1);
                    goto fail;
                }
                strcpy(ti3->class_names[j], strings[class_offset]);
                ti3->patch[i].class = j;
            }
        }
        if (spec_offset >= 0) {
            if (spectral_scaling != 1.0) {
                for (int j = 0; j < spec_band_count; j++) {
                    fields[spec_offset+j] /= spectral_scaling;
                }
            }
            ti3->patch[i].spectrum = spec_alloc1(bands, &fields[spec_offset], spec_band_count);
            if (ti3->patch[i].spectrum == NULL) {
                fprintf(stderr, "Bad spectral data at sample data line %d.\n", i + 1);
                goto fail;
            }
        }
        ti3->patch_count++;
        if (ti3->patch_count == alloc_size) {
            alloc_size += 256;
            ti3 = realloc(ti3, sizeof(*ti3) + alloc_size * sizeof(struct cam2xyz_patch));
            memset(&ti3->patch[ti3->patch_count], 0, 256 * sizeof(struct cam2xyz_patch));
        }
    }
    fclose(stream);
    free(bands);
    free(fields);
    ti3 = realloc(ti3, sizeof(*ti3) + ti3->patch_count * sizeof(struct cam2xyz_patch));
    return ti3;
fail:
    if (stream != NULL) {
        fclose(stream);
    }
    free(ti3);
    free(bands);
    free(fields);
    free(strings);
    fprintf(stderr, "failed to parse \"%s\".\n", filename);
    if (error_is_fatal) {
        exit(EXIT_FAILURE);
    }
    return NULL;
}

bool
argyllio_ti3_write(const char filename[],
                   bool print_spectral_data,
                   const struct patch_set *ps,
                   bool error_is_fatal)
{
    int band_start = 450;
    int band_end = 650;
    if (print_spectral_data) {
        for (int i = 0; i < ps->patch_count; i++) {
            if (ps->patch[i].spectrum == NULL) {
                fprintf(stderr, "At least one patch lacks spectrum information, cannot write spectral data.\n");
                if (error_is_fatal) exit(EXIT_FAILURE);
                return false;
            }
            int bs = spec_band_start(ps->patch[i].spectrum);
            int be = spec_band_end(ps->patch[i].spectrum);
            if (bs < band_start) band_start = bs;
            if (be > band_end) band_end = be;
        }
    }
    const int band_spacing = 5;
    while (band_start % band_spacing != 0) band_start++;
    while (band_end % band_spacing != 0) band_end--;
    int band_count = (band_end - band_start) / band_spacing + 1;

    FILE *stream = fopen(filename, "wt");
    if (stream == NULL) {
        fprintf(stderr, "could not open \"%s\" for writing: %s.\n", filename, strerror(errno));
        if (error_is_fatal) exit(EXIT_FAILURE);
        return false;
    }

    struct tm tm;
    time_t tt = time(NULL);
    localtime_r(&tt, &tm);
    char timestr[128];
    asctime_r(&tm, timestr);
    timestr[strlen(timestr)-1] = '\0';
    fprintf(stream,
            "CTI3   \n"
            "\n"
            "DESCRIPTOR \"Argyll Calibration Target chart information 3\"\n"
            "ORIGINATOR \"DCamProf\"\n"
            "CREATED \"%s\"\n"
            "KEYWORD \"DEVICE_CLASS\"\n"
            "DEVICE_CLASS \"INPUT\"\n"
            "KEYWORD \"COLOR_REP\"\n"
            "COLOR_REP \"RGB_XYZ\"\n",
            timestr);
    if (print_spectral_data) {
        fprintf(stream,
                "KEYWORD \"SPECTRAL_BANDS\"\n"
                "SPECTRAL_BANDS \"%d\"\n"
                "KEYWORD \"SPECTRAL_START_NM\"\n"
                "SPECTRAL_START_NM \"%f\"\n"
                "KEYWORD \"SPECTRAL_END_NM\"\n"
                "SPECTRAL_END_NM \"%f\"\n",
                band_count,
                (double)band_start,
                (double)band_end);
    }
    fprintf(stream,
            "\n"
            "KEYWORD \"SAMPLE_CLASS\"\n"
        );
    if (print_spectral_data) {
        for (int i = 0; i < band_count; i++) {
            fprintf(stream, "KEYWORD \"SPEC_%d\"\n", band_start + band_spacing * i);
        }
    }
    fprintf(stream,
            "NUMBER_OF_FIELDS %d\n"
            "BEGIN_DATA_FORMAT\n"
            "SAMPLE_ID SAMPLE_CLASS RGB_R RGB_G RGB_B XYZ_X XYZ_Y XYZ_Z",
            7 + (print_spectral_data ? band_count : 0));
    if (print_spectral_data) {
        for (int i = 0; i < band_count; i++) {
            fprintf(stream, " SPEC_%d", band_start + band_spacing * i);
        }
    }
    fprintf(stream, "\nEND_DATA_FORMAT\n\n");

    fprintf(stream, "NUMBER_OF_SETS %zd\n"
            "BEGIN_DATA\n", ps->patch_count);
    for (int i = 0; i < ps->patch_count; i++) {
        const struct cam2xyz_patch *p = &ps->patch[i];
        fprintf(stream, "%d \"%s\" %.6G %.6G %.6G %.6G %.6G %.6G",
                i+1,
                ps->class_names[p->class],
                100*p->cam.v[0], 100*p->cam.v[1], 100*p->cam.v[2],
                100*p->xyz.v[0], 100*p->xyz.v[1], 100*p->xyz.v[2]);
        if (print_spectral_data) {
            for (int i = 0; i < band_count; i++) {
                if (p->spectrum != NULL) {
                    double e = 100.0 * spec_at(p->spectrum, band_start + band_spacing * i);
                    e *= 1e6;
                    e = round(e);
                    e /= 1e6;
                    char buf[32];
                    sprintf(buf, " %.6G", e);
                    if (strchr(buf, 'E') != NULL) {
                        fprintf(stream, " %.6f", e);
                    } else {
                        fprintf(stream, " %.6G", e);
                    }
                } else {
                    fprintf(stream, " 0");
                }
            }
        }
        fprintf(stream, "\n");
    }
    fprintf(stream, "END_DATA\n");
    fclose(stream);

    return true;
}
