/*
 * (c) Copyright 2015 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#ifndef COLMATH_H_
#define COLMATH_H_

#include <math.h>

#include <spectrum.h>
#include <matmath.h>

enum exif_lightsource {
    lsUnknown = 0,
    lsDaylight = 1,
    lsFluorescent = 2,
    lsTungsten = 3,
    lsFlash = 4,
    lsFineWeather = 9,
    lsCloudyWeather = 10,
    lsShade = 11,
    lsDaylightFluorescent = 12,	// D  5700 - 7100K
    lsDayWhiteFluorescent = 13,	// N  4600 - 5500K
    lsCoolWhiteFluorescent = 14, // W  3800 - 4500K
    lsWhiteFluorescent = 15, // WW 3250 - 3800K
    lsWarmWhiteFluorescent = 16, // L  2600 - 3250K
    lsStandardLightA = 17,
    lsStandardLightB = 18,
    lsStandardLightC = 19,
    lsD55 = 20,
    lsD65 = 21,
    lsD75 = 22,
    lsD50 = 23,
    lsISOStudioTungsten = 24,
    lsOther = 255
};

struct observer {
    spectrum_t *cmf[3];
};

static inline double
srgb_gamma_forward(const double x)
{
    if (x <= 0.0031308) {
        return x * 12.92;
    } else {
        return 1.055 * pow (x, 1.0 / 2.4) - 0.055;
    }
}

static inline double
srgb_gamma_inverse(const double y)
{
    if (y <= 0.0031308 * 12.92) {
        return y * (1.0 / 12.92);
    } else {
        return pow ((y + 0.055) * (1.0 / 1.055), 2.4);
    }
}

v3
xyz2srgb(v3 xyz, v3 wp);

uint32_t
xyz2srgb_u24(v3 xyz, v3 wp);

static inline double
lab_ft_forward(double t)
{
    if (t >= 8.85645167903563082e-3) {
        return pow(t, 1.0/3.0);
    } else {
        return t * (841.0/108.0) + 4.0/29.0;
    }
}

static inline double
lab_ft_inverse(double t)
{
    if (t >= 0.206896551724137931) {
        return t*t*t;
    } else {
        return 108.0 / 841.0 * (t - 4.0/29.0);
    }
}

const char *
xyz2name(v3 xyz, v3 wp, char buf[]);

v3
xyz2lab(v3 xyz, v3 wp);

v3
lab2xyz(v3 lab, v3 wp);

static inline v3
xyz2xyY(const v3 xyz)
{
    double x = xyz.v[0], y = xyz.v[1], z = xyz.v[2];
    return (v3){{ x / (x+y+z), y / (x+y+z), y }};
}

static inline v3
xyY2xyz(const v3 xyY)
{
    double x = xyY.v[0], y = xyY.v[1], Y = xyY.v[2];
    return (v3){{ x*Y / y, Y, (1-x-y)*Y/y }};
}

static inline v3
xyz2uvY(const v3 xyz)
{
    double x = xyz.v[0], y = xyz.v[1], z = xyz.v[2];
    return (v3){{ 4*x / (x + 15*y + 3*z), 9*y / (x + 15*y + 3*z), y }};
}

static inline v3
uvY2xyz(const v3 uvY)
{
    double u = uvY.v[0], v = uvY.v[1], Y = uvY.v[2];
    return (v3){{ Y*9*u / (4*v), Y, Y * (12 - 3*u - 20*v) / (4*v) }};
}

v3
xyz2luv(v3 xyz, v3 wp);

v3
luv2xyz(v3 luv, v3 wp);

double
spec2xyz_weight(const struct observer *o, const spectrum_t *illuminant);

v3
spec2xyz_ill(const spectrum_t *s, const struct observer *o, const spectrum_t *illuminant, double weight);

v3
spec2xyz(const spectrum_t *s, const struct observer *o);

v3
xy2xyz_lightest(double x,
                double y,
                const v3 wp);

v3
uv2xyz_lightest(double u,
                double v,
                const v3 wp);

m3x3
bradford_map_white(v3 src_wp,
                   v3 dst_wp);

double
ciede2000(v3 xyz1, v3 xyz2, v3 wp);

double
ciede2000_lch(v3 xyz1, v3 xyz2, v3 wp,
              double Kl, double Kc, double Kh);

double
simple_de(v3 v1, v3 v2);

#endif
