/*
 * (c) Copyright 2015 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <inttypes.h>
#include <stdbool.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <assert.h>
#include <strings.h>

#include <lcms2.h>

#include <nmsimplex.h>

#include <target.h>
#include <colmath.h>
#include <matmath.h>
#include <matopt.h>
#include <profio.h>
#include <target.h>
#include <spectraldb.h>
#include <xyz2spec.h>
#include <observers.h>
#include <argyllio.h>
#include <jsonio.h>
#include <dcamprof.h>
#include <dngref.h>
#include <lut.h>

#define debug(...)
//#define debug(...) fprintf(stderr, __VA_ARGS__)

static struct {
    v3 d50;
    int pe_cmp_idx;
} glob;

FILE *
open_log_file(const char dir[],
              const char name[])
{
    if (dir == NULL) {
        return NULL;
    }
    mkdir(dir, 0777);
    char path[strlen(dir) + strlen(name) + 32];
    sprintf(path, "%s/%s", dir, name);
    FILE *stream = fopen(path, "wt");
    if (stream == NULL) {
        fprintf(stderr, "Could not open \"%s\" for writing: %s.\n", path, strerror(errno));
        exit(EXIT_FAILURE);
    }
    return stream;
}

enum ca_transform {
    CAT_CAT02,
    CAT_BRADFORD
};

static v3
chromatic_adaptation_transform(enum ca_transform cat,
                               const v3 xyz,
                               const v3 src_wp,
                               const v3 dst_wp)
{
    if (memcmp(&src_wp, &dst_wp, sizeof(src_wp)) == 0) {
        return xyz;
    }
    switch (cat) {
    case CAT_CAT02: {
        cmsViewingConditions vc1, vc2;
        cmsJCh Out;
        cmsCIEXYZ In;
        cmsHANDLE h1, h2;
        vc1.whitePoint.X = src_wp.v[0] * 100;
        vc1.whitePoint.Y = src_wp.v[1] * 100;
        vc1.whitePoint.Z = src_wp.v[2] * 100;
        vc1.Yb = 20;
        vc1.La = 20;
        vc1.surround = AVG_SURROUND;
        vc1.D_value = 1.0;
        h1 = cmsCIECAM02Init(0, &vc1);
        vc2.whitePoint.X = dst_wp.v[0] * 100;
        vc2.whitePoint.Y = dst_wp.v[1] * 100;
        vc2.whitePoint.Z = dst_wp.v[2] * 100;
        vc2.Yb = 20;
        vc2.La = 20;
        vc2.surround = AVG_SURROUND;
        vc2.D_value = 1.0;
        h2 = cmsCIECAM02Init(0, &vc2);
        In.X = xyz.v[0] * 100;
        In.Y = xyz.v[1] * 100;
        In.Z = xyz.v[2] * 100;
        cmsCIECAM02Forward(h1, &In, &Out);
        cmsCIECAM02Reverse(h2, &Out, &In);
        cmsCIECAM02Done(h1);
        cmsCIECAM02Done(h2);
        In.X /= 100;
        In.Y /= 100;
        In.Z /= 100;
        /*{

            fprintf(stderr, "%f %f %f | %f %f %f\n", xyz.v[0],xyz.v[1],xyz.v[2],In.X, In.Y, In.Z);
            v3 out = chromatic_adaptation_transform(CAT_BRADFORD, xyz, src_wp, dst_wp);
            fprintf(stderr, "1 %f %f %f | %f %f %f\n", xyz.v[0],xyz.v[1],xyz.v[2],out.v[0],out.v[1],out.v[2]);
            }*/
        return (v3){{ In.X, In.Y, In.Z }};
    }
    case CAT_BRADFORD: {
        /*
        cmsCIEXYZ xyz_ = { .X = xyz.v[0], .Y = xyz.v[1], .Z = xyz.v[2] };
        cmsCIEXYZ src_wp_ = { .X = src_wp.v[0], .Y = src_wp.v[1], .Z = src_wp.v[2] };
        cmsCIEXYZ dst_wp_ = { .X = dst_wp.v[0], .Y = dst_wp.v[1], .Z = dst_wp.v[2] };
        cmsCIEXYZ res;
        cmsAdaptToIlluminant(&res, &src_wp_, &dst_wp_, &xyz_);
        return (v3){{ res.X, res.Y, res.Z }};
        */
        return m3x3_mul_v3(bradford_map_white(src_wp, dst_wp), xyz);
    }
    default: abort();
    }
    return xyz;
}

static struct patch_set *
exclude_darker_patches(struct patch_set *ps,
                       double uv_de_lim)
{
    double max_y = 0;
    for (int i = 0; i < ps->patch_count; i++) {
        if (ps->patch[i].xyz.v[1] > max_y) {
            max_y = ps->patch[i].xyz.v[1];
        }
    }
    for (int i = 0; i < ps->patch_count; i++) {
        for (int j = 0; j < ps->patch_count; j++) {
            if (i == j) continue;
            v3 xyz1 = ps->patch[i].xyz;
            v3 xyz2 = ps->patch[j].xyz;
            for (int k = 0; k < 3; k++) {
                if (xyz1.v[k] != 0) xyz1.v[k] /= max_y;
                if (xyz2.v[k] != 0) xyz2.v[k] /= max_y;
            }
            v3 ltu1 = xyz2lutspace(xyz1);
            v3 ltu2 = xyz2lutspace(xyz2);
            ltu1.v[0] = ltu2.v[0];
            double uv_de = simple_de(ltu1, ltu2);
            if (uv_de < uv_de_lim) {
                if (xyz1.v[0]+xyz1.v[1]+xyz1.v[2] < xyz2.v[0]+xyz2.v[1]+xyz2.v[2]) {
                    ps->patch[i].class = -1;
                } else {
                    ps->patch[j].class = -1;
                }
            }
        }
    }
    int orig_count = ps->patch_count;
    ps->patch_count = 0;
    for (int i = 0; i < orig_count; i++) {
        if (ps->patch[i].class >= 0) {
            if (i != ps->patch_count) {
                ps->patch[ps->patch_count] = ps->patch[i];
            }
            ps->patch_count++;
        }
    }
    if (ps->patch_count < orig_count) {
        fprintf(stderr,
                //2345678901234567890123456789012345678901234567890123456789012345678901234567890
                "Excluded %zd patches from target due to lighter patch with same chroma.\n"
                "  %zd patches left\n",
                orig_count - ps->patch_count, ps->patch_count);
    }
    return ps;
}

static struct patch_set *
exclude_weak_signal_patches(struct patch_set *ps,
                            bool verbose)
{
    double max_cam = 0;
    for (int i = 0; i < ps->patch_count; i++) {
        for (int j = 0; j < 3; j++) {
            if (ps->patch[i].cam.v[j] > max_cam) {
                max_cam = ps->patch[i].cam.v[j];
            }
        }
    }
    int orig_count = ps->patch_count;
    ps->patch_count = 0;
    for (int i = 0; i < orig_count; i++) {
        double cam[3];
        cam[0] = ps->patch[i].cam.v[0] / max_cam;
        cam[1] = ps->patch[i].cam.v[1] / max_cam;
        cam[2] = ps->patch[i].cam.v[2] / max_cam;
        if (cam[0] >= 0.0001 || cam[1] >= 0.0001 || cam[2] >= 0.0001) {
            if (i != ps->patch_count) {
                ps->patch[ps->patch_count] = ps->patch[i];
            }
            ps->patch_count++;
        }
    }
    if (ps->patch_count < orig_count && verbose) {
        fprintf(stderr,
                //2345678901234567890123456789012345678901234567890123456789012345678901234567890
                "Excluded %zd patches from target due to weak camera signal. %zd patches left\n", orig_count - ps->patch_count, ps->patch_count);
    }
    return ps;
}

static void
assign_optimization_weights(struct patch_set *ps,
                            const double class_weight[],
                            const double class_de_lim[],
                            const double (*class_de_klch)[3],
                            bool skip_base_weighting,
                            bool skip_matrix_de,
                            const v3 illuminant_xyz)
{
    double wsum = 0;
    bool de_differs = false;
#pragma omp parallel for
    for (int i = 0; i < ps->patch_count; i++) {
        if (!skip_base_weighting) {
            double de = 0.5; // start with some minimum
            for (int j = 0; j < ps->patch_count; j++) {
                if (i == j) continue;
                de += ciede2000_lch(ps->patch[j].xyz, ps->patch[i].xyz, illuminant_xyz, 1, 1, 1);
            }
            ps->patch[i].w = de;
        } else {
            ps->patch[i].w = 1.0;
        }
        ps->patch[i].w *= class_weight[ps->patch[i].class];
#pragma omp critical
        {
            wsum += ps->patch[i].w;
        }
        ps->patch[i].de = class_de_lim[ps->patch[i].class];
        ps->patch[i].k_lch[0] = class_de_klch[ps->patch[i].class][0];
        ps->patch[i].k_lch[1] = class_de_klch[ps->patch[i].class][1];
        ps->patch[i].k_lch[2] = class_de_klch[ps->patch[i].class][2];
        ps->patch[i].m_de = ps->patch[i].de;
        if (ps->patch[i].de != ps->patch[0].de) {
            de_differs = true;
        }
    }
    for (int i = 0; i < ps->patch_count; i++) {
        //fprintf(stderr, "%f\n", ps->patch[i].w);
        ps->patch[i].w /= wsum;
        if (!de_differs || skip_matrix_de) {
            // in the matrix case it makes no sense to not go for the best solution if all have the same DE
            ps->patch[i].m_de = 0;
        }
    }
}

static int
find_whitest_patch(const struct patch_set *ps,
                   v3 wp)
{
    int illuminant_class = -1;
    int w_idx = -1;
    for (int i = 0; i < TARGET_MAX_CLASS_COUNT; i++) {
        if (strcasecmp(ps->class_names[i], "illuminant") == 0) {
            illuminant_class = i;
            break;
        }
    }
    if (illuminant_class != -1) {
        for (int i = 0; i < ps->patch_count; i++) {
            if (ps->patch[i].class == illuminant_class) {
                w_idx = i;
                break;
            }
        }
    } else {
        bool *cand = calloc(1, sizeof(*cand) * ps->patch_count);
        bool has_cand = false;
        for (int i = 0; i < ps->patch_count; i++) {
            v3 xyz = ps->patch[i].xyz;
            if (xyz.v[1] < 0.1) {  // skip very dark patches
                continue;
            }
            v3 lab = xyz2lab(v3_norm(xyz), wp);
            double ab = sqrt(lab.v[1] * lab.v[1] + lab.v[2] * lab.v[2]);
            if (ab < 1.0) {
                has_cand = true;
                cand[i] = true;
            }
        }
        double min_ab = 100000000;
        double min_ab_all = 100000000;
        double max_y = 0;
        int w_idx_all = -1;
        for (int i = 0; i < ps->patch_count; i++) {
            v3 xyz = ps->patch[i].xyz;
            if (has_cand) {
                if (!cand[i]) continue;
                if (xyz.v[1] > max_y) {
                    max_y = xyz.v[1];
                    w_idx = i;
                }
            } else {
                v3 lab = xyz2lab(v3_norm(xyz), wp);
                double ab = sqrt(lab.v[1] * lab.v[1] + lab.v[2] * lab.v[2]);
                if (xyz.v[1] >= 0.1) {
                    if (ab < min_ab) {
                        min_ab = ab;
                        w_idx = i;
                    }
                }
                if (ab < min_ab_all) {
                    min_ab_all = ab;
                    w_idx_all = i;
                }
            }
        }
        free(cand);
        if (w_idx == -1) w_idx = w_idx_all;
    }
    return w_idx;
}

struct patch_set *
normalize_patch_set(const struct patch_set *ps_)
{
    struct patch_set *ps = malloc(sizeof(*ps) + sizeof(struct cam2xyz_patch) * ps_->patch_count);
    memcpy(ps, ps_, sizeof(*ps) + sizeof(struct cam2xyz_patch) * ps_->patch_count);

    double max_rgb = 0;
    double max_xyz = 0;
    for (int i = 0; i < ps->patch_count; i++) {
        for (int j = 0; j < 3; j++) {
            if (ps->patch[i].cam.v[j] > max_rgb) max_rgb = ps->patch[i].cam.v[j];
            if (ps->patch[i].xyz.v[j] > max_xyz) max_xyz = ps->patch[i].xyz.v[j];
        }
    }
    for (int i = 0; i < ps->patch_count; i++) {
        ps->patch[i].spectrum = spec_copy(ps_->patch[i].spectrum);
        for (int j = 0; j < 3; j++) {
            ps->patch[i].xyz.v[j] /= max_xyz;
            ps->patch[i].cam.v[j] /= max_rgb;
        }
    }
    return ps;
}

static bool
write_dcam_profile(const char filename[],
                   const struct dcam_profile *prof,
                   bool error_is_fatal)
{
    FILE *stream = fopen(filename, "wt");
    if (stream == NULL) {
        fprintf(stderr, "Could not open \"%s\" for writing: %s.\n", filename, strerror(errno));
        if (error_is_fatal) exit(EXIT_FAILURE);
        return false;
    }
    fprintf(stream, "{\n");
    fprintf(stream, "  \"CalibrationIlluminant1\": \"%s\",\n", exif_lightsource_ntoa(prof->illuminant_name));
    if (prof->illuminant_spec != NULL) {
        jsonio_spectrum_print(stream, "  ", "CalibrationIlluminantSpectrum1", prof->illuminant_spec);
        fprintf(stream, ",\n");
    }
    jsonio_3x3matrix_print(stream, "  ", "ColorMatrix1", prof->color_matrix);
    fprintf(stream, ",\n");
    jsonio_3x3matrix_print(stream, "  ", "ForwardMatrix1", prof->forward_matrix);
    fprintf(stream, ",\n");
    chromalut_json_print(stream, "  ", "ChromaLUT1", prof->lut);
    fprintf(stream, "\n}\n");
    fclose(stream);

    return true;
}

void
dcam_profile_delete(struct dcam_profile *prof)
{
    if (prof == NULL) {
        return;
    }
    chromalut_delete(prof->lut);
    free(prof);
}

struct dcp_lut {
    v3 *hsm;
    uint32_t *hsmdims;
    bool srgb_gamma;
};

struct patch_error {
    int idx;
    v3 cxyz;
    double errors[4];
};

struct patch_error_report {
    enum {
        LOOKUP_MATRIX,
        LOOKUP_LUT_NATIVE,
        LOOKUP_DCP_HSM
    } lookup_type;
    struct {
        double avg;
        double mean;
        double p90;
        double max;
        int worst[5];
        int best[5];
    } errors[4];
    int patch_count;
    const struct patch_set *ps;
    v3 ps_wp;
    struct patch_error e[];
};



static int
pe_cmp(const void *a_, const void *b_)
{
    const struct patch_error *a = a_;
    const struct patch_error *b = b_;
    if (a->errors[glob.pe_cmp_idx] < b->errors[glob.pe_cmp_idx]) {
        return -1;
    }
    if (a->errors[glob.pe_cmp_idx] > b->errors[glob.pe_cmp_idx]) {
        return 1;
    }
    return 0;
}

static struct patch_error_report *
test_patch_matching(const m3x3 cam2xyz,
                    chromalut_t *lut,
                    struct dcp_lut *dcp_lut,
                    const struct patch_set *ps,
                    const v3 ps_wp)
{
    struct patch_error_report *pe = calloc(1, sizeof(*pe) + sizeof(pe->e[0]) * ps->patch_count);
    double sum[4] = {0,0,0,0};
    if (lut) {
        pe->lookup_type = LOOKUP_LUT_NATIVE;
    } else if (dcp_lut && dcp_lut->hsm) {
        pe->lookup_type = LOOKUP_DCP_HSM;
    } else {
        pe->lookup_type = LOOKUP_MATRIX;
    }
    for (int i = 0; i < ps->patch_count; i++) {
        v3 cxyz = m3x3_mul_v3(cam2xyz, ps->patch[i].cam);
        if (lut) {
            cxyz = chromalut_lookup(lut, cxyz);
        } else if (dcp_lut && dcp_lut->hsm) {
            cxyz = dcp_lookup(dcp_lut->hsm, dcp_lut->hsmdims, dcp_lut->srgb_gamma, cxyz);
        }
        v3 xyz = ps->patch[i].xyz;
        pe->e[i].idx = i;
        pe->e[i].cxyz = cxyz;
        pe->e[i].errors[0] = ciede2000(xyz, cxyz, ps_wp);
        pe->e[i].errors[1] = ciede2000_lch(xyz, cxyz, ps_wp, 1, 100000, 100000);
        pe->e[i].errors[2] = ciede2000_lch(xyz, cxyz, ps_wp, 100000, 1, 100000);
        pe->e[i].errors[3] = ciede2000_lch(xyz, cxyz, ps_wp, 100000, 100000, 1);
        for (int j = 0; j < 4; j++) {
            if (pe->e[i].errors[j] > pe->errors[j].max) {
                pe->errors[j].max = pe->e[i].errors[j];
            }
            sum[j] += pe->e[i].errors[j];
        }
    }
    for (int j = 0; j < 4; j++) {
        pe->errors[j].avg = sum[j] / ps->patch_count;
    }
    pe->patch_count = ps->patch_count;
    pe->ps = ps;
    pe->ps_wp = ps_wp;
    struct patch_error *spe = malloc(sizeof(pe->e[0]) * ps->patch_count);
    memcpy(spe, pe->e, sizeof(pe->e[0]) * ps->patch_count);
    for (int j = 3; j >= 0; j--) {
        glob.pe_cmp_idx = j;
        qsort(spe, pe->patch_count, sizeof(spe[0]), pe_cmp);
        pe->errors[j].mean = spe[pe->patch_count/2].errors[j];
        pe->errors[j].p90 = spe[(9*pe->patch_count)/10].errors[j];
        for (int k = 0; k < sizeof(pe->errors[j].worst)/sizeof(pe->errors[j].worst[0]) && k < ps->patch_count; k++) {
            pe->errors[j].worst[k] = spe[pe->patch_count - 1 - k].idx;
        }
        for (int k = 0; k < sizeof(pe->errors[j].best)/sizeof(pe->errors[j].best[0]) && k < ps->patch_count; k++) {
            pe->errors[j].best[k] = spe[k].idx;
        }
    }
    free(spe);
    return pe;
}

static void
print_common_spectra(const char report_dir[],
                     const struct observer *obs,
                     const spectrum_t *ill,
                     const struct dcam_ssf *ssf)
{
    if (obs != NULL) {
        FILE *stream1 = open_log_file(report_dir, "cmf-x.dat");
        FILE *stream2 = open_log_file(report_dir, "cmf-y.dat");
        FILE *stream3 = open_log_file(report_dir, "cmf-z.dat");
        if (stream1 == NULL) return;
        spec_print(obs->cmf[0], stream1);
        spec_print(obs->cmf[1], stream2);
        spec_print(obs->cmf[2], stream3);
        fclose(stream1);
        fclose(stream2);
        fclose(stream3);
    }
    if (ssf != NULL) {
        FILE *stream1 = open_log_file(report_dir, "ssf-r.dat");
        FILE *stream2 = open_log_file(report_dir, "ssf-g.dat");
        FILE *stream3 = open_log_file(report_dir, "ssf-b.dat");
        if (stream1 == NULL) return;
        spec_print(ssf->ssf.cmf[0], stream1);
        spec_print(ssf->ssf.cmf[1], stream2);
        spec_print(ssf->ssf.cmf[2], stream3);
        fclose(stream1);
        fclose(stream2);
        fclose(stream3);
    }
    if (ill != NULL) {
        FILE *stream1 = open_log_file(report_dir, "illuminant.dat");
        if (stream1 == NULL) return;
        spec_print(ill, stream1);
        fclose(stream1);
    }
    {
        FILE *stream1 = open_log_file(report_dir, "illuminant-d50.dat");
        if (stream1 == NULL) return;
        spec_print(spectraldb_illuminant(lsD50), stream1);
        fclose(stream1);
    }
}

static void
print_patch_matching_errors(const struct patch_error_report *pe,
                            bool print_worst_best,
                            const char report_dir[],
                            const char report_filename[],
                            FILE *stream)
{
    const char *name = "Matrix";
    const char *pad  = "      ";
    if (pe->lookup_type == LOOKUP_LUT_NATIVE) {
        name = "Native LUT";
        pad  = "          ";
    }
    if (pe->lookup_type == LOOKUP_DCP_HSM) {
        name = "DCP HSM LUT";
        pad  = "           ";
    }
    fprintf(stream, "%s patch match average DE %.2f, DE LCh %.2f %.2f %.2f\n",
            name, pe->errors[0].avg, pe->errors[1].avg, pe->errors[2].avg, pe->errors[3].avg);
    fprintf(stream, "%s                mean DE %.2f, DE LCh %.2f %.2f %.2f\n",
            pad, pe->errors[0].mean, pe->errors[1].mean, pe->errors[2].mean, pe->errors[3].mean);
    fprintf(stream, "%s                 p90 DE %.2f, DE LCh %.2f %.2f %.2f\n",
            pad, pe->errors[0].p90, pe->errors[1].p90, pe->errors[2].p90, pe->errors[3].p90);
    fprintf(stream, "%s                 max DE %.2f, DE LCh %.2f %.2f %.2f\n",
            pad, pe->errors[0].max, pe->errors[1].max, pe->errors[2].max, pe->errors[3].max);

    char cnbuf[128];
    if (report_filename != NULL) {
        FILE *stream1 = open_log_file(report_dir, report_filename);
        if (stream1 != NULL) {
            for (int i = 0; i < pe->patch_count; i++) {
                const struct patch_error *e = &pe->e[i];
                const struct cam2xyz_patch *p = &pe->ps->patch[e->idx];
                fprintf(stream1, "   %5d RGB %.3f %.3f %.3f => XYZ %.3f %.3f %.3f (%.3f %.3f %.3f) DE %.2f DE LCh %.2f %.2f %.2f (%s)\n", e->idx,
                        p->cam.v[0], p->cam.v[1], p->cam.v[2], p->xyz.v[0], p->xyz.v[1], p->xyz.v[2],
                        e->cxyz.v[0], e->cxyz.v[1], e->cxyz.v[2],
                        e->errors[0],
                        e->errors[1],
                        e->errors[2],
                        e->errors[3],
                        xyz2name(p->xyz, pe->ps_wp, cnbuf));
            }
            fclose(stream1);
        }
    }

    if (!print_worst_best) {
        return;
    }
    int worst_count = sizeof(pe->errors[0].worst)/sizeof(pe->errors[0].worst[0]);
    if (worst_count > pe->patch_count) worst_count = pe->patch_count;
    for (int j = 0; j < 4; j++) {
        const char *error_type[] = {
            "Overall DE",
            "Lightness DE",
            "Chroma DE",
            "Hue DE",
        };
        fprintf(stream, "%d worst patches for %s:\n", worst_count, error_type[j]);
        for (int k = 0; k < worst_count; k++) {
            const struct patch_error *e = &pe->e[pe->errors[j].worst[k]];
            const struct cam2xyz_patch *p = &pe->ps->patch[e->idx];
            fprintf(stream, "   %5d RGB %.3f %.3f %.3f => XYZ %.3f %.3f %.3f (%.3f %.3f %.3f) DE %.2f DE LCh %.2f %.2f %.2f (%s)\n", e->idx,
                    p->cam.v[0], p->cam.v[1], p->cam.v[2], p->xyz.v[0], p->xyz.v[1], p->xyz.v[2],
                    e->cxyz.v[0], e->cxyz.v[1], e->cxyz.v[2],
                    e->errors[0],
                    e->errors[1],
                    e->errors[2],
                    e->errors[3],
                    xyz2name(p->xyz, pe->ps_wp, cnbuf));
        }
    }
    int best_count = sizeof(pe->errors[0].best)/sizeof(pe->errors[0].best[0]);
    if (best_count > pe->patch_count) best_count = pe->patch_count;
    fprintf(stream, "%d best patches for Overall DE:\n", best_count);
    for (int k = 0; k < best_count; k++) {
        const struct patch_error *e = &pe->e[pe->errors[0].best[k]];
        const struct cam2xyz_patch *p = &pe->ps->patch[e->idx];
        fprintf(stream, "   %5d RGB %.3f %.3f %.3f => XYZ %.3f %.3f %.3f (%.3f %.3f %.3f) DE %.2f DE LCh %.2f %.2f %.2f (%s)\n", e->idx,
                p->cam.v[0], p->cam.v[1], p->cam.v[2], p->xyz.v[0], p->xyz.v[1], p->xyz.v[2],
                e->cxyz.v[0], e->cxyz.v[1], e->cxyz.v[2],
                e->errors[0],
                e->errors[1],
                e->errors[2],
                e->errors[3],
                xyz2name(p->xyz, pe->ps_wp, cnbuf));
    }
}

static void
print_pm_errors(const m3x3 cam2xyz,
                chromalut_t *lut,
                struct dcp_lut *dcp_lut,
                const struct patch_set *ps,
                const v3 ps_wp,
                bool print_worst_best,
                const char report_dir[],
                const char report_filename[])
{
    struct patch_error_report *pe = test_patch_matching(cam2xyz, lut, dcp_lut, ps, ps_wp);
    print_patch_matching_errors(pe, print_worst_best, report_dir, report_filename, stderr);
    free(pe);
}

static double
test_matrix_lsq(const m3x3 *cam2xyz,
                const v3 *wp,
                const struct patch_set *ps,
                void *arg)
{
    double sum = 0;
    for (size_t i = 0; i < ps->patch_count; i++) {
        v3 v = m3x3_mul_v3(*cam2xyz, ps->patch[i].cam);
        double de = ciede2000_lch(v, ps->patch[i].xyz, *wp, ps->patch[i].k_lch[0], ps->patch[i].k_lch[1], ps->patch[i].k_lch[2]);
        if (de > ps->patch[i].m_de) {
            double wde = ps->patch[i].w * de;
            sum += wde * wde;
        }
    }
    return sqrt(sum); // optimizer works better with slower approach to zero
}

static m3x3
find_cam2xyz_matrix(enum exif_lightsource calib_ls,
                    const v3 calib_wp,
                    const struct patch_set *ps,
                    bool use_slow_matrix_optimizer)
{
    fprintf(stderr, "Finding a camera raw RGB to CIE XYZ matrix for illuminant %s...\n",
            exif_lightsource_ntoa(calib_ls));
    m3x3 cam2xyz;
    if (use_slow_matrix_optimizer) {
        cam2xyz = matopt_find_matrix_brute_force(calib_wp, ps, test_matrix_lsq, NULL);
    } else {
        cam2xyz = matopt_find_matrix(calib_wp, ps, test_matrix_lsq, NULL);
    }

    // preserve white point
    int w_idx = find_whitest_patch(ps, calib_wp);
    double de = ciede2000(v3_norm(ps->patch[w_idx].xyz), v3_norm(calib_wp), calib_wp);
    debug("whitest patch difference from calibration illuminant whitepoint %f\n", de);
    if (de > 2) {
        fprintf(stderr,
                //2345678901234567890123456789012345678901234567890123456789012345678901234567890
                "Warning: whitest (most neutral) patch in target differs DE %.2f from\n"
                "  calibration illuminant, matrix precision may suffer.\n", de);
    } else if (de > 1e-5) {
        fprintf(stderr,
                //2345678901234567890123456789012345678901234567890123456789012345678901234567890
                "Whitest patch in target differs DE %.2f from calibration illuminant,\n"
                "  close enough to calculate whitepoint preservation.\n", de);
    }
    struct patch_set *wb_ps = calloc(1, sizeof(*wb_ps) + sizeof(wb_ps->patch[0]));
    wb_ps->patch_count = 1;
    wb_ps->patch[0].m_de = 0.0;
    wb_ps->patch[0].de = 0.0;
    wb_ps->patch[0].k_lch[0] = 1;
    wb_ps->patch[0].k_lch[1] = 1;
    wb_ps->patch[0].k_lch[2] = 1;
    wb_ps->patch[0].w = 1.0;
    wb_ps->patch[0].xyz = ps->patch[w_idx].xyz;
    wb_ps->patch[0].cam = ps->patch[w_idx].cam;
    matopt_refine_matrix(&cam2xyz, calib_wp, wb_ps, test_matrix_lsq, NULL);
    v3 calcxyz = m3x3_mul_v3(cam2xyz, wb_ps->patch[0].cam);
    de = ciede2000(wb_ps->patch[0].xyz, calcxyz, calib_wp);
    if (de > 1e-5) {
        fprintf(stderr,
                //2345678901234567890123456789012345678901234567890123456789012345678901234567890
                "Warning: did not succeeed to match whitest patch, DE %.2f.\n"
                "  Precision may suffer.\n", de);
    }
    debug("DE %f\n", de);
    debug("%+.4f %+.4f %+.4f\n", cam2xyz.v[0][0], cam2xyz.v[0][1], cam2xyz.v[0][2]);
    debug("%+.4f %+.4f %+.4f\n", cam2xyz.v[1][0], cam2xyz.v[1][1], cam2xyz.v[1][2]);
    debug("%+.4f %+.4f %+.4f\n\n", cam2xyz.v[2][0], cam2xyz.v[2][1], cam2xyz.v[2][2]);
    free(wb_ps);
    return cam2xyz;
}

static m3x3
calculate_color_matrix(enum exif_lightsource calib_ls,
                       const v3 calib_wp,
                       const struct patch_set *ps_,
                       bool use_slow_matrix_optimizer,
                       const char report_dir[])
{
    struct patch_set *ps = normalize_patch_set(ps_);
    m3x3 cam2xyz = find_cam2xyz_matrix(calib_ls, calib_wp, ps, use_slow_matrix_optimizer);

    fprintf(stderr, "Inverting to get ColorMatrix:\n  {\n");
    m3x3 xyz2cam = invert_m3x3(cam2xyz);
    jsonio_3x3matrix_print(stderr, "    ", "ColorMatrix1", xyz2cam);
    fprintf(stderr, "\n  }\n");
    print_pm_errors(cam2xyz, NULL, NULL, ps, calib_wp, false, report_dir, "cm-patch-errors.txt");

    target_delete_patch_set(ps);
    return xyz2cam;
}

static m3x3
calculate_forward_matrix(const m3x3 *cam2xyz_precalc,
                         const struct patch_set *ps_,
                         bool use_slow_matrix_optimizer,
                         v3 *calc_wb,
                         const char report_dir[])
{
    const v3 d50 = glob.d50;
    struct patch_set *ps = normalize_patch_set(ps_);
    m3x3 cam2xyz;
    if (cam2xyz_precalc == NULL) {
        cam2xyz = find_cam2xyz_matrix(lsD50, d50, ps, use_slow_matrix_optimizer);
    } else {
        fprintf(stderr, "Using previously calculated RGB to XYZ D50 matrix.\n");
        cam2xyz = *cam2xyz_precalc;
    }

    fprintf(stderr, "Applying white-balance to get ForwardMatrix:\n");

    v3 camd50 = m3x3_mul_v3(invert_m3x3(cam2xyz), d50);
    if (calc_wb != NULL) {
        *calc_wb = v3_norm(camd50);
    }
    v3 mul;
    mul.v[0] = 1.0 / camd50.v[0];
    mul.v[1] = 1.0 / camd50.v[1];
    mul.v[2] = 1.0 / camd50.v[2];
    cam2xyz = m3x3_mul_m3x3(cam2xyz, invert_m3x3(v3_asdiagonal(mul)));
    for (int i = 0; i < ps->patch_count; i++) {
        ps->patch[i].cam.v[0] /= camd50.v[0];
        ps->patch[i].cam.v[1] /= camd50.v[1];
        ps->patch[i].cam.v[2] /= camd50.v[2];
    }
    struct patch_set *ps1 = normalize_patch_set(ps);
    target_delete_patch_set(ps);
    ps = ps1;

    fprintf(stderr, "  {\n");
    jsonio_3x3matrix_print(stderr, "    ", "ForwardMatrix1", cam2xyz);
    fprintf(stderr, "\n  }\n");
    print_pm_errors(cam2xyz, NULL, NULL, ps, d50, false, report_dir, "fm-patch-errors.txt");

    target_delete_patch_set(ps);
    return cam2xyz;
}

static double
norm_l_de(v3 xyz1, v3 xyz2, v3 wp)
{
    v3 lab1 = xyz2lab(xyz1, wp);
    v3 lab2 = xyz2lab(xyz2, wp);
    lab2.v[0] = lab1.v[0];
    xyz1 = lab2xyz(lab1, wp);
    xyz2 = lab2xyz(lab2, wp);
    return ciede2000(xyz1, xyz2, wp);
}

static int
double_cmp(const void *a_, const void *b_)
{
    double a = *(const double *)a_;
    double b = *(const double *)b_;
    if (a < b) return -1;
    if (a > b) return 1;
    return 0;
}

static void
print_lookup_reports(const char report_dir[],
                     chromalut_t *lut,
                     struct dcp_lut *dcp_lut,
                     const struct dcam_ssf *dcam_ssf,
                     const m3x3 cam2xyz,
                     const struct patch_set *ps_,
                     const v3 *ps_wp)
{
    if (report_dir == NULL) {
        return;
    }
    if (lut != NULL) {
        FILE *stream1 = open_log_file(report_dir, "nve-lut.dat");
        FILE *stream2 = open_log_file(report_dir, "nve-ref.dat");
        FILE *stream3 = open_log_file(report_dir, "nve-lutv.dat");
        if (stream1 == NULL) {
            return;
        }
        for (double u = 0.0; u <= 0.6; u += 0.01) {
            for (double v = 0.0; v <= 0.6; v += 0.01) {
                v3 diff = chromalut_lookup_diff(lut, u, v);
                double Ldiff = diff.v[0];
                double udiff = diff.v[1];
                double vdiff = diff.v[2];
                fprintf(stream1, "%f %f %f\n", u+udiff, v+vdiff, Ldiff);
                fprintf(stream2, "%f %f %f\n", u, v, 0.0);
                fprintf(stream3, "%f %f %f %f %f %f\n", u, v, 0.0, udiff, vdiff, Ldiff);
            }
            fprintf(stream1, "\n");
            fprintf(stream2, "\n");
            fprintf(stream3, "\n");
        }
        fclose(stream1);
        fclose(stream2);
        fclose(stream3);
    }
    if (dcp_lut && dcp_lut->hsm && dcp_lut->hsmdims[2] > 1) {
        fprintf(stderr, "No support for writing plot data for 3D HueSatMap LUTs.\n");
    }
    if (dcp_lut && dcp_lut->hsm && dcp_lut->hsmdims[2] == 1) {
        FILE *stream = open_log_file(report_dir, "hsm-ref.dat");
        if (stream == NULL) {
            return;
        }
        FILE *stream1 = open_log_file(report_dir, "hsm-lut.dat");
        FILE *stream2 = open_log_file(report_dir, "hsm-lutv.dat");
        m3x3 rgb2xyz = invert_m3x3(dcp_xyz_D50_to_prophoto_rgb);
        int hcount = dcp_lut->hsmdims[0];
        int scount = dcp_lut->hsmdims[1];
        for (int h = 0; h < hcount; h++) {
            for (int s = 0; s < scount; s++) {
                const v3 *entry = &dcp_lut->hsm[scount*h+s];
                v3 hsv = {{ (h+0.5) / hcount * 6.0, (s + 0.5) / scount, 0.5 }};
                v3 rgb = dcp_hsv2rgb(hsv);
                v3 xyz = m3x3_mul_v3(rgb2xyz, rgb);
                v3 ltu = xyz2lutspace(xyz);
                fprintf(stream, "%f %f 0.0\n", ltu.v[1], ltu.v[2]);
                v3 hsv1 = hsv;
                hsv1.v[0] += (entry->v[0] / 360.0 * 6.0);
                if (hsv1.v[0] < 0) hsv1.v[0] += 6.0;
                hsv1.v[1] *= entry->v[1];
                hsv1.v[2] *= entry->v[2];
                rgb = dcp_hsv2rgb(hsv1);
                xyz = m3x3_mul_v3(rgb2xyz, rgb);
                v3 ltu1 = xyz2lutspace(xyz);
                double ldiff = (ltu1.v[0]/ltu.v[0]-1)/10.0;
                fprintf(stream1, "%f %f %f\n", ltu1.v[1], ltu1.v[2], ldiff);
                fprintf(stream2, "%f %f %f %f %f %f\n", ltu.v[1], ltu.v[2], 0.0, ltu1.v[1] - ltu.v[1], ltu1.v[2] - ltu.v[2], ldiff);
            }
        }
        if (stream) {
            fclose(stream);
            fclose(stream1);
            fclose(stream2);
        }
    }

    if (ps_ != NULL) {
        FILE *stream1, *stream2 = NULL, *stream3 = NULL, *stream3b = NULL, *stream4 = NULL, *stream5 = NULL, *stream5b = NULL;
        stream1 = open_log_file(report_dir, "target-mtx.dat");
        if (stream1 == NULL) {
            return;
        }
        if (lut != NULL) {
            stream2 = open_log_file(report_dir, "target-nve-lut.dat");
            stream3 = open_log_file(report_dir, "target-nve-lutvm.dat");
            stream3b = open_log_file(report_dir, "target-nve-lutve.dat");
        }
        if (dcp_lut && dcp_lut->hsm) {
            stream4 = open_log_file(report_dir, "target-hsm-lut.dat");
            stream5 = open_log_file(report_dir, "target-hsm-lutvm.dat");
            stream5b = open_log_file(report_dir, "target-hsm-lutve.dat");
        }
        struct patch_set *ps = normalize_patch_set(ps_);
        for (int i = 0; i < ps->patch_count; i++) {
            v3 xyz = ps->patch[i].xyz;
            v3 cxyz = m3x3_mul_v3(cam2xyz, ps->patch[i].cam);
            v3 mtx = xyz2lutspace(cxyz);
            v3 ref = xyz2lutspace(xyz);
            fprintf(stream1, "%f %f %f\n", mtx.v[1], mtx.v[2], mtx.v[0]);
            if (lut) {
                v3 ltu = xyz2lutspace(chromalut_lookup(lut, cxyz));
                fprintf(stream2, "%f %f %f\n", ltu.v[1], ltu.v[2], ltu.v[0]);
                fprintf(stream3, "%f %f %f %f %f %f\n", mtx.v[1], mtx.v[2], mtx.v[0], ltu.v[1]-mtx.v[1], ltu.v[2]-mtx.v[2], ltu.v[0]-mtx.v[0]);
                fprintf(stream3b, "%f %f %f %f %f %f\n", ref.v[1], ref.v[2], ref.v[0], ltu.v[1]-ref.v[1], ltu.v[2]-ref.v[2], ltu.v[0]-ref.v[0]);
            }
            if (dcp_lut && dcp_lut->hsm) {
                v3 ltu = xyz2lutspace(dcp_lookup(dcp_lut->hsm, dcp_lut->hsmdims, dcp_lut->srgb_gamma, cxyz));
                fprintf(stream4, "%f %f %f\n", ltu.v[1], ltu.v[2], ltu.v[0]);
                fprintf(stream5, "%f %f %f %f %f %f\n", mtx.v[1], mtx.v[2], mtx.v[0], ltu.v[1]-mtx.v[1], ltu.v[2]-mtx.v[2], ltu.v[0]-mtx.v[0]);
                fprintf(stream5b, "%f %f %f %f %f %f\n", ref.v[1], ref.v[2], ref.v[0], ltu.v[1]-ref.v[1], ltu.v[2]-ref.v[2], ltu.v[0]-ref.v[0]);
            }
        }
        fclose(stream1);
        if (lut != NULL) {
            fclose(stream2);
            fclose(stream3);
            fclose(stream3b);
        }
        if (dcp_lut && dcp_lut->hsm) {
            fclose(stream4);
            fclose(stream5);
            fclose(stream5b);
        }

        if (dcam_ssf != NULL) {
            /*
unset key
set palette rgbformula 30,31,32
set cbrange [0:500]
plot 'ssf-csep.dat' pt 7 ps 2 lt palette, 'gmt-locus.dat' using 1:2:4 w l lw 4 lc rgb var, 'gmt-srgb.dat' w l, 'gmt-prophoto.dat' w l, 'gmt-pointer.dat' using 1:2:4 w l lw 2 lc rgb var
            */
            double max_uvdist;
            {
                double *uvdists = malloc(ps->patch_count * sizeof(*uvdists));
                for (int i = 0; i < ps->patch_count; i++) {
                    v3 xyz = ps->patch[i].xyz;
                    v3 ltu = xyz2lutspace(xyz);
                    double mindist = 10000;
                    for (int j = 0; j < ps->patch_count; j++) {
                        if (i == j) continue;
                        v3 ltu2 = xyz2lutspace(ps->patch[j].xyz);
                        double uvdist = sqrt((ltu2.v[2]-ltu.v[2])*(ltu2.v[2]-ltu.v[2]) + (ltu2.v[1]-ltu.v[1])*(ltu2.v[1]-ltu.v[1]));
                        if (uvdist < mindist) mindist = uvdist;
                    }
                    uvdists[i] = mindist;
                }
                qsort(uvdists, ps->patch_count, sizeof(double), double_cmp);
                double mean_min_dist = uvdists[ps->patch_count/2];
                free(uvdists);
                max_uvdist = mean_min_dist * 1.9;
            }

            FILE *stream1 = open_log_file(report_dir, "ssf-csep.dat");
            for (int i = 0; i < ps->patch_count; i++) {
                v3 xyz = ps->patch[i].xyz;
                v3 ltu = xyz2lutspace(xyz);
                v3 cam = ps->patch[i].cam;
                double minde = 1000;
                double mindiff = 10000000000;
                int test_count = 0;
                for (int j = 0; j < ps->patch_count; j++) {
                    if (i == j) continue;
                    {
                        v3 ltu2 = xyz2lutspace(ps->patch[j].xyz);
                        double uvdist = sqrt((ltu2.v[2]-ltu.v[2])*(ltu2.v[2]-ltu.v[2]) + (ltu2.v[1]-ltu.v[1])*(ltu2.v[1]-ltu.v[1]));
                        if (uvdist > max_uvdist) {
                            continue;
                        }
                    }
                    double diff = 0;
                    v3 cam1 = v3_norm(cam);
                    v3 cam2 = v3_norm(ps->patch[j].cam);
                    for (int j = 0; j < 3; j++) {
                        diff += fabs(cam1.v[j] - cam2.v[j]);
                    }
                    diff *= v3_max(cam);
                    diff /= 3;
                    double de = norm_l_de(xyz, ps->patch[j].xyz, *ps_wp);
                    if (de < minde) {
                        minde = de;
                    }
                    diff /= de;
                    if (diff < mindiff) {
                        mindiff = diff;
                    }
                    test_count++;
                }
                if (minde == 1000) {
                    continue;
                }
                // for one delta E chromaticity difference, the signal changes this much
                double sig_diff_per_de = mindiff;

                double csep = sig_diff_per_de * 65536;

                fprintf(stream1, "%f %f %f\n", ltu.v[1], ltu.v[2], csep);
            }
            fclose(stream1);
        }
        target_delete_patch_set(ps);
    }

    if (dcam_ssf != NULL) {
        /*
        FILE *stream1 = open_log_file(report_dir, "ssf-stretch.dat");
        if (stream1 == NULL) {
            return;
        }
        const m3x3 xyz2cam = invert_m3x3(cam2xyz);
        for (double u = 0.0; u <= 0.6; u += 0.01) {
            for (double v = 0.0; v <= 0.6; v += 0.01) {
                v3 xyz = uv2xyz_lightest(u, v, glob.d50);
                v3 cxyz = xyz;
                if (lut) {
                    //cxyz = chromalut_reverse_lookup(lut, xyz);
                }
                v3 cam = m3x3_mul_v3(xyz2cam, cxyz);
                v3 camlut = xyz2lutspace(cam);
                if (!isfinite(camlut.v[0]) || camlut.v[0] < 0 || camlut.v[1] < 0 || camlut.v[2] < 0) {
                    camlut.v[0] = 0;
                    camlut.v[1] = u;
                    camlut.v[2] = v;
                }
                fprintf(stream1, "%f %f %f\n", camlut.v[1], camlut.v[2], camlut.v[0]);
            }
            fprintf(stream1, "\n");
        }
        fclose(stream1);
        */

        /*
        if (0) {
            const int range = 100;
            int skip_count = 0;
            int hist[61][61];
            memset(hist, 0, sizeof(hist));
#pragma omp parallel for
            for (int r = 0; r < range; r++) {
                for (int g = 0; g < range; g++) {
                    for (int b = 0; b < range; b++) {
                        v3 cam = {{ (double)r / (range-1), (double)g / (range-1), (double)b / (range-1)}};
                        v3 cxyz = m3x3_mul_v3(cam2xyz, cam);
                        v3 ltu;
                        if (lut) {
                            ltu = xyz2lutspace(chromalut_lookup(lut, cxyz));
                        } else if (dcp_lut && dcp_lut->hsm) {
                            ltu = xyz2lutspace(dcp_lookup(dcp_lut->hsm, dcp_lut->hsmdims, dcp_lut->srgb_gamma, cxyz));
                        } else {
                            ltu = xyz2lutspace(cxyz);
                        }
                        int u = (int)(ltu.v[1] * 100);
                        int v = (int)(ltu.v[2] * 100);
#pragma omp critical
                        {
                            if (u < 0 || u > 60 || v < 0 || v > 60) {
                                skip_count++;
                            } else {
                                hist[u][v]++;
                            }
                        }
                    }
                }
            }
            fprintf(stderr, "%d\n", skip_count);
            FILE *stream1 = open_log_file(report_dir, "ssf-hist.dat");
            for (int u = 0; u <= 60; u++) {
                for (int v = 0; v <= 60; v++) {
                    fprintf(stream1, "%f %f %f\n", u*0.01, v*0.01, (double)hist[u][v]);
                }
                fprintf(stream1, "\n");
            }
            fclose(stream1);
        }
        */
    }
}

static v3 *
make_huesatmap(uint32_t hsmdims[3],
               int hcount,
               int scount,
               chromalut_t *lut,
               const char report_dir[])
{
    m3x3 rgb2xyz = invert_m3x3(dcp_xyz_D50_to_prophoto_rgb);
    v3 *hsm = malloc(hcount * scount * sizeof(*hsm));
    for (int h = 0; h < hcount; h++) {
        for (int s = 0; s < scount; s++) {
            v3 hsv = {{ (h+0.5) / hcount * 6.0, (s + 0.5) / scount, 0.5 }};
            v3 ref_hsv;
            { // calculate corrected HSV value
                v3 rgb = dcp_hsv2rgb(hsv);
                v3 xyz = m3x3_mul_v3(rgb2xyz, rgb);
                xyz = chromalut_lookup(lut, xyz);
                rgb = m3x3_mul_v3(dcp_xyz_D50_to_prophoto_rgb, xyz);
                ref_hsv = dcp_rgb2hsv(rgb);
            }
            //fprintf(stderr, "%f %f %f\n", hsv.v[0], hsv.v[1], hsv.v[2]);
            //fprintf(stderr, "%f %f %f\n", ref_hsv.v[0], ref_hsv.v[1], ref_hsv.v[2]);

            v3 *entry = &hsm[scount*h+s];
            entry->v[0] = (ref_hsv.v[0] - hsv.v[0]) / 6.0 * 360.0;
            if (entry->v[0] < -180) {
                entry->v[0] += 360.0;
            } else if (entry->v[0] > 180) {
                entry->v[0] -= 360.0;
            }
            entry->v[1] = ref_hsv.v[1] / hsv.v[1];
            entry->v[2] = ref_hsv.v[2] / hsv.v[2];
            //fprintf(stderr, "%f %f %f\n", entry->v[0], entry->v[1], entry->v[2]);
        }
        //fprintf(stderr, "\n");
    }
    hsmdims[0] = hcount;
    hsmdims[1] = scount;
    hsmdims[2] = 1;

    return hsm;
}

static struct profio_dcp *
make_dcp(struct dcam_profile *prof,
         const char *camera_model,
         bool make_lut,
         int hcount,
         int scount,
         bool skip_fwd,
         bool skip_observer_mapping,
         const char report_dir[])
{
    struct profio_dcp *dcp = calloc(1, sizeof(*dcp));
    dcp->cm[0] = prof->color_matrix;
    dcp->fm[0] = prof->forward_matrix;
    dcp->has.fm = !skip_fwd;
    if (make_lut) {
        dcp->huesatmap[0] = make_huesatmap(dcp->hsmdims, hcount, scount, prof->lut, report_dir);
    }
    if (!skip_fwd) {
        v3 test_d50 = m3x3_mul_v3(dcp->fm[0], (v3){{1,1,1}});
        double de = ciede2000(test_d50, dcp_d50(), dcp_d50());
        if (de > 0.1 && skip_observer_mapping) {
            fprintf(stderr,
                    //2345678901234567890123456789012345678901234567890123456789012345678901234567890
                    "Warning: forward matrix D50 whitepoint differs %.2f DE from the 1931_2 D50 used\n"
                    "  by Adobe's DNG Profiles. The profile may not work in some DCP products.\n",
                    de);
        }
        // if de is small enough we make a final adjustment to "perfect" anyway
        if (!skip_observer_mapping || de <= 0.1) {
            m3x3 fm = m3x3_mul_m3x3(dcp->fm[0], bradford_map_white(test_d50, dcp_d50()));
            if (de > 0.1) {
                fprintf(stderr,
                        //2345678901234567890123456789012345678901234567890123456789012345678901234567890
                        "The D50 whitepoint in the forward matrix differs %.2f DE from the 1931_2 D50\n"
                        "  used by most DCP compatible products, and was remapped to match.\n", de);
            }
            dcp->fm[0] = fm;
        }
    }
    dcp->camera_model = strdup(camera_model);
    dcp->profile_name = strdup("Generated by DCamProf");
    dcp->copyright = strdup("Public Domain");
    dcp->embed_policy = pepNoRestrictions;
    dcp->has.embed_policy = true;
    dcp->ill[0] = prof->illuminant_name;
    dcp->ill_count = 1;

    m3x3 dummy;
    struct dcp_lut dcp_lut;
    dcp_lut.hsm = dcp->huesatmap[0];
    dcp_lut.hsmdims = dcp->hsmdims;
    dcp_lut.srgb_gamma = dcp->huesatmap_srgb_gamma;
    memset(&dummy, 0, sizeof(dummy));
    print_lookup_reports(report_dir, prof->lut, &dcp_lut, NULL, dummy, NULL, NULL);
    return dcp;
}

static struct dcam_profile *
make_profile(enum exif_lightsource calib_ls,
             const v3 calib_wp,
             const spectrum_t *calib_spec,
             const m3x3 *precalc_cm,
             const m3x3 *precalc_fm,
             bool use_slow_matrix_optimizer,
             double min_chroma_dist,
             const double regularization[3],
             const struct patch_set *ps,
             const struct patch_set *psD50,
             const struct observer *obs,
             const struct dcam_ssf *dcam_ssf,
             const char report_dir[])
{
    v3 wb;
    m3x3 cm;
    if (precalc_cm != NULL) {
        cm = *precalc_cm;
    } else {
        cm = calculate_color_matrix(calib_ls, calib_wp, ps, use_slow_matrix_optimizer, report_dir);
    }
    wb = v3_norm(m3x3_mul_v3(cm, calib_wp));
    fprintf(stderr, "ColorMatrix optimal white balance for target: %g,%g,%g\n", wb.v[0], wb.v[1], wb.v[2]);
    m3x3 fm;
    if (precalc_fm != NULL) {
        fm = *precalc_fm;
    } else {
        m3x3 precalc_cam2xyz_;
        m3x3 *precalc_cam2xyz = NULL;
        if (memcmp(&calib_wp, &glob.d50, sizeof(calib_wp)) == 0 && precalc_cm == NULL) {
            precalc_cam2xyz = &precalc_cam2xyz_;
            *precalc_cam2xyz = invert_m3x3(cm);
        }
        fm = calculate_forward_matrix(precalc_cam2xyz, psD50, use_slow_matrix_optimizer, &wb, report_dir);
    }
    fprintf(stderr, "ForwardMatrix optimal white balance for target: %g,%g,%g\n", wb.v[0], wb.v[1], wb.v[2]);
    const m3x3 cam2xyzD50 = m3x3_mul_m3x3(fm, invert_m3x3(v3_asdiagonal(wb)));

    chromalut_t *lut = chromalut_new(cam2xyzD50, psD50, min_chroma_dist, regularization);

    {
        const v3 d50 = glob.d50;
        struct patch_set *ps1 = normalize_patch_set(psD50);
        print_pm_errors(cam2xyzD50, lut, NULL, ps1, d50, true, report_dir, "lut-patch-errors.txt");

        print_lookup_reports(report_dir, lut, NULL, dcam_ssf, cam2xyzD50, psD50, &d50);
        target_print("targetd50", ps1, d50, obs, report_dir);
        target_delete_patch_set(ps1);
    }

    struct dcam_profile *prof = calloc(1, sizeof(*prof));
    prof->illuminant_name = calib_ls;
    prof->illuminant_spec = spec_copy(calib_spec);
    prof->color_matrix = cm;
    prof->forward_matrix = fm;
    prof->lut = lut;

    return prof;
}

static double
find_cct(const struct observer *obs,
         double u,
         double v,
         double *diff_de)
{
    double min_de = 100000;
    double best_temp = 0;
    // FIXME: make binary search
#pragma omp parallel for
    for (int i = 500; i < 100000; i += 2) {
        double temp = (double)i;
        spectrum_t *bb = spectraldb_render_blackbody(temp, 300, 830, 5);
        v3 uvY = xyz2uvY(spec2xyz(bb, obs));
        free(bb);
        double de = sqrt((u-uvY.v[0])*(u-uvY.v[0]) + (v-uvY.v[1])*(v-uvY.v[1]));
#pragma omp critical
        if (de < min_de) {
            min_de = de;
            best_temp = temp;
        }
    }
    if (diff_de != NULL) {
        spectrum_t *bb = spectraldb_render_blackbody(best_temp, 300, 830, 5);
        v3 xyz1 = v3_norm(spec2xyz(bb, obs));
        v3 xyz2 = v3_norm(uv2xyz_lightest(u, v, (v3){{1,1,1}}));
        *diff_de = ciede2000(xyz1, xyz2, xyz1);
        free(bb);
    }
    return best_temp;
}

static void
test_profile(const struct profio_dcp *dcp,
             const struct dcam_profile *prof,
             const v3 *custom_wb,
             bool skip_lut,
             const v3 test_wp,
             const spectrum_t *test_spec,
             const struct patch_set *ps_,
             const struct patch_set *psD50_,
             const struct observer *obs,
             const struct dcam_ssf *dcam_ssf,
             const char report_dir[])
{
    const v3 d50 = glob.d50;
    uint32_t hsmdims[3];
    chromalut_t *lut = NULL;
    struct dcp_lut dcp_lut;
    memset(&dcp_lut, 0, sizeof(dcp_lut));
    m3x3 cam2xyzD50;
    v3 wb;
    if (custom_wb != NULL) {
        wb = v3_norm(*custom_wb);
        fprintf(stderr, "Using user-provided white balance for target: %g,%g,%g\n", wb.v[0], wb.v[1], wb.v[2]);
    } else if (test_spec != NULL && dcam_ssf != NULL) {
        wb = v3_norm(spec2xyz(test_spec, &dcam_ssf->ssf));
        fprintf(stderr, "Calculated white balance from illuminant spectra and camera SSFs: %g,%g,%g\n", wb.v[0], wb.v[1], wb.v[2]);
    } else {
        wb.v[0] = 1;
        wb.v[1] = 1;
        wb.v[2] = 1;
        int w_idx = find_whitest_patch(psD50_, d50);
        if (w_idx == -1) {
            fprintf(stderr, "Didn't find a white patch in the target, could not figure out white balance.\n");
            exit(EXIT_FAILURE);
        }
        fprintf(stderr, "Whitest patch in target has index %d.\n", w_idx);
        wb = v3_norm(psD50_->patch[w_idx].cam);
        v3 cam;
        v3 xyzD50 = v3_norm(psD50_->patch[w_idx].xyz);
        if (prof != NULL) {
            m3x3 xyz2cam = invert_m3x3(prof->forward_matrix);
            cam = m3x3_mul_v3(xyz2cam, xyzD50);
            cam = v3_norm(cam);
            wb.v[0] = psD50_->patch[w_idx].cam.v[0] / cam.v[0];
            wb.v[1] = psD50_->patch[w_idx].cam.v[1] / cam.v[1];
            wb.v[2] = psD50_->patch[w_idx].cam.v[2] / cam.v[2];
            wb = v3_norm(wb);
        } else if (dcp != NULL) {
            v3 old_wb = wb;
            for (int i = 0; i < 1000; i++) {
                m3x3 cam2xyzD50 = dcp_make_cam2xyz_d50(dcp, wb);
                v3 invwb = {{ 1.0/wb.v[0], 1.0/wb.v[1], 1.0/wb.v[2] }};
                m3x3 fm = m3x3_mul_m3x3(cam2xyzD50, invert_m3x3(v3_asdiagonal(invwb)));
                //jsonio_3x3matrix_print(stderr, "", "", fm);
                m3x3 xyz2cam = invert_m3x3(fm);
                cam = m3x3_mul_v3(xyz2cam, xyzD50);
                cam = v3_norm(cam);
                wb.v[0] = psD50_->patch[w_idx].cam.v[0] / cam.v[0];
                wb.v[1] = psD50_->patch[w_idx].cam.v[1] / cam.v[1];
                wb.v[2] = psD50_->patch[w_idx].cam.v[2] / cam.v[2];
                wb = v3_norm(wb);
                //fprintf(stderr, "%f\n", simple_de(old_wb, wb));
                if (simple_de(old_wb, wb) < 1e-7) {
                    break;
                }
                old_wb = wb;
            }
        }
        fprintf(stderr, "Calculated optimal white balance for target: %g,%g,%g\n", wb.v[0], wb.v[1], wb.v[2]);
    }
    if (prof != NULL) {
        cam2xyzD50 = m3x3_mul_m3x3(prof->forward_matrix, invert_m3x3(v3_asdiagonal(wb)));
        if (!skip_lut) {
            lut = prof->lut;
        }
    } else if (dcp != NULL) {
        cam2xyzD50 = dcp_make_cam2xyz_d50(dcp, wb);
        if (!skip_lut) {
            dcp_lut.hsm = dcp_make_lut(dcp, wb);
            if (dcp_lut.hsm != NULL) {
                memcpy(hsmdims, dcp->hsmdims, sizeof(hsmdims));
                dcp_lut.hsmdims = hsmdims;
                dcp_lut.srgb_gamma = dcp->huesatmap_srgb_gamma;
                if (dcp->looktable != NULL) {
                    fprintf(stderr, "The DCP has both HueSatMap and LookTable, skipping LookTable.\n");
                }
            } else if (dcp->looktable != NULL) {
                fprintf(stderr, "The DCP has no HueSatMap but has a LookTable, applying that instead.\n");
                size_t count = dcp->lookdims[0]*dcp->lookdims[1]*dcp->lookdims[2];
                dcp_lut.hsm = malloc(count * sizeof(*dcp_lut.hsm));
                memcpy(dcp_lut.hsm, dcp->looktable, count * sizeof(*dcp_lut.hsm));
                memcpy(hsmdims, dcp->lookdims, sizeof(hsmdims));
                dcp_lut.hsmdims = hsmdims;
                dcp_lut.srgb_gamma = dcp->looktable_srgb_gamma;
            }
        }
        if (dcp->tonecurve != NULL) {
            fprintf(stderr, "The DCP has a tone curve (ignored).\n");
        }
        if (dcp->has.baseline_exposure_offset) {
            fprintf(stderr, "The DCP has a baseline exposure offset (ignored).\n");
        }
    }

    { // test white point
        v3 wb_wp;
        if (prof != NULL) {
            wb_wp = m3x3_mul_v3(invert_m3x3(prof->color_matrix), wb);
        } else if (dcp != NULL) {
            double xy[2];
            dcp_neutral2xy(dcp, wb, xy);
            wb_wp = xy2xyz_lightest(xy[0], xy[1], (v3){{1,1,1}});
        }
        v3 xyY = xyz2xyY(wb_wp);
        v3 uvY = xyz2uvY(wb_wp);
        v3 invwb;
        invwb.v[0] = 1.0 / wb.v[0];
        invwb.v[1] = 1.0 / wb.v[1];
        invwb.v[2] = 1.0 / wb.v[2];
        invwb = v3_norm(invwb);
        double cct_de;
        double cct = find_cct(obs, uvY.v[0], uvY.v[1], &cct_de);
        double test_cct_de;
        v3 test_uvY = xyz2uvY(v3_norm(test_wp));
        double test_cct = find_cct(obs, test_uvY.v[0], test_uvY.v[1], &test_cct_de);
        fprintf(stderr,
                "White balance whitepoint according to profile:\n"
                "  raw RGB: %.4f, %.4f, %.4f\n"
                "      XYZ: %.4f, %.4f, %.4f\n"
                "       xy: %.4f %.4f\n"
                "     u'v': %.4f %.4f\n"
                "      CCT: %uK (%.2f DE off-axis)\n"
                " test CCT: %uK (%.2f DE off-axis)\n",
                invwb.v[0], invwb.v[1], invwb.v[2],
                wb_wp.v[0], wb_wp.v[1], wb_wp.v[2], xyY.v[0], xyY.v[1], uvY.v[0], uvY.v[1], (int)cct, cct_de, (int)test_cct, test_cct_de);
    }
    /*
    fprintf(stderr, "The derived white-balanced camera raw RGB to XYZ D50 matrix:\n");
    fprintf(stderr, "  {\n");
    jsonio_3x3matrix_print(stderr, "    ", "cam2xyzD50", cam2xyzD50);
    fprintf(stderr, "\n  }\n");
    */

    struct patch_set *ps = normalize_patch_set(ps_);
    struct patch_set *psD50 = normalize_patch_set(psD50_);

    target_print("target", ps, test_wp, obs, report_dir);
    target_print("targetd50", psD50, d50, obs, report_dir);

    print_lookup_reports(report_dir, lut, dcp ? &dcp_lut : NULL, dcam_ssf, cam2xyzD50, psD50, &d50);

    struct patch_error_report *pe = test_patch_matching(cam2xyzD50, lut, dcp ? &dcp_lut : NULL, psD50, d50);
    print_patch_matching_errors(pe, true, report_dir, "fm-patch-errors.txt", stderr);

    target_delete_patch_set(ps);
    target_delete_patch_set(psD50);
    free(pe);
    free(dcp_lut.hsm);
}

static void
regenerate_patches(const char filename[],
                   struct patch_set *ps,
                   bool regenerate_xyz,
                   const struct dcam_ssf *dcam_ssf,
                   const struct observer *observer,
                   const spectrum_t *xyz_illuminant,
                   enum exif_lightsource xyz_illuminant_ls,
                   const spectrum_t *cam_illuminant,
                   enum exif_lightsource cam_illuminant_ls)
{
    if (regenerate_xyz && dcam_ssf != NULL) {
        if (xyz_illuminant_ls != cam_illuminant_ls) {
            fprintf(stderr,
                    //2345678901234567890123456789012345678901234567890123456789012345678901234567890
                    "Re-generating target reference XYZ for illuminant %s and camera raw \n"
                    "  RGB values for illuminant %s...\n",
                    exif_lightsource_ntoa(xyz_illuminant_ls), exif_lightsource_ntoa(cam_illuminant_ls));
        } else {
            fprintf(stderr,
                    //2345678901234567890123456789012345678901234567890123456789012345678901234567890
                    "Re-generating target reference XYZ and camera raw RGB values for\n"
                    "  illuminant %s...\n", exif_lightsource_ntoa(xyz_illuminant_ls));
        }
    } else if (regenerate_xyz) {
        fprintf(stderr, "Re-generating target reference XYZ values for illuminant %s...\n", exif_lightsource_ntoa(xyz_illuminant_ls));
    } else if (dcam_ssf) {
        fprintf(stderr, "Re-generating target camera raw RGB values for illuminant %s...\n", exif_lightsource_ntoa(cam_illuminant_ls));
    } else {
        // do nothing
        return;
    }

    for (int i = 0; i < ps->patch_count; i++) {
        if (ps->patch[i].spectrum == NULL) {
            fprintf(stderr, "target \"%s\" lacks spetral data, cannot regenerate XYZ or RGB values.\n", filename);
            exit(EXIT_FAILURE);
        }
        if (dcam_ssf != NULL) {
            ps->patch[i].cam = spec2xyz_ill(ps->patch[i].spectrum, &dcam_ssf->ssf, cam_illuminant, 1);
        }
        if (regenerate_xyz) {
            ps->patch[i].xyz = spec2xyz_ill(ps->patch[i].spectrum, observer, xyz_illuminant, 1);
        }
    }
}

static v3
parse_arg_illuminant(const char arg[],
                     const struct observer *observer,
                     bool *used_observer,
                     spectrum_t **ill_spec)
{
    if (ill_spec != NULL) {
        *ill_spec = NULL;
    }
    v3 p_ill = {{ 0,0,0 }};
    if (strchr(arg, ',') != NULL) {
        // x,y,z syntax
        p_ill.v[0] = atof(arg);
        char *p = strchr(arg, ',')+1;
        p_ill.v[1] = atof(p);
        p = strchr(p, ',');
        if (p != NULL) {
            p_ill.v[2] = atof(p+1);
        } else {
            v3 xyY = {{p_ill.v[0], p_ill.v[1], 1.0}};
            p_ill = xyY2xyz(xyY);
        }
    } else {
        // spectrum
        spectrum_t *s = jsonio_illuminant_parse(arg, false);
        if (s != NULL) {
            p_ill = spec2xyz(s, observer);
            *used_observer = true;
            if (ill_spec != NULL) {
                *ill_spec = s;
            } else {
                free(s);
            }
        } else if ((int)exif_lightsource_aton(arg) >= 0) {
            fprintf(stderr,
                    "Illuminant \"%s\" is a known EXIF illuminant but there is no spectrum\n"
                    "  for it in the built-in database, so no XYZ coordinates can be found.\n"
                    "  Please choose a different illuminant, or provide illuminant spectrum\n"
                    "  as a file.\n", arg);
            exit(EXIT_FAILURE);
        }
    }
    p_ill = v3_norm(p_ill);
    if (p_ill.v[0] <= 0.0 || p_ill.v[1] <= 0.0 || p_ill.v[2] <= 0.0) {
        fprintf(stderr, "Invalid illuminant \"%s\".\n", arg);
        exit(EXIT_FAILURE);
    }
    return p_ill;
}

static struct patch_set *
regenerate_target_values(struct patch_set **ps_,
                         const char filename[],
                         enum exif_lightsource calib_ls,
                         const spectrum_t *calib_spec,
                         const v3 calib_wp,
                         const v3 xyz_ref_wp,
                         const struct observer *observer,
                         const struct dcam_ssf *dcam_ssf)
{
    struct patch_set *ps = *ps_;
    struct patch_set *psD50 = target_copy(ps);

    if (ps->patch[0].spectrum != NULL) {
        if (calib_spec == NULL) {
            if (dcam_ssf != NULL) {
                fprintf(stderr,
                        //2345678901234567890123456789012345678901234567890123456789012345678901234567890
                        "Spectrum for calibration illuminant is not known, cannot regenerate camera raw\n"
                        "  RGB values. Either change calibration illuminant or do not provide\n"
                        "  camera SSFs (that is use the RGB values in the target file as is).\n");
                exit(EXIT_FAILURE);
            } else if (memcmp(&calib_wp, &xyz_ref_wp, sizeof(calib_wp)) != 0) {
                fprintf(stderr,
                        //2345678901234567890123456789012345678901234567890123456789012345678901234567890
                        "Warning: the target XYZ reference values' illuminant is different from the\n"
                        "  calibration illuminant, and while the target has spectra to allow\n"
                        "  re-generation the spectrum for the calibration illuminant is not known.\n"
                        "  Using a chromatic adaptation algorithm instead (CAT02) to transform XYZ\n"
                        "  reference values to match the calibration illuminant, which is not as precise.\n");
                for (int i = 0; i < ps->patch_count; i++) {
                    ps->patch[i].xyz = chromatic_adaptation_transform(CAT_CAT02, ps->patch[i].xyz, xyz_ref_wp, calib_wp);
                }
            }
        } else {
            regenerate_patches(filename, ps, true, dcam_ssf, observer, calib_spec, calib_ls, calib_spec, calib_ls);
        }
    } else if (memcmp(&calib_wp, &xyz_ref_wp, sizeof(calib_wp)) != 0) {
        fprintf(stderr,
                //2345678901234567890123456789012345678901234567890123456789012345678901234567890
                "Warning: the target XYZ reference values illuminant is different from the\n"
                "  calibration illuminant and the target lacks spectra so XYZ values cannot be\n"
                "  regenerated from scratch. Using a chromatic adaptation algorithm instead\n"
                "  (CAT02) to transform XYZ reference values to match the calibration illuminant.\n"
                "  This is not as precise as re-generation from spectral information.\n");
        for (int i = 0; i < ps->patch_count; i++) {
            ps->patch[i].xyz = chromatic_adaptation_transform(CAT_CAT02, ps->patch[i].xyz, xyz_ref_wp, calib_wp);
        }
    }

    ps = exclude_weak_signal_patches(ps, true);

    if (calib_ls == lsD50) {
        target_delete_patch_set(psD50);
        psD50 = target_copy(ps);
    } else if (psD50->patch[0].spectrum != NULL) {
        const spectrum_t *d50 = spectraldb_illuminant(lsD50);
        regenerate_patches(filename, psD50, true, dcam_ssf, observer, d50, lsD50, calib_spec, calib_ls);
        psD50 = exclude_weak_signal_patches(psD50, false);
    } else if (ciede2000(glob.d50, xyz_ref_wp, glob.d50) > 0.1) {
        fprintf(stderr,
                //2345678901234567890123456789012345678901234567890123456789012345678901234567890
                "Warning: the target XYZ reference values illuminant is different from D50 and\n"
                "  the target lacks spectra so XYZ values cannot be regenerated from scratch.\n"
                "  Using a chromatic adaptation transform instead (CAT02) to make values match.\n"
                "  This is not as precise as re-generation from spectral information.\n");
        for (int i = 0; i < psD50->patch_count; i++) {
            psD50->patch[i].xyz = chromatic_adaptation_transform(CAT_CAT02, psD50->patch[i].xyz, xyz_ref_wp, glob.d50);
        }
        psD50 = exclude_weak_signal_patches(psD50, false);
    }
    *ps_ = ps;
    return psD50;
}

static void
print_usage_and_exit(void)
{
    const char *usage_fmt =
        "DCamProf v" DCAMPROF_VERSION " - a digital camera profiling tool.\n"
        "\n"
        "Usage: dcamprof <command> [command-specific flags] <command args>\n"
        "\n"
        "Commands:\n"
        "\n"
        "  make-target [flags] <output.ti3>\n"
        "    flags:\n"
        "      -c <ssf.json> camera's spectral sensitivity functions\n"
        "      -o <observer> (default: 1931_2)\n"
        "      -i <illuminant> (default: D50)\n"
        "      -p <patches.ti3> include patch set\n"
        "      -a <name> assign (new) class name to previously included patch set (-p)\n"
        "      -g <generated grid spacing> (default: 0.03)\n"
        "      -d <distance> mimimum distance between patches of different classes (default: 0.02)\n"
        "      -b <distance> exclude patch if there is a lighter patch with same chromaticity\n"
        "         suggested chromaticity distance 0.004 (default: not active)\n"
        "      -X don't regenerate XYZ values of imported patch sets\n"
        "      -R don't regenerate RGB values of imported patch sets\n"
        "      -n exclude spectra in output (default: include if all inputs has it)\n"
        "      -r <dir> directory to save informational reports and plots\n"
        "    built-in patch sets:\n"
        "      from measured data: %s\n"
        "      generated borders: %s\n"
        "      generated grids: <border>-grid\n"
        "\n"
        "  make-profile [flags] <target.ti3> <output.json>\n"
        "    flags:\n"
        "      -w <patch class name | all> <max DE>[,matrix opt weight][,kL,kC,kH]\n"
        "         LUT/matrix optimizer parameters\n"
        "      -W skip patch normalization weighting\n"
        "      -M ignore DE weights for matrix optimization\n"
        "      -l <l,c> LUT luminance and chromaticity relax parameters (default: 0)\n"
        "      -d <distance> minimum patch chromaticity distance in LUT (default: 0.02)\n"
        "      -o <observer> (default: 1931_2)\n"
        "      -c <ssf.json> camera SSFs, used to re-generate target RGB values\n"
        "      -i <calibration illuminant> (default: D50)\n"
        "      -I <target XYZ reference values illuminant> (default: same as calibration illuminant)\n"
        "      -m <matrix.json | dcp.json> provide pre-calculated color matrix\n"
        "      -f <matrix.json | dcp.json> provide pre-calculated forward matrix\n"
        "      -s use slow but possibly better matrix optimization\n"
        "      -r <dir> directory to save informational reports and plots\n"
        "\n"
        "  test-profile [flags] <target.ti3> <profile.json | profile.dcp>\n"
        "    flags:\n"
        "      -o <observer> (default: 1931_2)\n"
        "      -c <ssf.json> camera SSFs, used to re-generate target RGB values\n"
        "      -i <test illuminant> (default: same as profile's calibration illuminant)\n"
        "      -I <target XYZ reference values illuminant> (default: same as test illuminant)\n"
        "      -w <r,g,b> | m<r,g,b> provide camera WB (default: derived from target)\n"
        "      -L skip LUT\n"
        "      -r <dir> directory to save informational reports and plots\n"
        "\n"
        "  make-dcp [flags] <profile.json> <output.dcp>\n"
        "    flags:\n"
        "      -c <unique camera name>\n"
        "      -h <hdiv,sdiv> hue and saturation divisions of DCP HSM (default: 90,30)\n"
        "      -F skip forward matrix\n"
        "      -L skip LUT (= matrix-only profile)\n"
        "      -O disable forward matrix whitepoint remapping\n"
        "      -r <dir> directory to save informational reports and plots\n"
        "\n"
        "  dcp2json <input.dcp> [<output.json>]\n"
        "\n"
        "  json2dcp <input.json> <output.dcp>\n"
        "\n"
        "Illuminants:\n"
        "  <x,y> whitepoint coordinate in xyY color space\n"
        "  <X,Y,Z> whitepoint coordinate in XYZ color space\n"
        "  <exif lightsource> integer or string\n"
        "  <spectrum.json> (requires observer)\n"
        "  built-in illuminant spectra: %s, and <temp>K for blackbodies\n"
        "\n"
        "Observers:\n"
        "  <observer.json> (not implemented yet)\n"
        "  built-in observers: %s\n";
    fprintf(stderr, usage_fmt, target_measured_list(), target_generated_list(),
            spectraldb_illuminant_list(), observer_list());
    exit(EXIT_FAILURE);
}

static void
init_observer_dependent_globals(const struct observer *obs)
{
    glob.d50 = spec2xyz(spectraldb_illuminant(lsD50), obs);
    glob.d50 = v3_norm(glob.d50);
}

int
main(int argc,
     char *argv[])
{
    if (argc < 3) { // minimum three: dcamprof <command> <arg>
        print_usage_and_exit();
    }

    bool parse_flags = true;
    int ai = 2;
    if (strcmp(argv[1], "dcp2json") == 0) {

        /******************
          DCP to JSON
         */

        if (ai != argc - 2 && ai != argc - 1) {
            print_usage_and_exit();
        }
        struct profio_dcp *dcp = profio_dcp_parse(argv[ai+0], true);
        FILE *out = stdout;
        if (ai == argc - 2) {
            out = fopen(argv[ai+1], "wt");
            if (out == NULL) {
                fprintf(stderr, "Could not open \"%s\" for writing: %s.\n", argv[ai+1], strerror(errno));
                exit(EXIT_FAILURE);
            }
        }
        profio_dcp_json(dcp, out);
        if (out != stdout) {
            fclose(out);
        }
        profio_dcp_delete(dcp);

    } else if (strcmp(argv[1], "json2dcp") == 0) {

        /******************
          JSON to DCP
         */
        if (ai != argc - 2) print_usage_and_exit();
        struct profio_dcp *dcp = jsonio_dcp_parse(argv[ai+0], true);
        profio_dcp_write(dcp, argv[ai+1], true);
        profio_dcp_delete(dcp);

    } else if (strcmp(argv[1], "make-target") == 0) {

        /******************
          Make target
         */
        const char *ps_name[argc];
        const char *ps_class_name[argc];
        memset(ps_class_name, 0, sizeof(ps_class_name));
        int ps_count = 0;
        const char *report_dir = NULL;
        struct dcam_ssf *dcam_ssf = NULL;
        const struct observer *observer = observer_get(OBSERVER_1931_2);
        spectrum_t *illuminant = spec_copy(spectraldb_illuminant(lsD50));
        enum exif_lightsource illuminant_ls = lsD50;
        double min_chroma_dist = 0.02;
        double min_chroma_dist_lighter = -1;
        bool exclude_spectra = false;
        bool regenerate_xyz = true;
        bool regenerate_cam = true;
        double grid_spacing = 0.03;
        for (; ai < argc-1 && parse_flags && argv[ai][0] == '-'; ai++) {
            switch (argv[ai][1]) {
            case '\0': ai -= 2; parse_flags = false; break;
            case 'c': dcam_ssf = jsonio_dcam_ssf_parse(argv[++ai], true); break;
            case 'i':
                free(illuminant);
                illuminant = jsonio_illuminant_parse(argv[++ai], true);
                if ((int)exif_lightsource_aton(argv[ai]) >= 0) {
                    illuminant_ls = exif_lightsource_aton(argv[ai]);
                } else {
                    illuminant_ls = lsOther;
                }
                break;
            case 'o': observer = observer_get_byname(argv[++ai], true); break;
            case 'p': ps_name[ps_count++] = argv[++ai]; break;
            case 'a': {
                if (ps_count == 0) {
                    fprintf(stderr, "Cannot assign class name, no patch set included.\n");
                    exit(EXIT_FAILURE);
                }
                ps_class_name[ps_count-1] = argv[++ai];
                break;
            }
            case 'b': min_chroma_dist_lighter = atof(argv[++ai]); break;
            case 'd': min_chroma_dist = atof(argv[++ai]); break;
            case 'g': grid_spacing = atof(argv[++ai]); break;
            case 'X': regenerate_xyz = false; break;
            case 'R': regenerate_cam = false; break;
            case 'n': exclude_spectra = true; break;
            case 'r': report_dir = argv[++ai]; break;
            default:
                fprintf(stderr, "unknown flag \"%s\".\n", argv[ai]);
                print_usage_and_exit();
                break;
            }
        }
        init_observer_dependent_globals(observer);
        if (ai != argc - 1) print_usage_and_exit();
        if (ps_count == 0) {
            fprintf(stderr, "No patch sets included (-p), nothing to do.\n");
            exit(EXIT_FAILURE);
        }
        const char *load_ps[ps_count];
        const char *load_ps_names[ps_count];
        int load_ps_idx[ps_count];
        int load_ps_count = 0;
        const char *meas_ps[ps_count];
        const char *meas_ps_names[ps_count];
        int meas_ps_idx[ps_count];
        int meas_ps_count = 0;
        const char *gene_ps[ps_count];
        const char *gene_ps_names[ps_count];
        int gene_ps_idx[ps_count];
        int gene_ps_count = 0;
        const char *grid_ps[ps_count];
        const char *grid_ps_names[ps_count];
        int grid_ps_idx[ps_count];
        int grid_ps_count = 0;
        for (int i = 0; i < ps_count; i++) {
            bool is_measured;
            if (target_isvalid_target_name(ps_name[i], &is_measured)) {
                if (is_measured) {
                    meas_ps_names[meas_ps_count] = ps_class_name[i];
                    meas_ps_idx[meas_ps_count] = i;
                    meas_ps[meas_ps_count++] = ps_name[i];
                } else {
                    gene_ps_names[gene_ps_count] = ps_class_name[i];
                    gene_ps_idx[gene_ps_count] = i;
                    gene_ps[gene_ps_count++] = ps_name[i];
                }
            } else {
                bool found = false;
                const char *p = strstr(ps_name[i], "-grid");
                if (p == NULL) {
                    p = strstr(ps_name[i], "-GRID");
                }
                if (p != NULL && p[5] == '\0') {
                    char name[strlen(ps_name[i])];
                    strcpy(name, ps_name[i]);
                    name[strlen(name)-5] = '\0';
                    if (target_isvalid_target_name(name, &is_measured)) {
                        if (is_measured) {
                            fprintf(stderr, "Cannot generate grid on built-in measurement based patch set \"%s\".\n", ps_name[i]);
                            exit(EXIT_FAILURE);
                        }
                        found = true;
                        grid_ps_names[grid_ps_count] = ps_class_name[i];
                        grid_ps_idx[grid_ps_count] = i;
                        grid_ps[grid_ps_count++] = ps_name[i];
                    }
                }
                if (!found) {
                    load_ps_names[load_ps_count] = ps_class_name[i];
                    load_ps_idx[load_ps_count] = i;
                    load_ps[load_ps_count++] = ps_name[i];
                }
            }
        }
        print_common_spectra(report_dir, observer, illuminant, dcam_ssf);
        struct patch_set *inps[ps_count];
        for (int i = 0; i < load_ps_count; i++) {
            fprintf(stderr, "Loading patch set from \"%s\"...\n", load_ps[i]);
            struct patch_set *ps = argyllio_ti3_parse(load_ps[i], true);
            if (!exclude_spectra) {
                for (int j = 0; j < ps->patch_count; j++) {
                    if (ps->patch[j].spectrum == NULL) {
                        fprintf(stderr, "Patch set \"%s\" lacks spectral information; disabling spectra in output.\n", load_ps[i]);
                        exclude_spectra = true;
                        break;
                    }
                }
            }
            if (ps->class_names[0][0] == '\0' && load_ps_names[i] == NULL) {
                fprintf(stderr, "Patch set in \"%s\" has no class name(s) and can thus not be identified. Provide one with the (-a) parameter.\n", load_ps[i]);
                exit(EXIT_FAILURE);
            }
            if (load_ps_names[i] != NULL) {
                target_set_name(ps, load_ps_names[i]);
            }
            regenerate_patches(load_ps[i], ps, regenerate_xyz, regenerate_cam ? dcam_ssf : NULL, observer, illuminant, illuminant_ls, illuminant, illuminant_ls);
            inps[load_ps_idx[i]] = ps;
        }
        for (int i = 0; i < meas_ps_count; i++) {
            fprintf(stderr, "Loading built-in measurement-based patch set \"%s\"...\n", meas_ps[i]);
            inps[meas_ps_idx[i]] = target_make_virtual(meas_ps[i], dcam_ssf, observer, illuminant, 0.0, false, false, NULL, report_dir);
            if (meas_ps_names[i] != NULL) {
                target_set_name(inps[meas_ps_idx[i]], meas_ps_names[i]);
            }
        }
        for (int i = 0; i < gene_ps_count; i++) {
            fprintf(stderr, "Generating chromaticity border patch set \"%s\"...\n", gene_ps[i]);
            inps[gene_ps_idx[i]] = target_make_virtual(gene_ps[i], dcam_ssf, observer, illuminant, grid_spacing, false, false, NULL, report_dir);
            if (gene_ps_names[i] != NULL) {
                target_set_name(inps[gene_ps_idx[i]], gene_ps_names[i]);
            }
        }
        for (int i = 0; i < grid_ps_count; i++) {
            fprintf(stderr, "Generating chromaticity grid patch set \"%s\"...\n", grid_ps[i]);
            char name[strlen(grid_ps[i])+1];
            strcpy(name, grid_ps[i]);
            name[strlen(name)-5] = '\0';
            inps[grid_ps_idx[i]] = target_make_virtual(name, dcam_ssf, observer, illuminant, grid_spacing, true, false, NULL, report_dir);
            if (grid_ps_names[i] != NULL) {
                target_set_name(inps[grid_ps_idx[i]], grid_ps_names[i]);
            }
        }
        struct patch_set *ps;
        if (ps_count > 1) {
            int tot_count = 0;
            for (int i = 0; i < ps_count; i++) {
                tot_count += inps[i]->patch_count;
            }
            fprintf(stderr, "Merging %d patch sets with %d patches in total...\n", ps_count, tot_count);
            ps = target_merge(inps, ps_count, min_chroma_dist, true);
            if (ps->patch_count != tot_count) {
                int class_count = 0;
                bool has_class[TARGET_MAX_CLASS_COUNT];
                memset(has_class, 0, sizeof(has_class));
                for (int i = 0; i < ps->patch_count; i++) {
                    if (!has_class[ps->patch[i].class]) {
                        has_class[ps->patch[i].class] = true;
                        class_count++;
                    }
                }
                fprintf(stderr, "%zd patches were excluded due to nearby neighbors of different class.\n%zd patches in %d class%s left in merged set.\n",
                        tot_count - ps->patch_count, ps->patch_count, class_count, class_count == 1 ? "" : "es");
            }
        } else {
            ps = inps[0];
        }
        if (min_chroma_dist_lighter >= 0) {
            ps = exclude_darker_patches(ps, min_chroma_dist_lighter);
        }

        fprintf(stderr, "Writing output to \"%s\"...\n", argv[ai]);
        argyllio_ti3_write(argv[ai], !exclude_spectra, ps, true);
        const v3 wp = v3_norm(spec2xyz(illuminant, observer));
        target_print("target", ps, wp, observer, report_dir);
        target_delete_patch_set(ps);
        free(illuminant);
        fprintf(stderr, "Complete!\n");

    } else if (strcmp(argv[1], "make-profile") == 0) {

        /******************
          Make profile
         */
        const char *report_dir = NULL;
        enum exif_lightsource calib_ls = lsD50;
        spectrum_t *calib_spec = spec_copy(spectraldb_illuminant(lsD50));
        const struct observer *observer = observer_get(OBSERVER_1931_2);
        bool used_observer = false;
        bool use_slow_matrix_optimizer = false;
        v3 calib_wp = {{0,0,0}};
        v3 xyz_ref_wp = {{0,0,0}};
        double min_chroma_dist = 0.02;
        bool skip_base_weighting = false;
        bool skip_matrix_de = false;
        struct dcam_ssf *dcam_ssf = NULL;
        m3x3 *precalc_fm = NULL;
        m3x3 *precalc_cm = NULL;
        double regularization[3] = { 0, 0, 0 };
        int class_count = 0;
        char *class_name[TARGET_MAX_CLASS_COUNT];
        double class_weights[TARGET_MAX_CLASS_COUNT][5];
        memset(class_name, 0, sizeof(class_name));
        for (int j = 0; j < TARGET_MAX_CLASS_COUNT; j++) {
            class_weights[j][0] = 0.0;
            class_weights[j][1] = 1.0;
            class_weights[j][2] = 1.0;
            class_weights[j][3] = 1.0;
            class_weights[j][4] = 1.0;
        }
        for (; ai < argc-1 && parse_flags && argv[ai][0] == '-'; ai++) {
            switch (argv[ai][1]) {
            case '\0': ai -= 2; parse_flags = false; break;
            case 'W': skip_base_weighting = true; break;
            case 'M': skip_matrix_de = true; break;
            case 'w':
                if (class_count == TARGET_MAX_CLASS_COUNT) {
                    fprintf(stderr, "Too many patch classes.\n");
                    exit(EXIT_FAILURE);
                }
                class_name[class_count] = argv[++ai];
                char *p = argv[++ai];
                class_weights[class_count][0] = atof(p);
                p = strchr(p, ','); if (p) class_weights[class_count][1] = atof(++p);
                if (p) p = strchr(p, ','); if (p) class_weights[class_count][2] = atof(++p);
                if (p) p = strchr(p, ','); if (p) class_weights[class_count][3] = atof(++p);
                if (p) p = strchr(p, ','); if (p) class_weights[class_count][4] = atof(++p);
                class_count++;
                break;
            case 'd': min_chroma_dist = atof(argv[++ai]); break;
            case 'c': dcam_ssf = jsonio_dcam_ssf_parse(argv[++ai], true); break;
            case 'l': {
                char *p = argv[++ai];
                regularization[0] = atof(p);
                if ((p = strchr(p, ',')) != NULL) {
                    regularization[1] = regularization[2] = atof(++p);
                }
                break;
            }
            case 'i':
                free(calib_spec);
                calib_spec = NULL;
                calib_wp = parse_arg_illuminant(argv[++ai], observer, &used_observer, &calib_spec);
                if ((int)exif_lightsource_aton(argv[ai]) >= 0) {
                    calib_ls = exif_lightsource_aton(argv[ai]);
                } else {
                    calib_ls = lsOther;
                }
                break;
            case 'I': xyz_ref_wp = parse_arg_illuminant(argv[++ai], observer, &used_observer, NULL); break;
            case 'o':
                observer = observer_get_byname(argv[++ai], true);
                if (used_observer) {
                    fprintf(stderr, "Observer (-o) must be given before any illuminants on the command line.\n");
                    exit(EXIT_FAILURE);
                }
                break;
            case 'm': free(precalc_cm); precalc_cm = jsonio_3x3matrix_parse(argv[++ai], "ColorMatrix1", true); break;
            case 'f': free(precalc_fm); precalc_fm = jsonio_3x3matrix_parse(argv[++ai], "ForwardMatrix1", true); break;
            case 's': use_slow_matrix_optimizer = true; break;
            case 'r': report_dir = argv[++ai]; break;
            default:
                fprintf(stderr, "unknown flag \"%s\".\n", argv[ai]);
                print_usage_and_exit();
                break;
            }
        }
        init_observer_dependent_globals(observer);
        if (calib_wp.v[0] == 0 && calib_wp.v[1] == 0 && calib_wp.v[2] == 0) {
            calib_wp = glob.d50;
        }
        if (xyz_ref_wp.v[0] == 0 && xyz_ref_wp.v[1] == 0 && xyz_ref_wp.v[2] == 0) {
            xyz_ref_wp = calib_wp;
        }
        if (ai != argc - 2) print_usage_and_exit();
        if (calib_ls != lsD50) {
            fprintf(stderr, "Reading target and generating values for the calibration illuminant %s and for D50...\n", exif_lightsource_ntoa(calib_ls));
        } else {
            fprintf(stderr, "Reading target and generating values for the calibration illuminant %s...\n", exif_lightsource_ntoa(calib_ls));
        }
        struct patch_set *ps = argyllio_ti3_parse(argv[ai], true);
        double class_de[TARGET_MAX_CLASS_COUNT];
        double class_wt[TARGET_MAX_CLASS_COUNT];
        double class_klch[TARGET_MAX_CLASS_COUNT][3];
        for (int j = 0; j < TARGET_MAX_CLASS_COUNT; j++) {
            class_de[j] = 0.0;
            class_wt[j] = 1.0;
            class_klch[j][0] = 1.0;
            class_klch[j][1] = 1.0;
            class_klch[j][2] = 1.0;
        }
        if (class_count > 0) {
            if (class_count == 1 && strcasecmp(class_name[0], "all") == 0) {
                for (int i = 0; i < TARGET_MAX_CLASS_COUNT; i++) {
                    class_de[i] = class_weights[0][0];
                    class_wt[i] = class_weights[0][1];
                    class_klch[i][0] = class_weights[0][2];
                    class_klch[i][1] = class_weights[0][3];
                    class_klch[i][2] = class_weights[0][4];
                }
            } else {
                for (int i = 0; i < class_count; i++) {
                    int j;
                    for (j = 0; j < TARGET_MAX_CLASS_COUNT; j++) {
                        if (strcasecmp(ps->class_names[j], class_name[i]) == 0) {
                            class_de[j] = class_weights[i][0];
                            class_wt[j] = class_weights[i][1];
                            class_klch[j][0] = class_weights[i][2];
                            class_klch[j][1] = class_weights[i][3];
                            class_klch[j][2] = class_weights[i][4];
                            break;
                        }
                    }
                    if (j == TARGET_MAX_CLASS_COUNT) {
                        fprintf(stderr, "Target \"%s\" doesn't include class \"%s\", cannot assign weight.\n", argv[ai], class_name[i]);
                        exit(EXIT_FAILURE);
                    }
                }
            }
        }
        struct patch_set *psD50 = regenerate_target_values(&ps, argv[ai], calib_ls, calib_spec, calib_wp, xyz_ref_wp, observer, dcam_ssf);

        print_common_spectra(report_dir, observer, calib_spec, dcam_ssf);
        assign_optimization_weights(ps, class_wt, class_de, (const double(*)[3])class_klch, skip_base_weighting, skip_matrix_de, xyz_ref_wp);
        assign_optimization_weights(psD50, class_wt, class_de, (const double(*)[3])class_klch, skip_base_weighting, skip_matrix_de, glob.d50);

        fprintf(stderr, "Making camera profile...\n");
        struct dcam_profile *prof = make_profile(calib_ls, calib_wp, calib_spec, precalc_cm, precalc_fm, use_slow_matrix_optimizer, min_chroma_dist, regularization, ps, psD50, observer, dcam_ssf, report_dir);

        target_delete_patch_set(ps);
        fprintf(stderr, "Writing output to \"%s\"...\n", argv[ai+1]);
        //profio_dcp_write(dcp, argv[ai+1], true);
        write_dcam_profile(argv[ai+1], prof, true);
        dcam_profile_delete(prof);
        free(calib_spec);
        fprintf(stderr, "Complete!\n");

    } else if (strcmp(argv[1], "test-profile") == 0) {

        /******************
          Test profile
         */
        const struct observer *observer = observer_get(OBSERVER_1931_2);
        bool used_observer = false;
        bool skip_lut = false;
        v3 custom_wb_;
        v3 *custom_wb = NULL;
        struct dcam_ssf *dcam_ssf = NULL;
        v3 xyz_ref_wp = {{0,0,0}};
        v3 test_ill_wp = {{0,0,0}};
        enum exif_lightsource test_ill_ls = lsOther;
        spectrum_t *test_ill_spec = NULL;
        const char *report_dir = NULL;
        for (; ai < argc-1 && parse_flags && argv[ai][0] == '-'; ai++) {
            switch (argv[ai][1]) {
            case '\0': ai -= 2; parse_flags = false; break;
            case 'o':
                observer = observer_get_byname(argv[++ai], true);
                if (used_observer) {
                    fprintf(stderr, "Observer (-o) must be given before any illuminant on the command line.\n");
                    exit(EXIT_FAILURE);
                }
                break;
            case 'w': {
                char *p = argv[++ai];
                bool invert = false;
                if (*p == 'm') {
                    invert = true;
                    p++;
                }
                custom_wb_.v[0] = custom_wb_.v[1] = custom_wb_.v[2] = 0;
                custom_wb_.v[0] = atof(p);
                if ((p = strchr(p, ',')) != NULL) {
                    custom_wb_.v[1] = atof(++p);
                    if ((p = strchr(p, ',')) != NULL) {
                        custom_wb_.v[2] = atof(++p);
                    }
                }
                if (custom_wb_.v[0] <= 0 || custom_wb_.v[1] <= 0 || custom_wb_.v[2] <= 0) {
                    fprintf(stderr, "Bad white balance value.\n");
                    exit(EXIT_FAILURE);
                }
                if (invert) {
                    custom_wb_.v[0] = 1.0 / custom_wb_.v[0];
                    custom_wb_.v[1] = 1.0 / custom_wb_.v[1];
                    custom_wb_.v[2] = 1.0 / custom_wb_.v[2];
                }
                custom_wb_ = v3_norm(custom_wb_);
                custom_wb = &custom_wb_;
                break;
            }
            case 'L': skip_lut = true; break;
            case 'c': dcam_ssf = jsonio_dcam_ssf_parse(argv[++ai], true); break;
            case 'i':
                test_ill_wp = parse_arg_illuminant(argv[++ai], observer, &used_observer, &test_ill_spec);
                if ((int)exif_lightsource_aton(argv[ai]) >= 0) {
                    test_ill_ls = exif_lightsource_aton(argv[ai]);
                } else {
                    test_ill_ls = lsOther;
                }
                break;
            case 'I': xyz_ref_wp = parse_arg_illuminant(argv[++ai], observer, &used_observer, NULL); break;
            case 'r': report_dir = argv[++ai]; break;
            default:
                fprintf(stderr, "unknown flag \"%s\".\n", argv[ai]);
                print_usage_and_exit();
                break;
            }
        }
        init_observer_dependent_globals(observer);
        if (ai != argc - 2) print_usage_and_exit();

        struct dcam_profile *prof = NULL;
        struct profio_dcp *dcp = NULL;
        const char *prof_filename = argv[ai+1];
        enum exif_lightsource calib_ls;
        fprintf(stderr, "Reading profile...\n");
        if (strstr(prof_filename, ".dcp") != NULL || strstr(prof_filename, ".DCP") != NULL) {
            dcp = profio_dcp_parse(prof_filename, true);
            calib_ls = dcp->ill[0];
        } else {
            prof = jsonio_profile_parse(prof_filename, true);
            calib_ls = prof->illuminant_name;
        }

        if (test_ill_wp.v[0] == 0 && test_ill_wp.v[1] == 0 && test_ill_wp.v[2] == 0) {
            test_ill_ls = calib_ls;
            if (prof != NULL && prof->illuminant_spec != NULL) {
                test_ill_spec = spec_copy(prof->illuminant_spec);
                test_ill_wp = v3_norm(spec2xyz(test_ill_spec, observer));
            } else {
                test_ill_wp = parse_arg_illuminant(exif_lightsource_ntoa(test_ill_ls), observer, &used_observer, &test_ill_spec);
            }
        }
        if (xyz_ref_wp.v[0] == 0 && xyz_ref_wp.v[1] == 0 && xyz_ref_wp.v[2] == 0) {
            xyz_ref_wp = test_ill_wp;
        }
        print_common_spectra(report_dir, observer, test_ill_spec, dcam_ssf);
        fprintf(stderr, "Reading target...\n");
        struct patch_set *ps = argyllio_ti3_parse(argv[ai], true);
        struct patch_set *psD50 = regenerate_target_values(&ps, argv[ai], test_ill_ls, test_ill_spec, test_ill_wp, xyz_ref_wp, observer, dcam_ssf);

        test_profile(dcp, prof, custom_wb, skip_lut, test_ill_wp, test_ill_spec, ps, psD50, observer, dcam_ssf, report_dir);
        profio_dcp_delete(dcp);
        dcam_profile_delete(prof);

    } else if (strcmp(argv[1], "make-dcp") == 0) {

        /******************
          Make DCP
         */
        const char *report_dir = NULL;
        const char *camera_name = "Unknown Camera";
        bool skip_lut = false;
        bool skip_fwd_mtx = false;
        bool skip_observer_mapping = false;
        int hcount = 90;
        int scount = 30;
        for (; ai < argc-1 && parse_flags && argv[ai][0] == '-'; ai++) {
            switch (argv[ai][1]) {
            case '\0': ai -= 2; parse_flags = false; break;
            case 'c': camera_name = argv[++ai]; break;
            case 'h': {
                char *p = argv[++ai];
                hcount = atoi(p);
                if ((p = strchr(p, ',')) == NULL) {
                    fprintf(stderr, "Missing saturation division value (expected huediv,satdiv eg \"-h 90,30\").\n");
                    exit(EXIT_FAILURE);
                }
                scount = atoi(++p);
                if (hcount <= 0 || scount <= 0 || hcount > 10000 || scount > 10000) {
                    fprintf(stderr, "Invalid hue or saturation division value.\n");
                    exit(EXIT_FAILURE);
                }
                break;
            }
            case 'O': skip_observer_mapping = true; break;
            case 'F': skip_fwd_mtx = true; break;
            case 'L': skip_lut = true; break;
            case 'r': report_dir = argv[++ai]; break;
            default:
                fprintf(stderr, "unknown flag \"%s\".\n", argv[ai]);
                print_usage_and_exit();
                break;
            }
        }
        if (ai != argc - 2) print_usage_and_exit();
        struct dcam_profile *prof = jsonio_profile_parse(argv[ai], true);
        struct profio_dcp *dcp = make_dcp(prof, camera_name, !skip_lut, hcount, scount, skip_fwd_mtx, skip_observer_mapping, report_dir);
        dcam_profile_delete(prof);
        profio_dcp_write(dcp, argv[ai+1], true);
        profio_dcp_delete(dcp);

    } else {
        fprintf(stderr, "Unknown command \"%s\". Run command without parameters to get usage documentation.\n", argv[1]);
        exit(EXIT_FAILURE);
    }
    return 0;
}
