/*
 * (c) Copyright 2015 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */

// utility functions derived from DNG reference code
#include <stdbool.h>

#include <colmath.h>
#include <dngref.h>

// XYZ D50 to Prophoto RGB, according to DNG SDK reference code
const m3x3 dcp_xyz_D50_to_prophoto_rgb =
{{ { +1.3457530096, -0.2556031200, -0.0510246682 },
   { -0.5444259963, +1.5080962194, +0.0204718467 },
   { +0.0000000000, +0.0000000000, +1.2119675456 } }};

static const double d50_xy[2] = { 0.3457, 0.3585 };

v3
dcp_lookup(const v3 *hsm,
           const uint32_t hsmdims[3],
           const bool srgb_gamma,
           const v3 cxyz)
{
    v3 rgb = m3x3_mul_v3(dcp_xyz_D50_to_prophoto_rgb, cxyz);
    for (int i = 0; i < 3; i++) {
        if (rgb.v[i] < 0) rgb.v[i] = 0;
        else if (rgb.v[i] > 1) rgb.v[i] = 1;
    }
    v3 hsv = dcp_rgb2hsv(rgb);

    v3 hsv1;

    { // DCP ref code
        uint32_t hueDivisions = hsmdims[0];
        uint32_t satDivisions = hsmdims[1];
        uint32_t valDivisions = hsmdims[2];

        double hScale = (hueDivisions < 2) ? 0.0 : (hueDivisions * (1.0 / 6.0));
        double sScale = (double) (satDivisions - 1);
        double vScale = (double) (valDivisions - 1);
        int32_t maxHueIndex0 = hueDivisions - 1;
        int32_t maxSatIndex0 = satDivisions - 2;
        int32_t maxValIndex0 = valDivisions - 2;
        const v3 *tableBase = hsm;
        int32_t hueStep = satDivisions;
        int32_t valStep = hueDivisions * hueStep;

        double h, s, v;

        h = hsv.v[0];
        s = hsv.v[1];
        v = hsv.v[2];

        double vEncoded = v;

        double hueShift;
        double satScale;
        double valScale;
        if (valDivisions < 2) {// Optimize most common case of "2.5D" table.
            double hScaled = h * hScale;
            double sScaled = s * sScale;
            int32_t hIndex0 = (int32_t) hScaled;
            int32_t sIndex0 = (int32_t) sScaled;
            if (sIndex0 > maxSatIndex0) sIndex0 = maxSatIndex0;
            int32_t hIndex1 = hIndex0 + 1;
            if (hIndex0 >= maxHueIndex0) {
                hIndex0 = maxHueIndex0;
                hIndex1 = 0;
            }
            double hFract1 = hScaled - (double) hIndex0;
            double sFract1 = sScaled - (double) sIndex0;
            double hFract0 = 1.0f - hFract1;
            double sFract0 = 1.0f - sFract1;
            const v3 *entry00 = tableBase + hIndex0 * hueStep + sIndex0;
            const v3 *entry01 = entry00 + (hIndex1 - hIndex0) * hueStep;
            double hueShift0 = hFract0 * entry00->v[0] + hFract1 * entry01->v[0];
            double satScale0 = hFract0 * entry00->v[1] + hFract1 * entry01->v[1];
            double valScale0 = hFract0 * entry00->v[2] + hFract1 * entry01->v[2];

            entry00++;
            entry01++;

            double hueShift1 = hFract0 * entry00->v[0] + hFract1 * entry01->v[0];
            double satScale1 = hFract0 * entry00->v[1] + hFract1 * entry01->v[1];
            double valScale1 = hFract0 * entry00->v[2] + hFract1 * entry01->v[2];
            hueShift = sFract0 * hueShift0 + sFract1 * hueShift1;
            satScale = sFract0 * satScale0 + sFract1 * satScale1;
            valScale = sFract0 * valScale0 + sFract1 * valScale1;
        } else {
            double hScaled = h		  * hScale;
            double sScaled = s		  * sScale;
            if (srgb_gamma) {
                double val = v;
                if (val < 0.0) val = 0.0;
                if (val > 1.0) val = 1.0;
                vEncoded = srgb_gamma_forward(val);
            }
            double vScaled = vEncoded * vScale;
            int32_t hIndex0 = (int32_t) hScaled;
            int32_t sIndex0 = (int32_t) sScaled;
            int32_t vIndex0 = (int32_t) vScaled;
            if (sIndex0 > maxSatIndex0) sIndex0 = maxSatIndex0;
            if (vIndex0 > maxValIndex0) vIndex0 = maxValIndex0;
            int32_t hIndex1 = hIndex0 + 1;
            if (hIndex0 >= maxHueIndex0) {
                hIndex0 = maxHueIndex0;
                hIndex1 = 0;
            }
            double hFract1 = hScaled - (double) hIndex0;
            double sFract1 = sScaled - (double) sIndex0;
            double vFract1 = vScaled - (double) vIndex0;
            double hFract0 = 1.0 - hFract1;
            double sFract0 = 1.0 - sFract1;
            double vFract0 = 1.0 - vFract1;
            const v3 *entry00 = tableBase + vIndex0 * valStep +  hIndex0 * hueStep + sIndex0;
            const v3 *entry01 = entry00 + (hIndex1 - hIndex0) * hueStep;
            const v3 *entry10 = entry00 + valStep;
            const v3 *entry11 = entry01 + valStep;
            double hueShift0 = vFract0 * (hFract0 * entry00->v[0] + hFract1 * entry01->v[0]) + vFract1 * (hFract0 * entry10->v[0] + hFract1 * entry11->v[0]);
            double satScale0 = vFract0 * (hFract0 * entry00->v[1] + hFract1 * entry01->v[1]) + vFract1 * (hFract0 * entry10->v[1] + hFract1 * entry11->v[1]);
            double valScale0 = vFract0 * (hFract0 * entry00->v[2] + hFract1 * entry01->v[2]) + vFract1 * (hFract0 * entry10->v[2] + hFract1 * entry11->v[2]);
            entry00++;
            entry01++;
            entry10++;
            entry11++;

            double hueShift1 = vFract0 * (hFract0 * entry00->v[0] + hFract1 * entry01->v[0]) + vFract1 * (hFract0 * entry10->v[0] + hFract1 * entry11->v[0]);
            double satScale1 = vFract0 * (hFract0 * entry00->v[1] + hFract1 * entry01->v[1]) + vFract1 * (hFract0 * entry10->v[1] + hFract1 * entry11->v[1]);
            double valScale1 = vFract0 * (hFract0 * entry00->v[2] + hFract1 * entry01->v[2]) + vFract1 * (hFract0 * entry10->v[2] + hFract1 * entry11->v[2]);
            hueShift = sFract0 * hueShift0 + sFract1 * hueShift1;
            satScale = sFract0 * satScale0 + sFract1 * satScale1;
            valScale = sFract0 * valScale0 + sFract1 * valScale1;
        }

        hueShift *= (6.0 / 360.0);	// Convert to internal hue range.
        h += hueShift;
        if (h > 6.0) h -= 6.0;
        if (h < 0) h += 6.0;
        s = s * satScale;
        if (s > 1.0) s = 1.0;

        vEncoded = vEncoded * valScale;
        if (vEncoded < 0) vEncoded = 0;
        if (vEncoded > 1) vEncoded = 1;

        if (srgb_gamma) {
            v = srgb_gamma_inverse(vEncoded);
        } else {
            v = vEncoded;
        }

        hsv1 = (v3){{ h, s, v }};
    }

    // output 
    v3 rgb1 = dcp_hsv2rgb(hsv1);
    const m3x3 rgb2xyz = invert_m3x3(dcp_xyz_D50_to_prophoto_rgb);
    v3 xyz1 = m3x3_mul_v3(rgb2xyz, rgb1);
    return xyz1;
}

// rgb range 0 - 1, hsv range 0 - 1 for s and v, 0 - 6 for h
v3
dcp_rgb2hsv(v3 rgb)
{
    double h, s, v;
    double largest = rgb.v[0];
    double smallest = rgb.v[0];
    for (int i = 1; i < 3; i++) {
        if (rgb.v[i] > largest) largest = rgb.v[i];
        if (rgb.v[i] < smallest) smallest = rgb.v[i];
    }
    v = largest;
    double span = largest - smallest;
    if (span > 0.0) {
        if (rgb.v[0] == v) {
            h = (rgb.v[1] - rgb.v[2]) / span;
            if (h < 0.0) {
                h += 6.0;
            }
        } else if (rgb.v[1] == v) {
            h = 2.0 + (rgb.v[2] - rgb.v[0]) / span;
        } else {
            h = 4.0 + (rgb.v[0] - rgb.v[1]) / span;
        }
        s = span / v;
    } else {
        h = s = 0.0;
    }
    return (v3){{ h, s, v }};
}

v3
dcp_hsv2rgb(const v3 hsv)
{
    double h = hsv.v[0], s = hsv.v[1], v = hsv.v[2], r, g, b;
    if (s > 0.0f) {
        if (h < 0.0f) {
            h += 6.0f;
        }
        if (h >= 6.0f) {
            h -= 6.0f;
        }
        int  i = (int)h;
        double f = h - (double)i;
        double p = v * (1.0f - s);
#define q       (v * (1.0f - s * f))
#define t       (v * (1.0f - s * (1.0f - f)))
        switch (i) {
        case 0: r = v; g = t; b = p; break;
        case 1: r = q; g = v; b = p; break;
        case 2: r = p; g = v; b = t; break;
        case 3: r = p; g = q; b = v; break;
        case 4: r = t; g = p; b = v; break;
        case 5: r = v; g = p; b = q; break;
        default: r = g = b = 0; // never reached
        }
#undef q
#undef t
    } else {
        r = v;
        g = v;
        b = v;
    }
    return (v3){{ r, g, b }};
}

struct ruvt {
    double r;
    double u;
    double v;
    double t;
};

static const double kTintScale = -3000.0;
static const struct ruvt kTempTable [] =
	{
	{   0, 0.18006, 0.26352, -0.24341 },
	{  10, 0.18066, 0.26589, -0.25479 },
	{  20, 0.18133, 0.26846, -0.26876 },
	{  30, 0.18208, 0.27119, -0.28539 },
	{  40, 0.18293, 0.27407, -0.30470 },
	{  50, 0.18388, 0.27709, -0.32675 },
	{  60, 0.18494, 0.28021, -0.35156 },
	{  70, 0.18611, 0.28342, -0.37915 },
	{  80, 0.18740, 0.28668, -0.40955 },
	{  90, 0.18880, 0.28997, -0.44278 },
	{ 100, 0.19032, 0.29326, -0.47888 },
	{ 125, 0.19462, 0.30141, -0.58204 },
	{ 150, 0.19962, 0.30921, -0.70471 },
	{ 175, 0.20525, 0.31647, -0.84901 },
	{ 200, 0.21142, 0.32312, -1.0182 },
	{ 225, 0.21807, 0.32909, -1.2168 },
	{ 250, 0.22511, 0.33439, -1.4512 },
	{ 275, 0.23247, 0.33904, -1.7298 },
	{ 300, 0.24010, 0.34308, -2.0637 },
	{ 325, 0.24702, 0.34655, -2.4681 },
	{ 350, 0.25591, 0.34951, -2.9641 },
	{ 375, 0.26400, 0.35200, -3.5814 },
	{ 400, 0.27218, 0.35407, -4.3633 },
	{ 425, 0.28039, 0.35577, -5.3762 },
	{ 450, 0.28863, 0.35714, -6.7262 },
	{ 475, 0.29685, 0.35823, -8.5955 },
	{ 500, 0.30505, 0.35907, -11.324 },
	{ 525, 0.31320, 0.35968, -15.628 },
	{ 550, 0.32129, 0.36011, -23.325 },
	{ 575, 0.32931, 0.36038, -40.770 },
	{ 600, 0.33724, 0.36051, -116.45 }
	};

static void
dngref_XYCoord2Temperature(const double whiteXY[2],
                           double *temp,
                           double *tint)
{
    double fTemperature = 0;
    double fTint = 0;

    // Convert to uv space.
    double u = 2.0 * whiteXY[0] / (1.5 - whiteXY[0] + 6.0 * whiteXY[1]);
    double v = 3.0 * whiteXY[1] / (1.5 - whiteXY[0] + 6.0 * whiteXY[1]);

    // Search for line pair coordinate is between.
    double last_dt = 0.0;
    double last_dv = 0.0;
    double last_du = 0.0;

    for (uint32_t index = 1; index <= 30; index++) {
        // Convert slope to delta-u and delta-v, with length 1.
        double du = 1.0;
        double dv = kTempTable [index] . t;
        double len = sqrt (1.0 + dv * dv);
        du /= len;
        dv /= len;

        // Find delta from black body point to test coordinate.
        double uu = u - kTempTable [index] . u;
        double vv = v - kTempTable [index] . v;

        // Find distance above or below line.
        double dt = - uu * dv + vv * du;

        // If below line, we have found line pair.
        if (dt <= 0.0 || index == 30) {
            // Find fractional weight of two lines.
            if (dt > 0.0)
                dt = 0.0;
            dt = -dt;
            double f;
            if (index == 1)
            {
                f = 0.0;
            }
            else
            {
                f = dt / (last_dt + dt);
            }

            // Interpolate the temperature.
            fTemperature = 1.0E6 / (kTempTable [index - 1] . r * f +
                                    kTempTable [index    ] . r * (1.0 - f));

            // Find delta from black body point to test coordinate.
            uu = u - (kTempTable [index - 1] . u * f +
                      kTempTable [index    ] . u * (1.0 - f));
            vv = v - (kTempTable [index - 1] . v * f +
                      kTempTable [index    ] . v * (1.0 - f));
            // Interpolate vectors along slope.
            du = du * (1.0 - f) + last_du * f;
            dv = dv * (1.0 - f) + last_dv * f;
            len = sqrt (du * du + dv * dv);
            du /= len;
            dv /= len;

            // Find distance along slope.
            fTint = (uu * du + vv * dv) * kTintScale;
            break;
        }
        // Try next line pair.
        last_dt = dt;
        last_du = du;
        last_dv = dv;
    }
    if (temp != NULL)
        *temp = fTemperature;
    if (tint != NULL)
        *tint = fTint;
}

static double
dngref_get_mix(const struct profio_dcp *dcp,
               const double whiteXY[2])
{
    const double fTemperature1 = exif_lightsource_temp(dcp->ill[0]);
    const double fTemperature2 = dcp->ill_count == 2 ? exif_lightsource_temp(dcp->ill[1]) : 0;
    // Convert to temperature/offset space.
    double wbtemp;
    dngref_XYCoord2Temperature(whiteXY, &wbtemp, NULL);

    // Find fraction to weight the first calibration.
    double g;
    if (wbtemp <= fTemperature1 || dcp->ill_count == 1) {
        g = 1.0;
    } else if (wbtemp >= fTemperature2) {
        g = 0.0;
    } else {
        double invT = 1.0 / wbtemp;
        g = (invT                  - (1.0 / fTemperature2)) /
            ((1.0 / fTemperature1) - (1.0 / fTemperature2));
    }
    return g;
}

static m3x3
dngref_FindXYZtoCamera(const struct profio_dcp *dcp,
                       const double whiteXY[2],
                       m3x3 *forwardMatrix,
                       m3x3 *reductionMatrix,
                       m3x3 *cameraCalibration)
{
    const double g = dngref_get_mix(dcp, whiteXY);
    // Interpolate the color matrix.
    m3x3 colorMatrix;
    if (g >= 1.0)
        colorMatrix = dcp->cm[0];
    else if (g <= 0.0)
        colorMatrix = dcp->cm[1];
    else {
        colorMatrix = m3x3_add_m3x3(k_mul_m3x3(g, dcp->cm[0]), k_mul_m3x3(1.0 - g, dcp->cm[1]));
    }

    // Interpolate forward matrix, if any.
    if (forwardMatrix) {
        bool has1 = dcp->has.fm;
        bool has2 = has1 && dcp->ill_count == 2;
        if (has1 && has2) {
            if (g >= 1.0)
                *forwardMatrix = dcp->fm[0];
            else if (g <= 0.0)
                *forwardMatrix = dcp->fm[1];
            else
                *forwardMatrix = m3x3_add_m3x3(k_mul_m3x3(g, dcp->fm[0]), k_mul_m3x3(1.0 - g, dcp->fm[1]));
        } else if (has1) {
            *forwardMatrix = dcp->fm[0];
        } else if (has2) {
            *forwardMatrix = dcp->fm[1];
        } else {
            memset(forwardMatrix, 0, sizeof(*forwardMatrix));
        }
    }

    /*
    // Interpolate reduction matrix, if any.
    if (reductionMatrix) {
        bool has1 = fReductionMatrix1.NotEmpty ();
        bool has2 = fReductionMatrix2.NotEmpty ();
        if (has1 && has2) {
            if (g >= 1.0)
                *reductionMatrix = fReductionMatrix1;
            else if (g <= 0.0)
                *reductionMatrix = fReductionMatrix2;
            else
                *reductionMatrix = (g      ) * fReductionMatrix1 +
                    (1.0 - g) * fReductionMatrix2;
        } else if (has1) {
            *reductionMatrix = fReductionMatrix1;
        } else if (has2) {
            *reductionMatrix = fReductionMatrix2;
        } else {
            reductionMatrix->Clear ();
        }
    }
    // Interpolate camera calibration matrix.
    if (cameraCalibration) {
        if (g >= 1.0)
            *cameraCalibration = fCameraCalibration1;
        else if (g <= 0.0)
            *cameraCalibration = fCameraCalibration2;
        else
            *cameraCalibration = (g      ) * fCameraCalibration1 +
                (1.0 - g) * fCameraCalibration2;
    }
    */

    // Return the interpolated color matrix.
    return colorMatrix;
}

static void
dngref_XYZtoXY(v3 XYZ, double XY[2])
{
    double X = XYZ.v[0];
    double Y = XYZ.v[1];
    double Z = XYZ.v[2];
    double total = X + Y + Z;
    if (total > 0.0) {
        XY[0] = X / total;
        XY[1] = Y / total;
    } else {
        XY[0] = 0.3457;
        XY[1] = 0.3585;
    }
}

static v3
dngref_XYtoXYZ(const double XY[2])
{
    double temp[2] = { XY[0], XY[1] };
    // Restrict xy coord to someplace inside the range of real xy coordinates.
    // This prevents math from doing strange things when users specify
    // extreme temperature/tint coordinates.
    for (int i = 0; i < 2; i++) {
        if (temp[i] < 0.000001) temp[i] = 0.000001;
        if (temp[i] > 0.999999) temp[i] = 0.999999;
    }
    if (temp[0] + temp[1] > 0.999999) {
        double scale = 0.999999 / (temp[0] + temp[1]);
        temp[0] *= scale;
        temp[1] *= scale;
    }
    v3 XYZ;
    XYZ.v[0] = temp[0] / temp[1];
    XYZ.v[1] = 1.0;
    XYZ.v[2] = (1.0 - temp[0] - temp[1]) / temp[1];
    return XYZ;
}

void
dcp_neutral2xy(const struct profio_dcp *dcp,
               v3 neutral,
               double xy[2])
{
    const int kMaxPasses = 30;
    double lastXY[2] = { 0.3457, 0.3585 }; // D50
    for (int pass = 0; pass < kMaxPasses; pass++) {
        m3x3 xyzToCamera = dngref_FindXYZtoCamera(dcp, lastXY, NULL, NULL, NULL);

        double nextXY[2];
        v3 nextXYZ = m3x3_mul_v3(invert_m3x3(xyzToCamera), neutral);
        dngref_XYZtoXY(nextXYZ, nextXY);

        if (fabs(nextXY[0] - lastXY[0]) +
            fabs(nextXY[1] - lastXY[1]) < 0.0000001)
        {
            xy[0] = nextXY[0];
            xy[1] = nextXY[1];
            return;
        }
        if (pass == kMaxPasses - 1) {
            nextXY[0] = (lastXY[0] + nextXY[0]) * 0.5;
            nextXY[1] = (lastXY[1] + nextXY[1]) * 0.5;
        }
        lastXY[0] = nextXY[0];
        lastXY[1] = nextXY[1];
    }
    xy[0] = lastXY[0];
    xy[1] = lastXY[1];
}

/*
static m3x3
dngref_SetWhiteXY(struct profio_dcp *dcp,
                  const double whiteXY[2])
{
    // Interpolate an matric values for this white point.
    m3x3 colorMatrix;
    m3x3 forwardMatrix;
    m3x3 reductionMatrix;
    m3x3 cameraCalibration;
    colorMatrix = dngref_FindXYZtoCamera(dcp,
                                         whiteXY,
                                         &forwardMatrix,
                                         &reductionMatrix,
                                         &cameraCalibration);
    // Find the camera white values.
    v3 fCameraWhite = m3x3_mul_v3(colorMatrix, XYtoXYZ(whiteXY));

    double whiteScale = 1.0 / v3_max(fCameraWhite);
    for (uint32 j = 0; j < 3; j++) {
        // We don't support non-positive values for camera neutral values.
        fCameraWhite.v[j] = whiteScale * fCameraWhite.v[j];
        if (fCameraWhite.v[j] < 0.001) fCameraWhite.v[j] = 0.001;
        if (fCameraWhite.v[j] > 1.0) fCameraWhite.v[j] = 1.0;
    }

    // Find PCS to Camera transform. Scale matrix so PCS white can just be
    // reached when the first camera channel saturates
    const double d50_xy[2] = { 0.3457, 0.3585 };
    const v3 PCStoXYZ = dngref_XYtoXYZ(d50_xy);
    m3x3 fPCStoCamera = m3x3_mul_m3x3(colorMatrix, bradford_map_white(PCStoXYZ, dngref_XYtoXYZ(whiteXY)));

    double scale = v3_max(mul_m3x3_v3(fPCStoCamera, PCStoXYZ));

    fPCStoCamera = k_mul_m3x3(1.0 / scale, fPCStoCamera);

    // If we have a forward matrix, then just use that.
    m3x3 fCameraToPCS;
    if (dcp->has.fm) {
        //dng_matrix individualToReference = Invert (fAnalogBalance * cameraCalibration);
        //dng_vector refCameraWhite = individualToReference * fCameraWhite;
        fCameraToPCS = m3x3_mul_m3x3(fm, invert_m3x3(v3_asdiagonal(fCameraWhite)));
        //fCameraToPCS = forwardMatrix * Invert (refCameraWhite.AsDiagonal ()) * individualToReference;
    } else { // Else we need to use the adapt in XYZ method.
        // Invert this PCS to camera matrix.  Note that if there are more than three
        // camera channels, this inversion is non-unique.
        //fCameraToPCS = Invert (fPCStoCamera, reductionMatrix);
        fCameraToPCS = invert_m3x3(fPCStoCamera);
    }
    return fCameraToPCS; // camera RGB to XYZ D50
}
*/

m3x3
dcp_make_cam2xyz_d50(const struct profio_dcp *dcp,
                     const v3 wb)
{
    m3x3 cm = dcp->cm[0];
    m3x3 fm = dcp->fm[0];
    if (dcp->ill_count == 2) {
        double xy[2];
        dcp_neutral2xy(dcp, wb, xy);
        cm = dngref_FindXYZtoCamera(dcp, xy, &fm, NULL, NULL);
    }
    m3x3 cam2xyz;
    if (!dcp->has.fm) {
        v3 wp = v3_norm(m3x3_mul_v3(invert_m3x3(cm), wb));
        m3x3 map = bradford_map_white(dcp_d50(), wp);
        cam2xyz = invert_m3x3(m3x3_mul_m3x3(cm, map));
    } else {
        // just apply white balance to forward matrix
        cam2xyz = m3x3_mul_m3x3(fm, invert_m3x3(v3_asdiagonal(wb)));
    }
    return cam2xyz;
}

v3 *
dcp_make_lut(const struct profio_dcp *dcp,
             const v3 wb)
{
    size_t count = dcp->hsmdims[0]*dcp->hsmdims[1]*dcp->hsmdims[2];
    v3 *hsm = malloc(count * sizeof(*hsm));
    if (dcp->ill_count == 1 || dcp->huesatmap[1] == NULL) {
        if (dcp->huesatmap[0] == NULL) {
            return NULL;
        }
        memcpy(hsm, dcp->huesatmap[0], count * sizeof(*hsm));
        return hsm;
    }
    double xy[2];
    dcp_neutral2xy(dcp, wb, xy);
    double mix = dngref_get_mix(dcp, xy);

    if (mix >= 1.0) {
        memcpy(hsm, dcp->huesatmap[0], count * sizeof(*hsm));
        return hsm;
    } else if (mix <= 0.0) {
        memcpy(hsm, dcp->huesatmap[1], count * sizeof(*hsm));
        return hsm;
    }

    // Interpolate between the tables.
    double w1 = mix;
    double w2 = 1.0 - mix;
    for (size_t i = 0; i < count; i++) {
        for (int k = 0; k < 3; k++) {
            hsm[i].v[k] = w1 * dcp->huesatmap[0][i].v[k] + w2 * dcp->huesatmap[1][i].v[k];
        }
    }
    return hsm;
}

v3
dcp_d50(void)
{
    return xy2xyz_lightest(d50_xy[0], d50_xy[1], (v3){{1,1,1}});
}
