/*
 * (c) Copyright 2015 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <jsonio.h>
#include <lut.h>
#include <spectraldb.h>
#include <cJSON.h>

static bool
file_exists(const char filename[])
{
    struct stat buf;
    return (stat(filename, &buf) == 0);
}

static char *
file2string(const char filename[])
{
    FILE *stream = fopen(filename, "rt");
    if (stream == NULL) {
        fprintf(stderr, "Could not open file \"%s\": %s\n", filename, strerror(errno));
        return NULL;
    }
    size_t bufsize = 4096;
    size_t datasize = 0, ret;
    char *buf = (char *)malloc(bufsize);
    while ((ret = fread(&buf[datasize], 1, bufsize - datasize, stream)) != 0) {
        datasize += ret;
        if (datasize == bufsize) {
            bufsize *= 2;
            buf = (char *)realloc(buf, bufsize);
        }
    }
    if (!feof(stream)) {
        fclose(stream);
        free(buf);
        fprintf(stderr, "Failed to read file \"%s\"\n", filename);
        return NULL;
    }
    fclose(stream);
    buf = (char *)realloc(buf, datasize + 1);
    buf[datasize] = '\0';
    return buf;
}

static cJSON *
json_parse(const char filename[])
{
    char *str = file2string(filename);
    if (str == NULL) {
        return NULL;
    }
    cJSON_Minify(str); // cJSON requires comments to be removed before parsing
    cJSON *js = cJSON_Parse(str);
    free(str);
    if (js == NULL) {
        fprintf(stderr, "Invalid JSON data in file \"%s\"\n", filename);
        return NULL;
    }
    return js;
}

static void
json_print(cJSON *js)
{
    char *js_str = cJSON_Print(js);
    printf("%s\n", js_str);
    free(js_str);
}

static bool
get_and_test(cJSON **js_,
             cJSON *jsroot,
             const char name[],
             int type,
             bool is_required)
{
    cJSON *js = cJSON_GetObjectItem(jsroot, name);
    *js_ = NULL;
    if (js == NULL) {
        if (is_required) {
            fprintf(stderr, "Required field \"%s\" is missing.\n", name);
            return false;
        }
        return true;
    }
    if (js->type != type) {
        fprintf(stderr, "Type mismatch for field \"%s\".\n", name);
        return false;
    }
    *js_ = js;
    return true;
}

static double *
json_numberarray(cJSON *js,
                 int *count_)
{
    if (js == NULL) {
        fprintf(stderr, "Missing array.\n");
        return NULL;
    }
    if (js->type != cJSON_Array) {
        fprintf(stderr, "Expected array, got other type.\n");
        return NULL;
    }
    cJSON *ji;
    int count = 0;
    for (ji = js->child; ji != NULL; ji = ji->next) {
        if (ji->type != cJSON_Number) {
            fprintf(stderr, "Expected Number in array, got other type.\n");
            return NULL;
        }
        count++;
    }
    if (count == 0) {
        fprintf(stderr, "Zero length array, expected more.\n");
        return NULL;
    }
    int i;
    double *a = malloc(sizeof(*a) * count);
    for (ji = js->child, i = 0; ji != NULL; ji = ji->next, i++) {
        a[i] = ji->valuedouble;
    }
    *count_ = count;
    return a;
}

static double *
json_spectrum_bands_array(cJSON *js,
                          int *count_)
{
    int count;
    double *bands = json_numberarray(js, &count);
    if (bands == NULL) {
        fprintf(stderr, "Could not parse JSON number array.\n");
        return NULL;
    }
    if (count != 3 || bands[2] > bands[1]) {
        for (int i = 0; i < count; i++) {
            if (bands[i] < 0 || (i > 0 && bands[i] <= bands[i-1])) {
                fprintf(stderr, "Invalid bands for spectrum.\n");
                free(bands);
                return NULL;
            }
        }
        *count_ = count;
        return bands;
    }
    if (bands[0] >= bands[1]) {
        fprintf(stderr, "Invalid bands for spectrum.\n");
        free(bands);
        return NULL;
    }
    int bspec[3] = { (int)bands[0], (int)bands[1], (int)bands[2] };
    free(bands);
    if (bspec[2] == 0 || (bspec[1] - bspec[0]) % bspec[2] != 0) {
        fprintf(stderr, "Invalid bands for spectrum.\n");
        return NULL;
    }
    count = (bspec[1] - bspec[0]) / bspec[2] + 1;
    if (count <= 2 || count > 1000000) {
        fprintf(stderr, "Invalid bands for spectrum.\n");
        return NULL;
    }
    double *ob = malloc(count * sizeof(ob[0]));
    for (int i = 0; i < count; i++) {
        ob[i] = bspec[0] + i * bspec[2];
    }
    ob[count-1] = bspec[1];
    if (ob[count-2] > ob[count-1]) {
        fprintf(stderr, "Invalid bands for spectrum.\n");
        free(ob);
        return NULL;
    }
    *count_ = count;
    return ob;
}

struct dcam_ssf *
jsonio_dcam_ssf_parse(const char filename[],
                      bool parse_error_is_fatal)
{
    cJSON *js = json_parse(filename);
    if (js == NULL) {
        if (parse_error_is_fatal) exit(EXIT_FAILURE);
        return NULL;
    }
    // json_print(js);
    int band_count;
    double *band_f = NULL;
    spectrum_t *ssfs[3] = { NULL, NULL, NULL };
    cJSON *camera_name = cJSON_GetObjectItem(js, "camera_name");
    cJSON *bands = cJSON_GetObjectItem(js, "ssf_bands");
    cJSON *spec[3];
    spec[0] = cJSON_GetObjectItem(js, "red_ssf");
    spec[1] = cJSON_GetObjectItem(js, "green_ssf");
    spec[2] = cJSON_GetObjectItem(js, "blue_ssf");

    if (camera_name == NULL || camera_name->type != cJSON_String) {
        goto fail;
    }
    band_f = json_spectrum_bands_array(bands, &band_count);
    if (band_f == NULL) {
        goto fail;
    }
    for (int i = 0; i < 3; i++) {
        int amp_count;
        double *amp = json_numberarray(spec[i], &amp_count);
        if (amp == NULL || amp_count != band_count) {
            free(amp);
            goto fail;
        }
        ssfs[i] = spec_alloc1(band_f, amp, band_count);
        free(amp);
        if (ssfs[i] == NULL) {
            goto fail;
        }
    }
    free(band_f);
    double maxa = 0;
    for (int i = 0; i < 3; i++) {
        if (spec_max(ssfs[i]) > maxa) maxa = spec_max(ssfs[i]);
    }
    if (maxa != 1.0) {
        for (int i = 0; i < 3; i++) {
            spec_scalar_multiply(ssfs[i], 1.0 / maxa);
        }
    }
    struct dcam_ssf *ssf = calloc(1, sizeof(*ssf));
    ssf->camera_name = strdup(camera_name->valuestring);
    ssf->ssf.cmf[0] = ssfs[0];
    ssf->ssf.cmf[1] = ssfs[1];
    ssf->ssf.cmf[2] = ssfs[2];
    cJSON_Delete(js);
    return ssf;
fail:
    free(ssfs[0]);
    free(ssfs[1]);
    free(ssfs[2]);
    free(band_f);
    cJSON_Delete(js);
    fprintf(stderr, "Could not parse Digital Camera SSF file \"%s\".\n", filename);
    if (parse_error_is_fatal) exit(EXIT_FAILURE);
    return NULL;
}

spectrum_t *
jsonio_spectrum_parse(const char filename[],
                      bool parse_error_is_fatal)
{
    cJSON *js = json_parse(filename);
    if (js == NULL) {
        if (parse_error_is_fatal) exit(EXIT_FAILURE);
        return NULL;
    }
    cJSON *js_bands = cJSON_GetObjectItem(js, "bands");
    cJSON *js_scale = cJSON_GetObjectItem(js, "scale");
    cJSON *js_spectrum = cJSON_GetObjectItem(js, "spectrum");
    int band_count;
    double *bands = NULL;
    double scale = 1.0;
    if (js_scale != NULL) {
        if (js_scale->type != cJSON_Number) {
            fprintf(stderr, "\"scale\" is not a number.\n");
            goto fail;
        }
        scale = js->valuedouble;
    }
    bands = json_spectrum_bands_array(js_bands, &band_count);
    if (bands == NULL) {
        goto fail;
    }
    int amp_count;
    double *amp = json_numberarray(js_spectrum, &amp_count);
    if (amp == NULL || amp_count != band_count) {
        free(amp);
        goto fail;
    }
    if (scale != 1.0) {
        for (int i = 0; i < band_count; i++) {
            amp[i] *= scale;
        }
    }
    spectrum_t *s = spec_alloc1(bands, amp, band_count);
    if (s == NULL) {
        free(amp);
        goto fail;
    }
    free(amp);
    cJSON_Delete(js);
    return s;
fail:
    free(bands);
    fprintf(stderr, "Could not parse Spectrum file \"%s\".\n", filename);
    cJSON_Delete(js);
    if (parse_error_is_fatal) exit(EXIT_FAILURE);
    return NULL;
}

void
jsonio_spectrum_print(FILE *stream,
                      const char indent[],
                      const char name[],
                      const spectrum_t *s)
{
    fprintf(stream,
            "%s\"%s\": {\n"
            "%s  \"bands\": [ %g",
            indent, name, indent, s->v[0].f);
    for (int i = 1; i < s->band_count; i++) {
        fprintf(stream, ", %g", s->v[i].f);
    }
    fprintf(stream,
            "],\n"
            "%s  \"spectrum\": [ %g",
            indent, s->v[0].e);
    for (int i = 1; i < s->band_count; i++) {
        fprintf(stream, ", %g", s->v[i].e);
    }
    fprintf(stream,
            "]\n"
            "%s}",
            indent);
}

spectrum_t *
jsonio_illuminant_parse(const char filename[],
                        bool parse_error_is_fatal)
{
    const spectrum_t *ill = NULL;
    enum exif_lightsource ls = exif_lightsource_aton(filename);
    if ((int)ls > 0) {
        ill = spectraldb_illuminant(ls);
    } else {
        ill = spectraldb_illuminant_byname(filename, false);
    }
    if (ill != NULL) {
        return spec_copy(ill);
    }
    if (file_exists(filename)) {
        return jsonio_spectrum_parse(filename, parse_error_is_fatal);
    }
    if (parse_error_is_fatal) {
        // run this to get error message
        spectraldb_illuminant_byname(filename, true);
        exit(EXIT_FAILURE);
    }
    return NULL;
}

static bool
matrix3x3_parse(m3x3 *m_,
                cJSON *mjs)
{
    if (mjs->type != cJSON_Array) {
        fprintf(stderr, "Expected array type, got other.\n");
        goto fail;
    }
    int row = 0;
    m3x3 m;
    for (cJSON *ji = mjs->child; ji != NULL; ji = ji->next) {
        if (ji->type != cJSON_Array) {
            fprintf(stderr, "Expected array in array, got other type.\n");
            goto fail;
        }
        if (++row > 3) goto dimension_fail;
        int col = 0;
        for (cJSON *ji1 = ji->child; ji1 != NULL; ji1 = ji1->next) {
            if (ji1->type != cJSON_Number) {
                fprintf(stderr, "Expected Number in inner array, got other type.\n");
                goto fail;
            }
            if (++col > 3) goto dimension_fail;
            m.v[row-1][col-1] = ji1->valuedouble;
        }
        if (col != 3) goto dimension_fail;
    }
    if (row != 3) goto dimension_fail;
    *m_ = m;
    return true;
dimension_fail:
    fprintf(stderr, "Expected 3x3 matrix, got other dimensions.\n");
    return false;
fail:
    return false;
}

m3x3 *
jsonio_3x3matrix_parse(const char filename[],
                       const char optional_matrix_name[],
                       bool parse_error_is_fatal)
{
    cJSON *js = json_parse(filename);
    if (js == NULL) {
        if (parse_error_is_fatal) exit(EXIT_FAILURE);
        return NULL;
    }
    cJSON *mjs = NULL;
    if (optional_matrix_name != NULL) {
        mjs = cJSON_GetObjectItem(js, optional_matrix_name);
    }
    if (mjs == NULL) {
        mjs = cJSON_GetObjectItem(js, "matrix");
        if (mjs == NULL) {
            mjs = js;
        }
    }
    m3x3 m;
    if (!matrix3x3_parse(&m, mjs)) {
        goto fail;
    }
    cJSON_Delete(js);
    m3x3 *m_ = malloc(sizeof(*m_));
    *m_ = m;
    return m_;

fail:
    fprintf(stderr, "Could not parse Matrix file \"%s\".\n", filename);
    cJSON_Delete(js);
    if (parse_error_is_fatal) exit(EXIT_FAILURE);
    return NULL;
}

void
jsonio_3x3matrix_print(FILE *stream,
                       const char indent[],
                       const char name[],
                       m3x3 m)
{
    fprintf(stream,
            "%s\"%s\": [\n"
            "%s  [ %9.6f, %9.6f, %9.6f ],\n"
            "%s  [ %9.6f, %9.6f, %9.6f ],\n"
            "%s  [ %9.6f, %9.6f, %9.6f ]\n"
            "%s]",
            indent, name,
            indent, m.v[0][0], m.v[0][1], m.v[0][2],
            indent, m.v[1][0], m.v[1][1], m.v[1][2],
            indent, m.v[2][0], m.v[2][1], m.v[2][2],
            indent);
}

static v3 *
parse_dcp_huesatmap(const uint32_t dims[],
                    cJSON *js)
{
    // { "HueDiv:"  0, "SatDiv:"  0, "ValDiv:"  0, "HueShift:" +0.000000, "SatScale:" 1.000000, "ValScale": 1.000000 },
    double dimsize = (double)dims[0] * (double)dims[1] * (double)dims[2];
    if (dimsize > 100 * 1024 * 1024 || dimsize < 1) {
        fprintf(stderr, "Bad table dimensions.\n");
        return NULL;
    }
    int divs[3] = { 0,0,0 };
    v3 *hsv = malloc(dims[0]*dims[1]*dims[2] * sizeof(*hsv));
    int k = 0;
    for (cJSON *ji = js->child; ji != NULL; ji = ji->next, k++) {
        if (ji->type != cJSON_Object) {
            fprintf(stderr, "Expected object array, got other type.\n");
            goto fail;
        }
        cJSON *fields[6];
        if (!get_and_test(&fields[0], ji, "HueDiv", cJSON_Number, true)) goto fail;
        if (!get_and_test(&fields[1], ji, "SatDiv", cJSON_Number, true)) goto fail;
        if (!get_and_test(&fields[2], ji, "ValDiv", cJSON_Number, true)) goto fail;
        if (!get_and_test(&fields[3], ji, "HueShift", cJSON_Number, true)) goto fail;
        if (!get_and_test(&fields[4], ji, "SatScale", cJSON_Number, true)) goto fail;
        if (!get_and_test(&fields[5], ji, "ValScale", cJSON_Number, true)) goto fail;
        for (int i = 0; i < 3; i++) {
            if (fields[i]->valueint != divs[i]) {
                fprintf(stderr, "Bad HueSatDiv indexing.\n");
                goto fail;
            }
            hsv[k].v[i] = fields[3+i]->valuedouble;
        }
        divs[2]++;
        if (divs[2] == dims[2]) {
            divs[2] = 0;
            divs[1]++;
            if (divs[1] == dims[1]) {
                divs[1] = 0;
                divs[0]++;
                if (divs[0] == dims[0] && ji->next != NULL) {
                    fprintf(stderr, "Expected object array, got other type.\n");
                    goto fail;
                }
            }
        }
    }
    return hsv;
fail:
    free(hsv);
    return NULL;
}

struct profio_dcp *
jsonio_dcp_parse(const char filename[],
                 bool parse_error_is_fatal)
{
    struct profio_dcp *dcp = NULL;
    cJSON *jsroot = json_parse(filename);
    if (jsroot == NULL) {
        if (parse_error_is_fatal) exit(EXIT_FAILURE);
        return NULL;
    }

    cJSON *camera_name, *profile_name, *copyright, *embed_policy, *calibration_sig, *ill[2], *cm[2], *fm[2], *black_render, *baseline_exp;
    cJSON *look_dims, *look_enc, *hsm_dims, *hsm_enc, *hsm[2], *look, *tonecurve;
    if (!get_and_test(&camera_name, jsroot, "UniqueCameraModel", cJSON_String, true)) goto fail;
    if (!get_and_test(&profile_name, jsroot, "ProfileName", cJSON_String, false)) goto fail;
    if (!get_and_test(&copyright, jsroot, "ProfileCopyright", cJSON_String, false)) goto fail;
    if (!get_and_test(&embed_policy, jsroot, "ProfileEmbedPolicy", cJSON_String, false)) goto fail;
    if (!get_and_test(&calibration_sig, jsroot, "CalibrationSignature", cJSON_String, false)) goto fail;
    if (!get_and_test(&ill[0], jsroot, "CalibrationIlluminant1", cJSON_String, true)) goto fail;
    if (!get_and_test(&ill[1], jsroot, "CalibrationIlluminant2", cJSON_String, false)) goto fail;
    if (!get_and_test(&cm[0], jsroot, "ColorMatrix1", cJSON_Array, true)) goto fail;
    if (!get_and_test(&cm[1], jsroot, "ColorMatrix2", cJSON_Array, ill[1] != NULL)) goto fail;
    if (!get_and_test(&fm[0], jsroot, "ForwardMatrix1", cJSON_Array, false)) goto fail;
    if (!get_and_test(&fm[1], jsroot, "ForwardMatrix2", cJSON_Array, fm[0] != NULL && ill[1] != NULL)) goto fail;
    if (!get_and_test(&black_render, jsroot, "DefaultBlackRender", cJSON_String, false)) goto fail;
    if (!get_and_test(&baseline_exp, jsroot, "BaselineExposureOffset", cJSON_Number, false)) goto fail;
    if (!get_and_test(&tonecurve, jsroot, "ProfileToneCurve", cJSON_Array, false)) goto fail;
    if (!get_and_test(&hsm[0], jsroot, "ProfileHueSatMap1", cJSON_Array, false)) goto fail;
    if (!get_and_test(&hsm[1], jsroot, "ProfileHueSatMap2", cJSON_Array, hsm[0] != NULL && ill[1] != NULL)) goto fail;
    if (!get_and_test(&hsm_dims, jsroot, "ProfileHueSatMapDims", cJSON_Array, hsm[0] != NULL)) goto fail;
    if (!get_and_test(&look, jsroot, "ProfileLookTable", cJSON_Array, false)) goto fail;
    if (!get_and_test(&look_dims, jsroot, "ProfileLookTableDims", cJSON_Array, look != NULL)) goto fail;
    if (!get_and_test(&hsm_enc, jsroot, "ProfileHueSatMapEncoding", cJSON_String, false)) goto fail;
    if (!get_and_test(&look_enc, jsroot, "ProfileLookTableEncoding", cJSON_String, false)) goto fail;

    dcp = calloc(1, sizeof(*dcp));
    dcp->camera_model = strdup(camera_name->valuestring);
    if (profile_name) dcp->profile_name = strdup(profile_name->valuestring);
    if (copyright) dcp->copyright = strdup(copyright->valuestring);
    if (calibration_sig) dcp->calibration_signature = strdup(calibration_sig->valuestring);
    dcp->ill_count = (ill[1] != NULL) ? 2 : 1;
    if (hsm_dims != NULL) {
        int count;
        double *array = json_numberarray(hsm_dims, &count);
        if (count != 3) {
            fprintf(stderr, "Expected three integer array for ProfileHueSatMapDims.\n");
            free(array);
            goto fail;
        }
        dcp->hsmdims[0] = (int)array[0];
        dcp->hsmdims[1] = (int)array[1];
        dcp->hsmdims[2] = (int)array[2];
        free(array);
    }
    if (look_dims != NULL) {
        int count;
        double *array = json_numberarray(look_dims, &count);
        if (count != 3) {
            fprintf(stderr, "Expected three integer array for ProfileLookTableDims.\n");
            free(array);
            goto fail;
        }
        dcp->lookdims[0] = (int)array[0];
        dcp->lookdims[1] = (int)array[1];
        dcp->lookdims[2] = (int)array[2];
        free(array);
    }
    for (int i = 0; i < dcp->ill_count; i++) {
        dcp->ill[i] = exif_lightsource_aton(ill[i]->valuestring);
        if (dcp->ill[i] < 0) {
            fprintf(stderr, "Unknown CalibrationIlluminant%d \"%s\".\n", i+1, ill[i]->valuestring);
            goto fail;
        }
        if (!matrix3x3_parse(&dcp->cm[i], cm[i])) {
            fprintf(stderr, "Failed to parse ColorMatrix%d.\n", i+1);
            goto fail;
        }
        if (fm[i] != NULL) {
            dcp->has.fm = true;
            if (!matrix3x3_parse(&dcp->fm[i], fm[i])) {
                fprintf(stderr, "Failed to parse ForwardMatrix%d.\n", i+1);
                goto fail;
            }
        }
        if (hsm[i] != NULL) {
            if ((dcp->huesatmap[i] = parse_dcp_huesatmap(dcp->hsmdims, hsm[i])) == NULL) {
                fprintf(stderr, "Failed to parse ProfileHueSatMap%d.\n", i+1);
                goto fail;
            }
        }
    }
    if (look != NULL) {
        if ((dcp->looktable = parse_dcp_huesatmap(dcp->lookdims, look)) == NULL) {
            fprintf(stderr, "Failed to parse ProfileLookTable.\n");
            goto fail;
        }
    }
    if (tonecurve) {
        dcp->tonecurve_len = 0;
        for (cJSON *ji = tonecurve->child; ji != NULL; ji = ji->next) {
            if (ji->type != cJSON_Array) {
                fprintf(stderr, "Expected array in array, got other type.\n");
                goto fail;
            }
            dcp->tonecurve_len++;
        }
        if (dcp->tonecurve_len < 2) {
            fprintf(stderr, "Too short ProfileToneCurve.\n");
            goto fail;
        }
        dcp->tonecurve = malloc(2 * dcp->tonecurve_len * sizeof(dcp->tonecurve[0]));
        int count = 0;
        for (cJSON *ji = tonecurve->child; ji != NULL; ji = ji->next) {
            int i = 0;
            for (cJSON *ji1 = ji->child; ji1 != NULL; ji1 = ji1->next) {
                if (ji1->type != cJSON_Number) {
                    fprintf(stderr, "Expected Number in inner array, got other type.\n");
                    goto fail;
                }
                if (i > 1) {
                    fprintf(stderr, "Expected length 2 of array element in ProfileToneCurve.\n");
                    goto fail;
                }
                dcp->tonecurve[2*count+i] = ji1->valuedouble;
                i++;
            }
            count++;
        }
        if (dcp->tonecurve[0+0] != 0 || dcp->tonecurve[0+1] != 0) {
            fprintf(stderr, "Expected [0,0] as first element in ProfileToneCurve.\n");
            goto fail;
        }
        if (dcp->tonecurve[dcp->tonecurve_len-1+0] != 1.0 || dcp->tonecurve[dcp->tonecurve_len-1+1] != 1.0) {
            fprintf(stderr, "Expected [1,1] as last element in ProfileToneCurve.\n");
            goto fail;
        }
    }
    if (look_enc) {
        dcp->has.looktable_srgb_gamma = true;
        if (strcasecmp(look_enc->valuestring, "sRGB") == 0) {
            dcp->looktable_srgb_gamma = true;
        } else if (strcasecmp(look_enc->valuestring, "Linear") == 0) {
            dcp->looktable_srgb_gamma = false;
        } else {
            fprintf(stderr, "Unknown ProfileLookTableEncoding \"%s\".\n", look_enc->valuestring);
            goto fail;
        }
    }
    if (hsm_enc) {
        dcp->has.huesatmap_srgb_gamma = true;
        if (strcasecmp(hsm_enc->valuestring, "sRGB") == 0) {
            dcp->huesatmap_srgb_gamma = true;
        } else if (strcasecmp(hsm_enc->valuestring, "Linear") == 0) {
            dcp->huesatmap_srgb_gamma = false;
        } else {
            fprintf(stderr, "Unknown ProfileHueSatMapEncoding \"%s\".\n", hsm_enc->valuestring);
            goto fail;
        }
    }
    if (baseline_exp) {
        dcp->has.baseline_exposure_offset = true;
        dcp->baseline_exposure_offset = baseline_exp->valuedouble;
    }
    if (black_render) {
        dcp->has.black_render_none = true;
        if (strcasecmp(black_render->valuestring, "None") == 0) {
            dcp->black_render_none = true;
        } else if (strcasecmp(black_render->valuestring, "Auto") == 0) {
            dcp->black_render_none = false;
        } else {
            fprintf(stderr, "Unknown DefaultBlackRender \"%s\".\n", black_render->valuestring);
            goto fail;
        }
    }
    if (embed_policy) {
        dcp->has.embed_policy = true;
        if (strcasecmp(embed_policy->valuestring, "Allow copying") == 0 ||
            strcasecmp(embed_policy->valuestring, "AllowCopying") == 0)
        {
            dcp->embed_policy = pepAllowCopying;
        } else if (strcasecmp(embed_policy->valuestring, "Embed if used") == 0 ||
                   strcasecmp(embed_policy->valuestring, "EmbedIfUsed") == 0)
        {
            dcp->embed_policy = pepEmbedIfUsed;
        } else if (strcasecmp(embed_policy->valuestring, "Embed never") == 0 ||
                   strcasecmp(embed_policy->valuestring, "EmbedNever") == 0)
        {
            dcp->embed_policy = pepEmbedNever;
        } else if (strcasecmp(embed_policy->valuestring, "No restrictions") == 0 ||
                   strcasecmp(embed_policy->valuestring, "NoRestrictions") == 0)
        {
            dcp->embed_policy = pepNoRestrictions;
        } else {
            fprintf(stderr, "Unknown ProfileEmbedPolicy \"%s\".\n", embed_policy->valuestring);
            goto fail;
        }
    }

    cJSON_Delete(jsroot);
    return dcp;

fail:
    fprintf(stderr, "Could not parse DCP file \"%s\".\n", filename);
    cJSON_Delete(jsroot);
    profio_dcp_delete(dcp);
    if (parse_error_is_fatal) exit(EXIT_FAILURE);
    return NULL;
}

struct dcam_profile *
jsonio_profile_parse(const char filename[],
                     bool parse_error_is_fatal)
{
    struct dcam_profile *prof = NULL;
    cJSON *jsroot = json_parse(filename);
    if (jsroot == NULL) {
        if (parse_error_is_fatal) exit(EXIT_FAILURE);
        return NULL;
    }
    cJSON *cm, *fm, *ill, *lut, *ill_spec;

    if (!get_and_test(&ill, jsroot, "CalibrationIlluminant1", cJSON_String, true)) goto fail;
    if (!get_and_test(&ill_spec, jsroot, "CalibrationIlluminantSpectrum1", cJSON_Object, false)) goto fail;
    if (!get_and_test(&cm, jsroot, "ColorMatrix1", cJSON_Array, true)) goto fail;
    if (!get_and_test(&fm, jsroot, "ForwardMatrix1", cJSON_Array, true)) goto fail;
    if (!get_and_test(&lut, jsroot, "ChromaLUT1", cJSON_Object, false)) goto fail;

    prof = calloc(1, sizeof(*prof));
    if (!matrix3x3_parse(&prof->color_matrix, cm)) {
        fprintf(stderr, "Failed to parse ColorMatrix1.\n");
        goto fail;
    }
    if (!matrix3x3_parse(&prof->forward_matrix, fm)) {
        fprintf(stderr, "Failed to parse ForwardMatrix1.\n");
        goto fail;
    }
    prof->illuminant_name = exif_lightsource_aton(ill->valuestring);
    if ((int)prof->illuminant_name < 0) {
        fprintf(stderr, "Unknown CalibrationIlluminant1 \"%s\".\n", ill->valuestring);
        goto fail;
    }
    if (ill_spec != NULL) {
        cJSON *jbands, *jspec;
        if (!get_and_test(&jbands, ill_spec, "bands", cJSON_Array, true)) goto fail;
        if (!get_and_test(&jspec, ill_spec, "spectrum", cJSON_Array, true)) goto fail;
        int band_count;
        double *band_f = json_spectrum_bands_array(jbands, &band_count);
        if (band_f == NULL) {
            goto fail;
        }
        int amp_count;
        double *amp = json_numberarray(jspec, &amp_count);
        if (amp == NULL || amp_count != band_count) {
            free(amp);
            free(band_f);
            goto fail;
        }
        prof->illuminant_spec = spec_alloc1(band_f, amp, band_count);
        free(band_f);
        free(amp);
        if (prof->illuminant_spec == NULL) {
            goto fail;
        }
    }

    if (lut != NULL) {
        cJSON *reg;
        cJSON *uv[3];
        if (!get_and_test(&reg, lut, "regularization", cJSON_Array, true)) goto fail;
        if (!get_and_test(&uv[0], lut, "uvU", cJSON_Array, true)) goto fail;
        if (!get_and_test(&uv[1], lut, "uvV", cJSON_Array, true)) goto fail;
        if (!get_and_test(&uv[2], lut, "uvL", cJSON_Array, true)) goto fail;

        double regularization[3];
        {
            int count;
            double *relax = json_numberarray(reg, &count);
            if (relax == NULL || count != 3) {
                fprintf(stderr, "Unexpected regularization content in LUT.\n");
                free(relax);
                goto fail;
            }
            regularization[0] = relax[0];
            regularization[1] = relax[1];
            regularization[2] = relax[2];
            free(relax);
        }
        int count[3];
        double *da[3] = { NULL, NULL, NULL };
        for (int i = 0; i < 3; i++) {
            da[i] = json_numberarray(uv[i], &count[i]);
            if (da[i] == NULL) {
                free(da[0]);free(da[1]);free(da[2]);
                goto fail;
            }
        }
        for (int i = 1; i < 3; i++) {
            if (count[i] != count[0] || count[i] % 3 != 0) {
                fprintf(stderr, "Unexpected array length in LUT.\n");
                free(da[0]);free(da[1]);free(da[2]);
                goto fail;
            }
        }
        int cp_count = count[0]/3;
        v3 *cp[3];
        for (int i = 0; i < 3; i++) {
            cp[i] = malloc(cp_count * sizeof(*cp[i]));
            for (int j = 0; j < cp_count; j++) {
                cp[i][j].v[0] = da[i][j*3+0];
                cp[i][j].v[1] = da[i][j*3+2]; // yes it is this order, 012 vs 021
                cp[i][j].v[2] = da[i][j*3+1];
            }
            free(da[i]);
        }
        prof->lut = chromalut_new_from_data(cp, cp_count, regularization);
        for (int i = 0; i < 3; i++) {
            free(cp[i]);
        }
        if (prof->lut == NULL) {
            fprintf(stderr, "Invalid LUT data.\n");
            goto fail;
        }
    }

    cJSON_Delete(jsroot);
    return prof;

fail:
    fprintf(stderr, "Could not parse camera profile in file \"%s\".\n", filename);
    cJSON_Delete(jsroot);
    if (prof != NULL) {
        free(prof->illuminant_spec);
        chromalut_delete(prof->lut);
        free(prof);
    }
    if (parse_error_is_fatal) exit(EXIT_FAILURE);
    return NULL;
}
