/*
 * (c) Copyright 2015 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdio.h>
#include <assert.h>
#include <sys/time.h>

#include <lcms2.h>

#include <dcamprof.h>
#include <lut.h>
#include <tps.h>
#include <nmsimplex.h>

struct chromalut_t_ {
    tps_t *tps[3];
    double regularization[3];
    v3 *cp[3];
    int cp_count;
};

v3
xyz2lutspace(v3 xyz)
{
    double x = xyz.v[0], y = xyz.v[1], z = xyz.v[2];
    // u' v' and L*
    double up = 4*x / (x + 15*y + 3*z);
    double vp = 9*y / (x + 15*y + 3*z);
    double L = 116*lab_ft_forward(y) - 16;

    return (v3){{ L*0.01, up, vp }};
}

v3
lutspace2xyz(v3 lutspace)
{
    double L = lutspace.v[0]*100.0, up = lutspace.v[1], vp = lutspace.v[2];
    double y = (L + 16)/116;
    y = lab_ft_inverse(y);
    double x = y*9*up / (4*vp);
    double z = y * (12 - 3*up - 20*vp) / (4*vp);

    return (v3){{ x, y, z }};
}

static v3
chromalut_lookup_(tps_t *tps[3],
                  v3 xyz)
{
    v3 lutspace = xyz2lutspace(xyz);
    v3 lso = lutspace;
    lso.v[0] *= 1.0 + 10.0 * tps_interpolate(tps[0], lutspace.v[1], lutspace.v[2]);
    lso.v[1] += tps_interpolate(tps[1], lutspace.v[1], lutspace.v[2]);
    lso.v[2] += tps_interpolate(tps[2], lutspace.v[1], lutspace.v[2]);
    return lutspace2xyz(lso);
}

static double
worst_stretch(const m3x3 cam2xyz,
              const struct patch_set *ps,
              v3 **cp,
              int cp_count,
              int j)
{
    double max_stretch = 0;
    v3 ltu;
    ltu = xyz2lutspace(m3x3_mul_v3(cam2xyz, ps->patch[j].cam));
    double c1m[3] = { 0, ltu.v[1], ltu.v[2] };
    ltu.v[0]  = cp[0][j].v[1];
    ltu.v[1] += cp[1][j].v[1];
    ltu.v[2] += cp[2][j].v[1];
    double c1L[3] = { ltu.v[0], ltu.v[1], ltu.v[2] };
    for (int i = 0; i < cp_count; i++) {
        if (i == j) continue;

        ltu = xyz2lutspace(m3x3_mul_v3(cam2xyz, ps->patch[i].cam));
        double c2m[3] = { 0, ltu.v[1], ltu.v[2] };
        ltu.v[0]  = cp[0][i].v[1];
        ltu.v[1] += cp[1][i].v[1];
        ltu.v[2] += cp[2][i].v[1];
        double c2L[3] = { ltu.v[0], ltu.v[1], ltu.v[2] };

        double cdm[3] = { c2m[0]-c1m[0], c2m[1]-c1m[1], c2m[2]-c1m[2] };
        double cdL[3] = { c2L[0]-c1L[0], c2L[1]-c1L[1], c2L[2]-c1L[2] };
        double d1m = sqrt(cdm[0]*cdm[0] + cdm[1]*cdm[1] + cdm[2]*cdm[2]);
        double d1L = sqrt(cdL[0]*cdL[0] + cdL[1]*cdL[1] + cdL[2]*cdL[2]);
        double stretch;
        if (d1L < d1m) {
            stretch = d1m / d1L;
        } else {
            stretch = d1L / d1m;
        }
        if (stretch > max_stretch) {
            max_stretch = stretch;
        }
    }
    return max_stretch;
}

chromalut_t *
chromalut_new(const m3x3 cam2xyz,
              const struct patch_set *ps_,
              const double min_chroma_distance,
              const double regularization[3])
{
    const v3 d50 = (v3){{ cmsD50X, cmsD50Y, cmsD50Z }};
    struct patch_set *ps = normalize_patch_set(ps_);

    fprintf(stderr, "Making 2.5D chromaticity-addressed lookup table for XYZ correction...\n");

    if (min_chroma_distance > 0.0) {
        struct patch_set *ps1 = malloc(sizeof(*ps) + ps->patch_count * sizeof(ps->patch[0]));
        *ps1 = *ps;
        ps1->patch_count = 0;
        int *cgrp = calloc(1, sizeof(*cgrp) * ps->patch_count);
        bool *used = calloc(1, sizeof(*used) * ps->patch_count);
        int grouped = 0;
        for (int i = 0; i < ps->patch_count; i++) {
            if (used[i]) continue;
            ps1->patch[ps1->patch_count++] = ps->patch[i];
            used[i] = true;
            const int base_i = ps1->patch_count - 1;
            cgrp[base_i] = 1;
            for (int j = i+1; j < ps->patch_count; j++) {
                if (used[j]) continue;
                v3 ltu1, ltu2;
                ltu1 = xyz2lutspace(m3x3_mul_v3(cam2xyz, ps->patch[i].cam));
                ltu2 = xyz2lutspace(m3x3_mul_v3(cam2xyz, ps->patch[j].cam));
                ltu1.v[0] = ltu2.v[0];
                double uv_de = simple_de(ltu1, ltu2);
                if (uv_de < min_chroma_distance) {
                    cgrp[base_i]++;
                    ps1->patch[ps1->patch_count++] = ps->patch[j];
                    used[j] = true;
                    grouped++;
                }
            }
        }
        assert(ps1->patch_count == ps->patch_count);
        free(used);
        if (grouped > 0) {
            ps->patch_count = 0;
            int largest_grp = 0;
            for (int i = 0; i < ps1->patch_count;  i += cgrp[i]) {
                if (cgrp[i] > 1) {
                    if (cgrp[i] > largest_grp) largest_grp = cgrp[i];
                    int cc[TARGET_MAX_CLASS_COUNT];
                    memset(cc, 0, sizeof(cc));
                    struct cam2xyz_patch patch;
                    memset(&patch, 0, sizeof(patch));
                    for (int j = i; j < i + cgrp[i]; j++) {
                        for (int k = 0; k < 3; k++) {
                            patch.cam.v[k] += ps1->patch[j].cam.v[k];
                            patch.xyz.v[k] += ps1->patch[j].xyz.v[k];
                        }
                        patch.de += ps1->patch[j].de;
                        cc[ps1->patch[j].class]++;
                    }
                    for (int k = 0; k < 3; k++) {
                        patch.cam.v[k] /= cgrp[i];
                        patch.xyz.v[k] /= cgrp[i];
                    }
                    patch.de /= cgrp[i];
                    int max_class = 0;
                    for (int k = 0; k < TARGET_MAX_CLASS_COUNT; k++) {
                        if (cc[k] > cc[max_class]) {
                            max_class = k;
                        }
                    }
                    for (int j = i; j < i + cgrp[i]; j++) {
                        if (ps1->patch[j].class == cc[max_class]) {
                            ps->patch[ps->patch_count] = ps1->patch[j];
                            break;
                        }
                    }
                    ps->patch[ps->patch_count].cam = patch.cam;
                    ps->patch[ps->patch_count].xyz = patch.xyz;
                    ps->patch[ps->patch_count].de = patch.de;
                    ps->patch_count++;
                } else {
                    ps->patch[ps->patch_count] = ps1->patch[i];
                    ps->patch_count++;
                }
            }
            fprintf(stderr,
                    //2345678901234567890123456789012345678901234567890123456789012345678901234567890
                    "%.2f%% of the patches was put in a chromaticity group due to nearby neighbor.\n"
                    "  Largest chromaticity group contains %d patches. Patch count reduced from\n"
                    "  %zd to %zd. Note that patch matching cannot reach 100%% when chromaticity\n"
                    "  groups are formed, as the LUT matches the average within a group.\n",
                    100.0 * (double)grouped / (double)ps1->patch_count, largest_grp, ps1->patch_count, ps->patch_count);
        }
        free(ps1);
        free(cgrp);
    }

    v3 *cp[3];
    for (int i = 0; i < 3; i++) {
        cp[i] = malloc(ps->patch_count * sizeof(v3));
    }

    /*
    double csum = 0, rsum = 0;
    for (int i = 0; i < ps->patch_count; i++) {
        v3 cam = m3x3_mul_v3(cam2xyz, ps->patch[i].cam);
        v3 ref = ps->patch[i].xyz;
        cam = xyz2lutspace(cam);
        ref = xyz2lutspace(ref);
        csum += cam.v[0];
        rsum += ref.v[0];
    }
    double f = rsum / csum;
    */
    v3 *ref_a = malloc(ps->patch_count * sizeof(v3));
    for (int i = 0; i < ps->patch_count; i++) {
        v3 cam = m3x3_mul_v3(cam2xyz, ps->patch[i].cam);
        v3 ref = ps->patch[i].xyz;
        cam = xyz2lutspace(cam);
        ref = xyz2lutspace(ref);
        ref_a[i] = ref;
        cp[0][i] = cp[1][i] = cp[2][i] = (v3){{ cam.v[1], 0, cam.v[2] }};
        //cp[0][i].v[1] = ref.v[0] - f * cam.v[0];
        cp[0][i].v[1] = (ref.v[0] / cam.v[0] - 1.0) * 0.1;
        cp[1][i].v[1] = ref.v[1] - cam.v[1];
        cp[2][i].v[1] = ref.v[2] - cam.v[2];
        //fprintf(stderr, "cam %f %f %f\n", cam.v[0],cam.v[1],cam.v[2]);
        //fprintf(stderr, "ref %f %f %f\n", ref.v[0],ref.v[1],ref.v[2]);
    }

    double relax[3] = { 0.0, 0.0, 0.0 };
    if (regularization != NULL) {
        relax[0] = regularization[0];
        relax[1] = regularization[1];
        relax[2] = regularization[2];
    }
    { // workaround: test if we need to add regularization
        tps_t *tps_exact[3];
        for (int i = 0; i < 3; i++) {
            tps_exact[i] = tps_new(cp[i], ps->patch_count, relax[i]);
            if (tps_exact[i] == NULL) {
                // tps seems to sometimes have some issue with 0 regularization
                relax[i] = 0.000000000001;
                tps_exact[i] = tps_new(cp[i], ps->patch_count, relax[i]);
            }
            if (tps_exact[i] == NULL) {
                relax[i] = 0.0;
                do {
                    relax[i] += 0.0005;
                    tps_exact[i] = tps_new(cp[i], ps->patch_count, relax[i]);
                } while (tps_exact[i] == NULL);
                fprintf(stderr, "Exact TPS solution for axis %d was not possible, relaxed with %f.\n", i, relax[i]);
            }
        }
        for (int i = 0; i < 3; i++) {
            tps_delete(tps_exact[i]);
        }
    }

    double max_de_lim = 0.0;
    for (int j = 0; j < ps->patch_count; j++) {
        if (ps->patch[j].de > max_de_lim) {
            max_de_lim = ps->patch[j].de;
        }
    }
    if (max_de_lim > 0.0) { // relax LUT stretching according to provide max acceptable DE values

        v3 *cpa[3];
        for (int i = 0; i < 3; i++) {
            cpa[i] = malloc(ps->patch_count * sizeof(v3));
            memcpy(cpa[i], cp[i], ps->patch_count * sizeof(v3));
        }

        double old_de[ps->patch_count];
        for (int j = 0; j < ps->patch_count; j++) {
            old_de[j] = -1;
        }
        fprintf(stderr, "Relaxing LUT stretch with up to %.2f DE. Iterating over %zd patches%s...\n", max_de_lim, ps->patch_count,
                (ps->patch_count > 100) ? ", this can take very long time" : "");
        struct timeval tv, tvnext;
        gettimeofday(&tv, NULL);
        tvnext = tv;
        tvnext.tv_sec += 3;
        for (;;) {
            tps_t *tps_approx[3];
            for (int i = 0; i < 3; i++) {
                tps_approx[i] = tps_new(cpa[i], ps->patch_count, relax[i]);
            }

            double difference = 0;
            double de_sum = 0;
            for (int j = 0; j < ps->patch_count; j++) {
                v3 cam = m3x3_mul_v3(cam2xyz, ps->patch[j].cam);
                v3 ref = ps->patch[j].xyz;
                v3 approx = chromalut_lookup_(tps_approx, cam);
                double de = ciede2000(approx, ref, d50);
                de_sum += de;
                difference += fabs(de - old_de[j]);
                old_de[j] = de;
            }
            if (difference == 0.0) {
                for (int i = 0; i < 3; i++) {
                    memcpy(cp[i], cpa[i], ps->patch_count * sizeof(v3));
                    free(cpa[i]);
                    tps_delete(tps_approx[i]);
                }
                int zero_count = 0;
                for (int j = 0; j < ps->patch_count; j++) {
                    if (cp[0][j].v[1] == 0 && cp[1][j].v[1] == 0 && cp[2][j].v[1] == 0) {
                        zero_count++;
                    }
                }
                fprintf(stderr,
                        //2345678901234567890123456789012345678901234567890123456789012345678901234567890
                        "Average DE for the %zd tested patches increased to %.2f after LUT relax.\n"
                        "  %.2f%% could do without LUT correction.\n",
                        ps->patch_count, de_sum / ps->patch_count, (100.0 * zero_count) / ps->patch_count);
                break;
            }
            gettimeofday(&tv, NULL);
            if (tv.tv_sec >= tvnext.tv_sec) {
                fprintf(stderr, ".");
                tvnext = tv;
                tvnext.tv_sec += 1;
            }
            for (int j = 0; j < ps->patch_count; j++) {
                v3 cam = m3x3_mul_v3(cam2xyz, ps->patch[j].cam);
                v3 ref = ps->patch[j].xyz;
                v3 apx = chromalut_lookup_(tps_approx, cam);
                double de = ciede2000(apx, ref, d50);
                apx = xyz2lutspace(apx);
                cam = xyz2lutspace(cam);
                ref = xyz2lutspace(ref);
                for (int i = 0; i < 3; i++) {
                    double pre_relax[3] = { cpa[0][j].v[1], cpa[1][j].v[1], cpa[2][j].v[1] };
                    double stretch_before = worst_stretch(cam2xyz, ps, cpa, ps->patch_count, j);
                    if (de > ps->patch[j].de) {
                        if (fabs(cpa[i][j].v[1] - cp[i][j].v[1]) <= 0.00005) {
                            cpa[i][j].v[1] = cp[i][j].v[1]; // full correction
                        } else if (cpa[i][j].v[1] < 0) {
                            cpa[i][j].v[1] -= 0.00005;
                        } else {
                            cpa[i][j].v[1] += 0.00005;
                        }
                    } else if (de < ps->patch[j].de - 0.1) {
                        if (fabs(cpa[i][j].v[1]) <= 0.00005) {
                            cpa[i][j].v[1] = 0.0; // no correction
                        } else if (cpa[i][j].v[1] < 0) {
                            cpa[i][j].v[1] += 0.00005;
                        } else {
                            cpa[i][j].v[1] -= 0.00005;
                        }
                    }
                    double stretch_after = worst_stretch(cam2xyz, ps, cpa, ps->patch_count, j);
                    if (stretch_after > stretch_before) {
                        //fprintf(stderr, "w");
                        cpa[0][j].v[1] = pre_relax[0];
                        cpa[1][j].v[1] = pre_relax[1];
                        cpa[2][j].v[1] = pre_relax[2];
                    }
                }
            }
            /*
            if (0){ // test approx
                for (int j = 0; j < ps->patch_count; j++) {
                    v3 cam = m3x3_mul_v3(cam2xyz, ps->patch[j].cam);
                    v3 ref = ps->patch[j].xyz;
                    v3 approx = chromalut_lookup_(tps_approx, cam);
                    double de = ciede2000(approx, ref, d50);
                    fprintf(stderr, "%d DE %.2f (%.2f)\n", j, de, ps->patch[j].de);
                }
                getchar();
            }
            */
            for (int i = 0; i < 3; i++) {
                tps_delete(tps_approx[i]);
            }
        }

        /*
        for (;;) {
            for (int i = 0; i < 3; i++) {
                tps_approx[i] = tps_new(cpa[i], ps->patch_count, relax[i]);
            }
            bool complete = true;

            { // test approx
                for (int j = 0; j < ps->patch_count; j++) {
                    v3 cam = m3x3_mul_v3(cam2xyz, ps->patch[j].cam);
                    v3 ref = ps->patch[j].xyz;
                    v3 approx = chromalut_lookup_(tps_approx, cam);
                    double de = ciede2000(approx, ref, d50);
                    //fprintf(stderr, "%d DE %f %d\n", j, de, rdir[j]);
                }
                //getchar();
            }

            {
                complete = true;
                for (int j = 0; j < ps->patch_count; j++) {
                    v3 cam = m3x3_mul_v3(cam2xyz, ps->patch[j].cam);
                    v3 ref = ps->patch[j].xyz;
                    v3 apx = chromalut_lookup_(tps_approx, cam);
                    double de = ciede2000(apx, ref, d50);
                    apx = xyz2lutspace(apx);
                    cam = xyz2lutspace(cam);
                    ref = xyz2lutspace(ref);
                    if (de > ps->patch[j].de && rdir[j] != +2) {
                        for (int i = 0; i < 3; i++) {
                            if (cpa[i][j].v[1] < 0) {
                                cpa[i][j].v[1] -= 0.00005;
                            } else {
                                cpa[i][j].v[1] += 0.00005;
                            }
                        }
                        rdir[j] = -1;
                        complete = false;
                    } else if (de < ps->patch[j].de - 0.1 && rdir[j] != +2) {
                        for (int i = 0; i < 3; i++) {
                            if (fabs(cpa[i][j].v[1]) <= 0.00005) {
                                cpa[i][j].v[1] = 0.0;
                            } else if (cpa[i][j].v[1] < 0) {
                                cpa[i][j].v[1] += 0.00005;
                            } else {
                                cpa[i][j].v[1] -= 0.00005;
                            }
                        }
                        bool finished = true;
                        for (int i = 0; i < 3; i++) {
                            if (cpa[i][j].v[1] != 0.0) {
                                finished = false;
                            }
                        }
                        if (!finished) {
                            rdir[j] = +1;
                            complete = false;
                        } else {
                            rdir[j] = +2;
                        }
                    } else {
                        rdir[j] = +2;
                    }
                }
            }

            static int cc = 0;
            if (++cc == 1000) break;
            if (complete) {
                break;
            }
            for (int i = 0; i < 3; i++) {
                tps_delete(tps_approx[i]);
            }
        }
        */
    }

    chromalut_t *lut = calloc(1, sizeof(*lut));
    for (int i = 0; i < 3; i++) {
        lut->regularization[i] = relax[i];
        lut->tps[i] = tps_new(cp[i], ps->patch_count, relax[i]);
        if (lut->tps[i] == NULL) {
            fprintf(stderr, "Failed to generate TPS! Aborting\n");
            abort();
        }
        lut->cp[i] = cp[i];
    }
    lut->cp_count = ps->patch_count;

    if (0) {
        for (int j = 0; j < ps->patch_count; j++) {
            v3 cam = m3x3_mul_v3(cam2xyz, ps->patch[j].cam);
            v3 ref = ps->patch[j].xyz;
            v3 apx = chromalut_lookup_(lut->tps, cam);
            double de = ciede2000(apx, ref, d50);
            fprintf(stderr, "%d DE %.2f (%.2f)\n", j, de, ps->patch[j].de);
        }
    }

    return lut;
}

void
chromalut_delete(chromalut_t *lut)
{
    if (lut == NULL) {
        return;
    }
    for (int i = 0; i < 3; i++) {
        tps_delete(lut->tps[i]);
        free(lut->cp[i]);
    }
    free(lut);
}

v3
chromalut_lookup_diff(chromalut_t *lut,
                      double u,
                      double v)
{
    double Ldiff = tps_interpolate(lut->tps[0], u, v);
    double udiff = tps_interpolate(lut->tps[1], u, v);
    double vdiff = tps_interpolate(lut->tps[2], u, v);
    return (v3){{ Ldiff, udiff, vdiff }};
}

v3
chromalut_lookup(chromalut_t *lut,
                 v3 xyz)
{
    if (lut == NULL) {
        return xyz;
    }
    return chromalut_lookup_(lut->tps, xyz);
}

struct chromalut_reverse_lookup_simplex_fun_arg {
    chromalut_t *lut;
    v3 xyz;
};

static double
chromalut_reverse_lookup_simplex_fun(double m[],
                                     void *arg)
{
    struct chromalut_reverse_lookup_simplex_fun_arg *a = (struct chromalut_reverse_lookup_simplex_fun_arg *)arg;
    v3 xyz = {{ m[0], m[1], m[2] }};
    xyz = chromalut_lookup(a->lut, xyz);
    return simple_de(xyz, a->xyz);
}

v3
chromalut_reverse_lookup(chromalut_t *lut,
                         v3 xyz)
{
    double m[3] = { xyz.v[0], xyz.v[1], xyz.v[2] };
    if (m[0] < 0 || m[1] < 0 || m[2] < 0) {
        return (v3){{ -1, -1, -1 }};
    }

    struct chromalut_reverse_lookup_simplex_fun_arg arg;
    arg.lut = lut;
    arg.xyz = xyz;
    double min = simplex(chromalut_reverse_lookup_simplex_fun, &arg, m, 3, 1.0e-7, 0.01, NULL);
    if (min > 0.0001) {
        //fprintf(stderr, "%f %f %f %f\n", min, m[0], m[1], m[2]);
        return (v3){{ -1, -1, -1 }};
    }
    return (v3){{ m[0], m[1], m[2] }};
}

static void
print_cp_array(FILE *stream,
               const v3 *cp,
               int cp_count)
{
    fprintf(stream, "[ ");
    for (int i = 0; i < cp_count; i++) {
        fprintf(stream, "%g, %g, %g%s", cp[i].v[0], cp[i].v[2], cp[i].v[1], i < cp_count - 1 ? ", " : "");
    }
    fprintf(stream, " ]");
}

void
chromalut_json_print(FILE *stream,
                     const char indent[],
                     const char name[],
                     chromalut_t *lut)
{
    fprintf(stream, "%s\"%s\": {\n", indent, name);
    fprintf(stream, "%s  \"regularization\": [ %g, %g, %g ],\n", indent, lut->regularization[0], lut->regularization[1], lut->regularization[2]);
    fprintf(stream, "%s  \"uvU\": ", indent);
    print_cp_array(stream, lut->cp[0], lut->cp_count);
    fprintf(stream, ",\n%s  \"uvV\": ", indent);
    print_cp_array(stream, lut->cp[1], lut->cp_count);
    fprintf(stream, ",\n%s  \"uvL\": ", indent);
    print_cp_array(stream, lut->cp[2], lut->cp_count);
    fprintf(stream, "\n%s}", indent);
}

chromalut_t *
chromalut_new_from_data(v3 *cp[3],
                        int cp_count,
                        double regularization[3])
{
    chromalut_t *lut = calloc(1, sizeof(*lut));
    for (int i = 0; i < 3; i++) {
        lut->regularization[i] = regularization[i];
        lut->tps[i] = tps_new(cp[i], cp_count, lut->regularization[i]);
        if (lut->tps[i] == NULL) {
            fprintf(stderr, "Failed to generate TPS! Aborting\n");
            abort();
        }
        lut->cp[i] = malloc(cp_count * sizeof(v3));
        memcpy(lut->cp[i], cp[i], cp_count * sizeof(v3));
    }
    return lut;
}
