/*
 * (c) Copyright 2015 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <inttypes.h>
#include <inttypes.h>

#include <profio.h>

#define debug(...)
//#define debug(...) fprintf(stderr, __VA_ARGS__)

#define bit_swap16(value)                                               \
    (((((uint16_t)(value)) & 0xFF00) >> 8) |                            \
     ((((uint16_t)(value)) & 0x00FF) << 8))
#define bit_swap32(value)                                               \
    (((((uint32_t)(value)) & 0xFF000000) >> 24) |                       \
     ((((uint32_t)(value)) & 0x00FF0000) >>  8) |                       \
     ((((uint32_t)(value)) & 0x0000FF00) <<  8) |                       \
     ((((uint32_t)(value)) & 0x000000FF) << 24))

#define GET4(data) do {                         \
        if (fread(data, 1, 4, stream) != 4) {   \
            goto fail;                          \
        }                                       \
        if (do_swap) {                          \
            *data = bit_swap32(*(data));        \
        }                                       \
    } while (0)
#define GET2(data) do {                         \
        if (fread(data, 1, 2, stream) != 2) {   \
            goto fail;                          \
        }                                       \
        if (do_swap) {                          \
            *data = bit_swap16(*(data));        \
        }                                       \
    } while (0)

typedef	enum {
	TIFF_NOTYPE	= 0,	/* placeholder */
	TIFF_BYTE	= 1,	/* 8-bit unsigned integer */
	TIFF_ASCII	= 2,	/* 8-bit bytes w/ last byte null */
	TIFF_SHORT	= 3,	/* 16-bit unsigned integer */
	TIFF_LONG	= 4,	/* 32-bit unsigned integer */
	TIFF_RATIONAL	= 5,	/* 64-bit unsigned fraction */
	TIFF_SBYTE	= 6,	/* !8-bit signed integer */
	TIFF_UNDEFINED	= 7,	/* !8-bit untyped data */
	TIFF_SSHORT	= 8,	/* !16-bit signed integer */
	TIFF_SLONG	= 9,	/* !32-bit signed integer */
	TIFF_SRATIONAL	= 10,	/* !64-bit signed fraction */
	TIFF_FLOAT	= 11,	/* !32-bit IEEE floating point */
	TIFF_DOUBLE	= 12,	/* !64-bit IEEE floating point */
	TIFF_IFD	= 13	/* %32-bit unsigned integer (offset) */
} TIFFDataType;

#define TIFFTAG_UNIQUECAMERAMODEL 50708
#define TIFFTAG_COLORMATRIX1      50721
#define TIFFTAG_COLORMATRIX2      50722
#define TIFFTAG_REDUCTIONMATRIX1  50725 //
#define TIFFTAG_REDUCTIONMATRIX2  50726 //
#define TIFFTAG_CALIBRATIONILLUMINANT1 50778
#define TIFFTAG_CALIBRATIONILLUMINANT2 50779
#define TIFFTAG_PROFILECALIBRATIONSIGNATURE 50932
#define TIFFTAG_PROFILENAME 50936
#define TIFFTAG_PROFILEHUESATMAPDIMS 50937
#define TIFFTAG_PROFILEHUESATMAPDATA1 50938
#define TIFFTAG_PROFILEHUESATMAPDATA2 50939
#define TIFFTAG_PROFILETONECURVE 50940
#define TIFFTAG_PROFILEEMBEDPOLICY 50941
#define TIFFTAG_PROFILECOPYRIGHT 50942
#define TIFFTAG_FORWARDMATRIX1 50964
#define TIFFTAG_FORWARDMATRIX2 50965
#define TIFFTAG_PROFILELOOKTABLEDIMS 50981
#define TIFFTAG_PROFILELOOKTABLEDATA 50982
#define TIFFTAG_PROFILEHUESATMAPENCODING 51107 // long 1
#define TIFFTAG_PROFILELOOKTABLEENCODING 51108
#define TIFFTAG_BASELINEEXPOSUREOFFSET 51109 // rational 1
#define TIFFTAG_DEFAULTBLACKRENDER 51110 // long 1


struct tiff_tag {
    uint16_t id;
    uint16_t type;
    uint32_t count;
    uint32_t data;
    bool parsed_subdir;
    uint32_t ext_data_size;
    uint8_t *ext_data;
};

static int
tag_value_size(uint16_t type)
{
    int vsize = 1;
    switch (type) {
    case TIFF_UNDEFINED:
    case TIFF_SBYTE:
    case TIFF_BYTE:
    case TIFF_ASCII:
        vsize = 1;
        break;
    case TIFF_SHORT:
    case TIFF_SSHORT:
        vsize = 2;
        break;
    case TIFF_LONG:
    case TIFF_SLONG:
    case TIFF_FLOAT:
        vsize = 4;
        break;
    case TIFF_RATIONAL:
    case TIFF_SRATIONAL:
    case TIFF_DOUBLE:
        vsize = 8;
        break;
    case TIFF_IFD:
        vsize = 4;
        break;
    }
    return vsize;
}

struct profio_dcp *
profio_dcp_parse(const char filename[],
                 bool error_is_fatal)
{
    // TIFF directory format
    uint8_t *tag_ext_data = NULL;
    struct profio_dcp *dcp = NULL;
    FILE *stream = fopen(filename, "r");
    if (stream == NULL) {
        fprintf(stderr, "Could not open \"%s\" for reading: %s.\n", filename, strerror(errno));
        if (error_is_fatal) exit(EXIT_FAILURE);
        return NULL;
    }

    uint8_t bo[2];
    bool little_endian;
    if (fread(bo, 1, 2, stream) != 2) {
        debug("failed, could not read 2 byte header\n");
        goto fail;
    }
    if (bo[0] == 'I' && bo[1] == 'I') {
        little_endian = true;
    } else if (bo[0] == 'M' && bo[1] == 'M') {
        little_endian = false;
    } else {
        debug("unknown tiff header %d %d\n", bo[0], bo[1]);
        fprintf(stderr, "Invalid TIFF header.\n");
        goto fail;
    }

    bool do_swap;
    { // test if we need to swap data to get host order
        uint16_t t = 1;
        memcpy(bo, &t, 2);
        if ((bo[0] == 1 && !little_endian) || (bo[1] == 1 && little_endian)) {
            do_swap = true;
        } else {
            do_swap = false;
        }
    }

    // goto first IFDOffset
    if (fseek(stream, 4, SEEK_SET) == -1) {
        debug("first IFD Offset seek failed\n");
        goto fail;
    }

    uint32_t ifdoffset;
    GET4(&ifdoffset);
    debug("IFD offset %d\n", ifdoffset);
    if (fseek(stream, ifdoffset, SEEK_SET) == -1) {
        debug("IFD Offset seek to %d failed\n", ifdoffset);
        goto fail;
    }
    dcp = calloc(1, sizeof(*dcp));
    dcp->ill_count = 1;
    uint16_t tag_count;
    uint32_t hsd_count[3] = { 0,0,0 };
    GET2(&tag_count);
    for (int i = 0; i < (int)tag_count; i++) {
        uint16_t tag_id;
        uint16_t tag_type;
        uint32_t tag_count;
        uint32_t tag_data;
        uint32_t tag_ext_data_size;
        GET2(&tag_id);
        GET2(&tag_type);
        GET4(&tag_count);
        GET4(&tag_data);
        const int vsize = tag_value_size(tag_type);
        debug("tag %05d|0x%04X\t%d\t%d\t0x%08X\t(%d)\t%d [%d - %d]\n", tag_id, tag_id, tag_type, tag_count, tag_data, vsize * tag_count, (int)ftell(stream), tag_data, tag_data + vsize * tag_count);
        if (vsize * tag_count > 4) {
            const long last_pos = ftell(stream);
            if (fseek(stream, tag_data, SEEK_SET) == -1) {
                debug("a\n");
                goto fail;
            }
            tag_ext_data_size = vsize * tag_count + (tag_type == TIFF_ASCII ? 1 : 0);
            tag_ext_data = malloc(tag_ext_data_size);
            size_t res = fread(tag_ext_data, 1, vsize * tag_count, stream);
            if (res != vsize * tag_count) {
                debug("failed to read %d bytes (%d / %d) at %d, returned %zd\n", vsize * tag_count, vsize, tag_count, tag_data, res);
                goto fail;
            }
            if (fseek(stream, last_pos, SEEK_SET) == -1) {
                debug("e\n");
                goto fail;
            }
            if (tag_type == TIFF_ASCII) {
                // ascii strings are generally zero-terminated, but here we guarantee it
                tag_ext_data[tag_count] = '\0';
            }
        } else {
            uint32_t raw_data = do_swap ? bit_swap32(tag_data) : tag_data;
            tag_ext_data_size = vsize * tag_count + (tag_type == TIFF_ASCII ? 1 : 0);
            if (tag_ext_data_size < 4) {
                tag_ext_data_size = 4;
            }
            tag_ext_data = malloc(tag_ext_data_size);
            memcpy(tag_ext_data, &raw_data, 4);
        }

        switch (tag_id) {
        case TIFFTAG_UNIQUECAMERAMODEL:
            dcp->camera_model = strdup((char *)tag_ext_data);
            break;
        case TIFFTAG_REDUCTIONMATRIX1:
        case TIFFTAG_REDUCTIONMATRIX2:
            fprintf(stderr, "ReductionMatrix support not implemented!\n");
            goto fail;
            break;
        case TIFFTAG_COLORMATRIX1:
        case TIFFTAG_COLORMATRIX2:
        case TIFFTAG_FORWARDMATRIX1:
        case TIFFTAG_FORWARDMATRIX2: {
            if (tag_type != TIFF_SRATIONAL || tag_count != 9) {
                debug("unexpected matrix content");
                goto fail;
            }
            int32_t raw[9*2];
            m3x3 cm;
            memcpy(raw, tag_ext_data, sizeof(raw));
            for (int row = 0; row < 3; row++) {
                for (int col = 0; col < 3; col++) {
                    if (do_swap) {
                        raw[2*(row*3+col)+0] = bit_swap32(raw[2*(row*3+col)+0]);
                        raw[2*(row*3+col)+1] = bit_swap32(raw[2*(row*3+col)+1]);
                    }
                    cm.v[row][col] = (double)raw[2*(row*3+col)+0] / raw[2*(row*3+col)+1];
                }
            }
            switch (tag_id) {
            case TIFFTAG_COLORMATRIX1: memcpy(&dcp->cm[0], &cm, sizeof(cm)); break;
            case TIFFTAG_COLORMATRIX2: memcpy(&dcp->cm[1], &cm, sizeof(cm)); break;
            case TIFFTAG_FORWARDMATRIX1: memcpy(&dcp->fm[0], &cm, sizeof(cm)); dcp->has.fm = true; break;
            case TIFFTAG_FORWARDMATRIX2: memcpy(&dcp->fm[1], &cm, sizeof(cm)); break;
            }
            break;
        }
        case TIFFTAG_CALIBRATIONILLUMINANT1:
        case TIFFTAG_CALIBRATIONILLUMINANT2: {
            if (tag_type != TIFF_SHORT) {
                debug("unexpected illuminant content");
                goto fail;
            }
            int ill_idx = 0;
            if (tag_id == TIFFTAG_CALIBRATIONILLUMINANT2) {
                ill_idx = 1;
                dcp->ill_count = 2;
            }
            uint16_t ill;
            memcpy(&ill, tag_ext_data, 2);
            dcp->ill[ill_idx] = do_swap ? bit_swap16(ill) : ill;
            break;
        }
        case TIFFTAG_PROFILECALIBRATIONSIGNATURE:
            dcp->calibration_signature = strdup((char *)tag_ext_data);
            break;
        case TIFFTAG_PROFILENAME:
            dcp->profile_name = strdup((char *)tag_ext_data);
            break;
        case TIFFTAG_PROFILEHUESATMAPDIMS:
        case TIFFTAG_PROFILELOOKTABLEDIMS:
            if (tag_type != TIFF_LONG || tag_count != 3) {
                debug("unexpected map dims content");
                goto fail;
            }
            uint32_t dims[3];
            for (int i = 0; i < 3; i++) {
                dims[i] = do_swap ? bit_swap32(((uint32_t *)tag_ext_data)[i]) : ((uint32_t *)tag_ext_data)[i];
            }
            if (tag_id == TIFFTAG_PROFILEHUESATMAPDIMS) {
                memcpy(dcp->hsmdims, dims, sizeof(dims));
            } else {
                memcpy(dcp->lookdims, dims, sizeof(dims));
            }
            break;
        case TIFFTAG_PROFILEHUESATMAPDATA1:
        case TIFFTAG_PROFILEHUESATMAPDATA2:
        case TIFFTAG_PROFILELOOKTABLEDATA:
            if (tag_type != TIFF_FLOAT || tag_count == 0 || tag_count % 3 != 0) {
                debug("unexpected map content");
                goto fail;
            }
            v3 *map = malloc((tag_count / 3) * sizeof(*map));
            for (int i = 0; i < (int)tag_count / 3; i++) {
                float row[3];
                if (do_swap) {
                    uint32_t raw[3];
                    memcpy(raw, &((uint32_t *)tag_ext_data)[3*i], 12);
                    for (int j = 0; j < 3; j++) {
                        raw[j] = bit_swap32(raw[j]);
                    }
                    memcpy(row, raw, 12);
                } else {
                    memcpy(row, &((uint32_t *)tag_ext_data)[3*i], 12);
                }
                map[i].v[0] = row[0];
                map[i].v[1] = row[1];
                map[i].v[2] = row[2];
            }
            switch (tag_id) {
            case TIFFTAG_PROFILEHUESATMAPDATA1: dcp->huesatmap[0] = map; hsd_count[0] = tag_count; break;
            case TIFFTAG_PROFILEHUESATMAPDATA2: dcp->huesatmap[1] = map; hsd_count[1] = tag_count;  break;
            case TIFFTAG_PROFILELOOKTABLEDATA: dcp->looktable = map; hsd_count[2] = tag_count; break;
            }
            break;
        case TIFFTAG_PROFILEHUESATMAPENCODING:
        case TIFFTAG_PROFILELOOKTABLEENCODING:
            if (tag_type != TIFF_LONG || tag_count != 1) {
                debug("unexpected table encoding content\n");
                goto fail;
            }
            if (tag_id == TIFFTAG_PROFILEHUESATMAPENCODING) {
                dcp->huesatmap_srgb_gamma = !!(*(uint32_t *)tag_ext_data);
                dcp->has.huesatmap_srgb_gamma = true;
            } else {
                dcp->looktable_srgb_gamma = !!(*(uint32_t *)tag_ext_data);
                dcp->has.looktable_srgb_gamma = true;
            }
            break;
        case TIFFTAG_PROFILETONECURVE:
            if (tag_type != TIFF_FLOAT || tag_count < 4 || tag_count % 2 != 0) {
                debug("unexpected table encoding content\n");
                goto fail;
            }
            dcp->tonecurve = malloc(tag_count * sizeof(double));
            for (int i = 0; i < tag_count; i++) {
                float val;
                if (do_swap) {
                    uint32_t raw = bit_swap32(((uint32_t *)tag_ext_data)[i]);
                    memcpy(&val, &raw, 4);
                } else {
                    val = ((float *)tag_ext_data)[i];
                }
                dcp->tonecurve[i] = val;
            }
            dcp->tonecurve_len = tag_count / 2;
            break;
        case TIFFTAG_PROFILEEMBEDPOLICY:
            if (tag_type != TIFF_LONG || tag_count != 1) {
                debug("unexpected embed policy content\n");
                goto fail;
            }
            dcp->embed_policy = do_swap ? bit_swap32(*(uint32_t *)tag_ext_data) : *(uint32_t *)tag_ext_data;
            dcp->has.embed_policy = true;
            if ((unsigned)dcp->embed_policy > 3) {
                debug("unexpected embed policy content %d\n", (int)dcp->embed_policy);
                goto fail;
            }
            break;
        case TIFFTAG_PROFILECOPYRIGHT:
            dcp->copyright = strdup((char *)tag_ext_data);
            break;
        case TIFFTAG_BASELINEEXPOSUREOFFSET:
            if (tag_type != TIFF_SRATIONAL || tag_count != 1) {
                debug("unexpected baseline exposure offset content\n");
                goto fail;
            }
            int32_t raw[2];
            memcpy(raw, tag_ext_data, sizeof(raw));
            if (do_swap) {
                raw[0] = bit_swap32(raw[0]);
                raw[1] = bit_swap32(raw[1]);
            }
            dcp->baseline_exposure_offset = (double)raw[0] / raw[1];
            dcp->has.baseline_exposure_offset = true;
            break;
        case TIFFTAG_DEFAULTBLACKRENDER:
            if (tag_type != TIFF_LONG || tag_count != 1) {
                debug("unexpected default black render content\n");
                goto fail;
            }
            dcp->black_render_none = !!(*(uint32_t *)tag_ext_data);
            dcp->has.black_render_none = true;
            break;
        default:
            debug("skipping unknown tag id %05d\n", tag_id);
            break;
        }
        free(tag_ext_data);
        tag_ext_data = NULL;
    }
    GET4(&ifdoffset);
    if (ifdoffset != 0) {
        debug("expected ifd termination, got %u\n", ifdoffset);
        goto fail;
    }
    if (hsd_count[0]/3 != dcp->hsmdims[0]*dcp->hsmdims[1]*dcp->hsmdims[2] ||
        (dcp->huesatmap[1] != NULL && hsd_count[1]/3 != dcp->hsmdims[0]*dcp->hsmdims[1]*dcp->hsmdims[2]) ||
        hsd_count[2]/3 != dcp->lookdims[0]*dcp->lookdims[1]*dcp->lookdims[2])
    {
        debug("mismatching table dimensions\n");
        goto fail;
    }
    debug("success!\n");
    fclose(stream);
    return dcp;

fail:
    free(tag_ext_data);
    profio_dcp_delete(dcp);
    fclose(stream);
    fprintf(stderr, "Could not parse DCP in file \"%s\".\n", filename);
    if (error_is_fatal) exit(EXIT_FAILURE);
    return NULL;
}

static void
write_tiff_tag(FILE *stream,
               uint16_t tag_id,
               uint16_t tag_type,
               uint32_t tag_count,
               const void *tag_data,
               uint32_t *data_offset)
{
    fwrite(&tag_id, 1, 2, stream);
    fwrite(&tag_type, 1, 2, stream);
    fwrite(&tag_count, 1, 4, stream);
    const int vsize = tag_value_size(tag_type);
    if (vsize * tag_count > 4) {
        fwrite(data_offset, 1, 4, stream);
        *data_offset += (vsize * tag_count + 3) & ~0x3;
    } else {
        uint32_t data = 0;;
        if (tag_type == TIFF_BYTE || tag_type == TIFF_ASCII) {
            for (uint32_t i = 0; i < tag_count; i++) {
                data |= ((const uint8_t *)tag_data)[i] << (i*8);
            }
        } else if (tag_type == TIFF_SHORT) {
            for (uint32_t i = 0; i < tag_count; i++) {
                data |= ((const uint16_t *)tag_data)[i] << (i*16);
            }
        } else {
            memcpy(&data, tag_data, 4);
        }
        fwrite(&data, 1, 4, stream);
    }
}

static void
write_tiff_data(FILE *stream,
                const void *data,
                size_t size)
{
    if (size <= 4) {
        return;
    }
    fwrite(data, 1, size, stream);
    size_t ps = ((size + 3) & ~0x3) - size;
    if (ps > 0) {
        uint32_t pad = 0;
        fwrite(&pad, 1, ps, stream);
    }
}

static void
make_srational_matrix(const m3x3 cm,
                      int32_t srational[18])
{
    for (int row = 0; row < 3; row++) {
        for (int col = 0; col < 3; col++) {
            srational[2*(row*3+col)+0] = (int32_t)round(10000 * cm.v[row][col]);
            srational[2*(row*3+col)+1] = 10000;
        }
    }
}

static float *
make_float_hsmap(const uint32_t dims[],
                 const v3 *hsv)
{
    float *m = malloc(dims[0]*dims[1]*dims[2]*sizeof(*m)*3);
    for (int i = 0; i < dims[0]*dims[1]*dims[2]; i++) {
        m[3*i+0] = hsv[i].v[0];
        m[3*i+1] = hsv[i].v[1];
        m[3*i+2] = hsv[i].v[2];
    }
    return m;
}

bool
profio_dcp_write(const struct profio_dcp *dcp,
                 const char filename[],
                 bool error_is_fatal)
{
    FILE *stream = fopen(filename, "w");
    if (stream == NULL) {
        fprintf(stderr, "failed to open \"%s\" for writing: %s\n", filename, strerror(errno));
        if (error_is_fatal) exit(EXIT_FAILURE);
        return false;
    }

    {
        uint8_t bo[2];
        uint16_t t = 1;
        memcpy(bo, &t, 2);
        if (bo[1] == 1) {
            fwrite("MMCR", 1, 4, stream);
        } else {
            fwrite("IIRC", 1, 4, stream);
        }
    }
    uint32_t ifdoffset = 8;
    fwrite(&ifdoffset, 1, 4, stream);
    uint16_t tag_count = 0;
    if (dcp->camera_model) tag_count++;
    tag_count += dcp->ill_count;
    tag_count += dcp->ill_count;
    if (dcp->calibration_signature) tag_count++;
    if (dcp->profile_name) tag_count++;
    if (dcp->huesatmap[0] != NULL) tag_count += 2;
    if (dcp->huesatmap[1] != NULL) tag_count++;
    if (dcp->has.huesatmap_srgb_gamma) tag_count++;
    if (dcp->tonecurve != NULL) tag_count++;
    if (dcp->has.embed_policy) tag_count++;
    if (dcp->copyright) tag_count++;
    if (dcp->has.fm) tag_count += dcp->ill_count;
    if (dcp->looktable != NULL) tag_count += 2;
    if (dcp->has.looktable_srgb_gamma) tag_count++;
    if (dcp->has.baseline_exposure_offset) tag_count++;
    if (dcp->has.black_render_none) tag_count++;
    debug("tag count: %d\n", tag_count);

    fwrite(&tag_count, 1, 2, stream);

    uint32_t data_offset = 8 + 2 + tag_count * 12 + 4;
    if (dcp->camera_model) write_tiff_tag(stream, TIFFTAG_UNIQUECAMERAMODEL, TIFF_ASCII, strlen(dcp->camera_model) + 1, dcp->camera_model, &data_offset);
    write_tiff_tag(stream, TIFFTAG_COLORMATRIX1, TIFF_SRATIONAL, 9, NULL, &data_offset);
    if (dcp->ill_count == 2) {
        write_tiff_tag(stream, TIFFTAG_COLORMATRIX2, TIFF_SRATIONAL, 9, NULL, &data_offset);
    }
    write_tiff_tag(stream, TIFFTAG_CALIBRATIONILLUMINANT1, TIFF_SHORT, 1, &dcp->ill[0], NULL);
    if (dcp->ill_count == 2) {
        write_tiff_tag(stream, TIFFTAG_CALIBRATIONILLUMINANT2, TIFF_SHORT, 1, &dcp->ill[1], NULL);
    }
    if (dcp->calibration_signature) write_tiff_tag(stream, TIFFTAG_PROFILECALIBRATIONSIGNATURE, TIFF_ASCII, strlen(dcp->calibration_signature) + 1, dcp->calibration_signature, &data_offset);
    if (dcp->profile_name) write_tiff_tag(stream, TIFFTAG_PROFILENAME, TIFF_ASCII, strlen(dcp->profile_name) + 1, dcp->profile_name, &data_offset);
    if (dcp->huesatmap[0] != NULL) {
        write_tiff_tag(stream, TIFFTAG_PROFILEHUESATMAPDIMS, TIFF_LONG, 3, NULL, &data_offset);
        write_tiff_tag(stream, TIFFTAG_PROFILEHUESATMAPDATA1, TIFF_FLOAT, dcp->hsmdims[0]*dcp->hsmdims[1]*dcp->hsmdims[2]*3, NULL, &data_offset);
        if (dcp->huesatmap[1] != NULL) {
            write_tiff_tag(stream, TIFFTAG_PROFILEHUESATMAPDATA2, TIFF_FLOAT, dcp->hsmdims[0]*dcp->hsmdims[1]*dcp->hsmdims[2]*3, NULL, &data_offset);
        }
    }
    if (dcp->tonecurve) write_tiff_tag(stream, TIFFTAG_PROFILETONECURVE, TIFF_FLOAT, dcp->tonecurve_len*2, NULL, &data_offset);
    uint32_t u32;
    if (dcp->has.embed_policy) {
        u32 = dcp->embed_policy;
        write_tiff_tag(stream, TIFFTAG_PROFILEEMBEDPOLICY, TIFF_LONG, 1, &u32, &data_offset);
    }
    if (dcp->copyright) write_tiff_tag(stream, TIFFTAG_PROFILECOPYRIGHT, TIFF_ASCII, strlen(dcp->copyright) + 1, dcp->copyright, &data_offset);
    if (dcp->has.fm) {
        write_tiff_tag(stream, TIFFTAG_FORWARDMATRIX1, TIFF_SRATIONAL, 9, NULL, &data_offset);
        if (dcp->ill_count == 2) {
            write_tiff_tag(stream, TIFFTAG_FORWARDMATRIX2, TIFF_SRATIONAL, 9, NULL, &data_offset);
        }
    }
    if (dcp->looktable != NULL) {
        write_tiff_tag(stream, TIFFTAG_PROFILELOOKTABLEDIMS, TIFF_LONG, 3, NULL, &data_offset);
        write_tiff_tag(stream, TIFFTAG_PROFILELOOKTABLEDATA, TIFF_FLOAT, dcp->lookdims[0]*dcp->lookdims[1]*dcp->lookdims[2]*3, NULL, &data_offset);
    }
    if (dcp->has.huesatmap_srgb_gamma) {
        u32 = !!dcp->huesatmap_srgb_gamma;
        write_tiff_tag(stream, TIFFTAG_PROFILEHUESATMAPENCODING, TIFF_LONG, 1, &u32, NULL);
    }
    if (dcp->has.looktable_srgb_gamma) {
        u32 = !!dcp->looktable_srgb_gamma;
        write_tiff_tag(stream, TIFFTAG_PROFILELOOKTABLEENCODING, TIFF_LONG, 1, &u32, NULL);
    }
    if (dcp->has.baseline_exposure_offset) {
        write_tiff_tag(stream, TIFFTAG_BASELINEEXPOSUREOFFSET, TIFF_SRATIONAL, 1, NULL, &data_offset);
    }
    if (dcp->has.black_render_none) {
        u32 = !!dcp->black_render_none;
        write_tiff_tag(stream, TIFFTAG_DEFAULTBLACKRENDER, TIFF_LONG, 1, &u32, NULL);
    }
    ifdoffset = 0;
    fwrite(&ifdoffset, 1, 4, stream);

    int32_t matrix[9*2];
    if (dcp->camera_model) write_tiff_data(stream, dcp->camera_model, strlen(dcp->camera_model) + 1);
    make_srational_matrix(dcp->cm[0], matrix);
    write_tiff_data(stream, matrix, 9*2*4);
    if (dcp->ill_count == 2) {
        make_srational_matrix(dcp->cm[1], matrix);
        write_tiff_data(stream, matrix, 9*2*4);
    }
    if (dcp->calibration_signature) write_tiff_data(stream, dcp->calibration_signature, strlen(dcp->calibration_signature) + 1);
    if (dcp->profile_name) write_tiff_data(stream, dcp->profile_name, strlen(dcp->profile_name) + 1);
    if (dcp->huesatmap[0] != NULL) {
        float *hsmap = make_float_hsmap(dcp->hsmdims, dcp->huesatmap[0]);
        write_tiff_data(stream, dcp->hsmdims, 3*4);
        write_tiff_data(stream, hsmap, dcp->hsmdims[0]*dcp->hsmdims[1]*dcp->hsmdims[2]*3*4);
        free(hsmap);
        if (dcp->huesatmap[1] != NULL) {
            hsmap = make_float_hsmap(dcp->hsmdims, dcp->huesatmap[1]);
            write_tiff_data(stream, hsmap, dcp->hsmdims[0]*dcp->hsmdims[1]*dcp->hsmdims[2]*3*4);
        }
    }
    if (dcp->tonecurve) {
        float *tc = malloc(sizeof(*tc)*2*dcp->tonecurve_len);
        for (int i = 0; i < 2*dcp->tonecurve_len; i++) {
            tc[i] = dcp->tonecurve[i];
        }
        write_tiff_data(stream, tc, dcp->tonecurve_len*2*4);
        free(tc);
    }
    if (dcp->copyright) write_tiff_data(stream, dcp->copyright, strlen(dcp->copyright) + 1);
    if (dcp->has.fm) {
        make_srational_matrix(dcp->fm[0], matrix);
        write_tiff_data(stream, matrix, 9*2*4);
        if (dcp->ill_count == 2) {
            make_srational_matrix(dcp->fm[1], matrix);
            write_tiff_data(stream, matrix, 9*2*4);
        }
    }
    if (dcp->looktable != NULL) {
        float *hsmap = make_float_hsmap(dcp->lookdims, dcp->looktable);
        write_tiff_data(stream, dcp->lookdims, 3*4);
        write_tiff_data(stream, hsmap, dcp->lookdims[0]*dcp->lookdims[1]*dcp->lookdims[2]*3*4);
        free(hsmap);
    }
    if (dcp->has.baseline_exposure_offset) {
        int32_t srat[2] = { 10000 * dcp->baseline_exposure_offset, 10000 };
        write_tiff_data(stream, srat, 2*4);
    }

    fclose(stream);
    return true;
}

const char *
exif_lightsource_ntoa(enum exif_lightsource ls)
{
    switch (ls) {
    case lsUnknown: return "Unknown";
    case lsDaylight: return "Daylight";
    case lsFluorescent: return "Fluorescent";
    case lsTungsten: return "Tungsten";
    case lsFlash: return "Flash";
    case lsFineWeather: return "FineWeather";
    case lsCloudyWeather: return "CloudyWeather";
    case lsShade: return "Shade";
    case lsDaylightFluorescent: return "DaylightFluorescent";	// D  5700 - 7100K
    case lsDayWhiteFluorescent: return "DayWhiteFluorescent";	// N  4600 - 5500K
    case lsCoolWhiteFluorescent: return "CoolWhiteFluorescent"; // W  3800 - 4500K
    case lsWhiteFluorescent: return "WhiteFluorescent"; // WW 3250 - 3800K
    case lsWarmWhiteFluorescent: return "WarmWhiteFluorescent"; // L  2600 - 3250K
    case lsStandardLightA: return "Std A";
    case lsStandardLightB: return "Std B";
    case lsStandardLightC: return "Std C";
    case lsD55: return "D55";
    case lsD65: return "D65";
    case lsD75: return "D75";
    case lsD50: return "D50";
    case lsISOStudioTungsten: return "ISOStudioTungsten";
    case lsOther: return "Other";
    default: return NULL;
    }
}

enum exif_lightsource
exif_lightsource_aton(const char str[])
{
    if (strcasecmp(str, "Unknown") == 0) {
        return lsUnknown;
    } else if (strcasecmp(str, "Daylight") == 0) {
        return lsDaylight;
    } else if (strcasecmp(str, "Fluorescent") == 0) {
        return lsFluorescent;
    } else if (strcasecmp(str, "Tungsten") == 0) {
        return lsTungsten;
    } else if (strcasecmp(str, "Flash") == 0) {
        return lsFlash;
    } else if (strcasecmp(str, "FineWeather") == 0) {
        return lsFineWeather;
    } else if (strcasecmp(str, "CloudyWeather") == 0) {
        return lsCloudyWeather;
    } else if (strcasecmp(str, "Shade") == 0) {
        return lsShade;
    } else if (strcasecmp(str, "DaylightFluorescent") == 0) {
        return lsDaylightFluorescent;
    } else if (strcasecmp(str, "DayWhiteFluorescent") == 0) {
        return lsDayWhiteFluorescent;
    } else if (strcasecmp(str, "CoolWhiteFluorescent") == 0) {
        return lsCoolWhiteFluorescent;
    } else if (strcasecmp(str, "WhiteFluorescent") == 0) {
        return lsWhiteFluorescent;
    } else if (strcasecmp(str, "WarmWhiteFluorescent") == 0) {
        return lsWarmWhiteFluorescent;
    } else if (strcasecmp(str, "Std A") == 0 || strcasecmp(str, "StdA") == 0) {
        return lsStandardLightA;
    } else if (strcasecmp(str, "Std B") == 0 || strcasecmp(str, "StdB") == 0) {
        return lsStandardLightB;
    } else if (strcasecmp(str, "Std C") == 0 || strcasecmp(str, "StdC") == 0) {
        return lsStandardLightC;
    } else if (strcasecmp(str, "D55") == 0) {
        return lsD55;
    } else if (strcasecmp(str, "D65") == 0) {
        return lsD65;
    } else if (strcasecmp(str, "D75") == 0) {
        return lsD75;
    } else if (strcasecmp(str, "D50") == 0) {
        return lsD50;
    } else if (strcasecmp(str, "ISOStudioTungsten") == 0) {
        return lsISOStudioTungsten;
    } else if (strcasecmp(str, "Other") == 0) {
        return lsOther;
    } else {
        char *ep = NULL;
        int ill = (int)strtol(str, &ep, 10);
        if (*ep != '\0' || exif_lightsource_ntoa(ill) == NULL) {
            return -1;
        }
        return (enum exif_lightsource)ill;
    }
    return -1;
}

double
exif_lightsource_temp(enum exif_lightsource ls)
{
    switch (ls) {
    case lsStandardLightA:
    case lsTungsten:
        return 2850.0;
    case lsISOStudioTungsten:
        return 3200.0;
    case lsD50:
        return 5000.0;
    case lsD55:
    case lsDaylight:
    case lsFineWeather:
    case lsFlash:
    case lsStandardLightB:
        return 5500.0;
    case lsD65:
    case lsStandardLightC:
    case lsCloudyWeather:
        return 6500.0;
    case lsD75:
    case lsShade:
        return 7500.0;
    case lsDaylightFluorescent:
        return (5700.0 + 7100.0) * 0.5;
    case lsDayWhiteFluorescent:
        return (4600.0 + 5500.0) * 0.5;
    case lsCoolWhiteFluorescent:
    case lsFluorescent:
        return (3800.0 + 4500.0) * 0.5;
    case lsWhiteFluorescent:
        return (3250.0 + 3800.0) * 0.5;
    case lsWarmWhiteFluorescent:
        return (2600.0 + 3250.0) * 0.5;
    default:
        return 0.0;
    }
}

void
profio_dcp_json(struct profio_dcp *dcp,
                FILE *stream)
{
    if (dcp == NULL || stream == NULL) {
        return;
    }
    fprintf(stream, "{\n");
    if (dcp->camera_model) {
        fprintf(stream, "  \"UniqueCameraModel\": \"%s\"", dcp->camera_model);
    }
    if (dcp->profile_name) {
        fprintf(stream, ",\n"
                "  \"ProfileName\": \"%s\"", dcp->profile_name);
    }
    if (dcp->copyright) {
        fprintf(stream, ",\n"
                "  \"ProfileCopyright\": \"%s\"", dcp->copyright);
    }
    if (dcp->has.embed_policy) {
        const char *ep = "Allow Copying";
        switch (dcp->embed_policy) {
        case pepAllowCopying: ep = "Allow copying"; break;
        case pepEmbedIfUsed: ep = "Embed if used"; break;
        case pepEmbedNever: ep = "Embed never"; break;
        case pepNoRestrictions: ep = "No restrictions"; break;
        }
        fprintf(stream, ",\n"
                "  \"ProfileEmbedPolicy\": \"%s\"", ep);
    }
    if (dcp->calibration_signature) {
        fprintf(stream, ",\n"
                "  \"CalibrationSignature\": \"%s\"", dcp->calibration_signature);
    }
    for (int i = 0; i < dcp->ill_count; i++) {
        fprintf(stream, ",\n"
                "  \"CalibrationIlluminant%d\": \"", i+1);
        const char *ill = exif_lightsource_ntoa(dcp->ill[i]);
        if (ill == NULL) {
            fprintf(stream, "Undefined %d", dcp->ill[i]);
        } else {
            fprintf(stream, "%s", ill);
        }
        fprintf(stream, "\"");
    }
    for (int i = 0; i < dcp->ill_count; i++) {
        fprintf(stream,
                ",\n"
                "  \"ColorMatrix%d\": [\n"
                "    [ %9.6f, %9.6f, %9.6f ],\n"
                "    [ %9.6f, %9.6f, %9.6f ],\n"
                "    [ %9.6f, %9.6f, %9.6f ]\n"
                "  ]",
                i+1,
                dcp->cm[i].v[0][0], dcp->cm[i].v[0][1], dcp->cm[i].v[0][2],
                dcp->cm[i].v[1][0], dcp->cm[i].v[1][1], dcp->cm[i].v[1][2],
                dcp->cm[i].v[2][0], dcp->cm[i].v[2][1], dcp->cm[i].v[2][2]);
    }
    if (dcp->has.fm) {
        for (int i = 0; i < dcp->ill_count; i++) {
            fprintf(stream,
                    ",\n"
                    "  \"ForwardMatrix%d\": [\n"
                    "    [ %9.6f, %9.6f, %9.6f ],\n"
                    "    [ %9.6f, %9.6f, %9.6f ],\n"
                    "    [ %9.6f, %9.6f, %9.6f ]\n"
                    "  ]",
                    i+1,
                    dcp->fm[i].v[0][0], dcp->fm[i].v[0][1], dcp->fm[i].v[0][2],
                    dcp->fm[i].v[1][0], dcp->fm[i].v[1][1], dcp->fm[i].v[1][2],
                    dcp->fm[i].v[2][0], dcp->fm[i].v[2][1], dcp->fm[i].v[2][2]);
        }
    }
    if (dcp->has.black_render_none) {
        fprintf(stream,
                ",\n"
                "  \"DefaultBlackRender\": \"%s\"",
                dcp->black_render_none ? "None" : "Auto");
    }
    if (dcp->has.baseline_exposure_offset) {
        fprintf(stream,
                ",\n"
                "  \"BaselineExposureOffset\": %f",
                dcp->baseline_exposure_offset);
    }
    if (dcp->huesatmap[0] != NULL) {
        fprintf(stream,
                ",\n"
                "  \"ProfileHueSatMapDims\": [ %u, %u, %u ]",
                dcp->hsmdims[0], dcp->hsmdims[1], dcp->hsmdims[2]);
    }
    if (dcp->has.huesatmap_srgb_gamma) {
        fprintf(stream,
                ",\n"
                "  \"ProfileHueSatMapEncoding\": \"%s\"",
                dcp->huesatmap_srgb_gamma ? "sRGB" : "Linear");
    }
    if (dcp->looktable != NULL) {
        fprintf(stream,
                ",\n"
                "  \"ProfileLookTableDims\": [ %u, %u, %u ]",
                dcp->lookdims[0], dcp->lookdims[1], dcp->lookdims[2]);
    }
    if (dcp->has.looktable_srgb_gamma) {
        fprintf(stream,
                ",\n"
                "  \"ProfileLookTableEncoding\": \"%s\"",
                dcp->looktable_srgb_gamma ? "sRGB" : "Linear");
    }
    for (int i = 0; i < dcp->ill_count; i++) {
        if (dcp->huesatmap[i] == NULL) {
            continue;
        }
        fprintf(stream,
                ",\n"
                "  \"ProfileHueSatMap%d\": [\n",
                i+1);
        int j = 0;
        for (int h = 0; h < dcp->hsmdims[0]; h++) {
            for (int s = 0; s < dcp->hsmdims[1]; s++) {
                for (int v = 0; v < dcp->hsmdims[2]; v++, j++) {
                    fprintf(stream, "    { \"HueDiv\": %2u, \"SatDiv\": %2u, \"ValDiv\": %2u, \"HueShift\": %10.6f, \"SatScale\": %.6f, \"ValScale\": %.6f }",
                            h, s, v, dcp->huesatmap[i][j].v[0], dcp->huesatmap[i][j].v[1], dcp->huesatmap[i][j].v[2]);
                    if (j < dcp->hsmdims[0]*dcp->hsmdims[1]*dcp->hsmdims[2]-1) {
                        fprintf(stream, ",\n");
                    }
                }
            }
        }
        fprintf(stream, "\n  ]");
    }
    if (dcp->looktable != NULL) {
        fprintf(stream,
                ",\n"
                "  \"ProfileLookTable\": [\n");
        int j = 0;
        for (int h = 0; h < dcp->lookdims[0]; h++) {
            for (int s = 0; s < dcp->lookdims[1]; s++) {
                for (int v = 0; v < dcp->lookdims[2]; v++, j++) {
                    fprintf(stream, "    { \"HueDiv\": %2u, \"SatDiv\": %2u, \"ValDiv\": %2u, \"HueShift\": %10.6f, \"SatScale\": %.6f, \"ValScale\": %.6f }",
                            h, s, v, dcp->looktable[j].v[0], dcp->looktable[j].v[1], dcp->looktable[j].v[2]);
                    if (j < dcp->lookdims[0]*dcp->lookdims[1]*dcp->lookdims[2]-1) {
                        fprintf(stream, ",\n");
                    }
                }
            }
        }
        fprintf(stream, "\n  ]");
    }
    if (dcp->tonecurve != NULL) {
        fprintf(stream,
                ",\n"
                "  \"ProfileToneCurve\": [\n");
        for (int j = 0; j < dcp->tonecurve_len; j++) {
            fprintf(stream, "    [ %f, %f ]", dcp->tonecurve[2*j+0], dcp->tonecurve[2*j+1]);
            if (j < dcp->tonecurve_len-1) {
                fprintf(stream, ",\n");
            }
        }
        fprintf(stream, "\n  ]");
    }
    fprintf(stream, "\n}\n");
}

void
profio_dcp_delete(struct profio_dcp *dcp)
{
    if (dcp == NULL) {
        return;
    }
    free(dcp->camera_model);
    free(dcp->profile_name);
    free(dcp->copyright);
    free(dcp->calibration_signature);
    free(dcp->huesatmap[0]);
    free(dcp->huesatmap[1]);
    free(dcp->looktable);
    free(dcp->tonecurve);
    free(dcp);
}
